/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_keyword_manager.h"
#include "sph_ifs_dithering_effects.h"
#include "sph_master_frame.h"
#include "sph_spectral_region.h"
#include "sph_ldt.h"
#include "sph_dataset_stats.h"
#include "sph_dataset.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ifs_keywords.h"
#include "sph_ifs_tags.h"
#include "sph_ifs_proc_apply_master_dark.h"
#include "sph_framecombination.h"

/*-----------------------------------------------------------------------------
 Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IFS_DITHERING_EFFECTS_GENERAL;
extern sph_error_code SPH_IFS_DITHERING_EFFECTS_FRAMES_MISSING;

const double SPH_IFS_DITHERING_EFFECTS_DITHER_TOLERANCE = 0.001;

static
int sph_ifs_dithering_effects_set_calib_wavelengths(
        sph_ifs_dithering_effects* self) {
    cpl_vector* lambdas = NULL;

    if (self->number_lines < 2) {
        sph_error_raise(SPH_IFS_DITHERING_EFFECTS_GENERAL, __FILE__, __func__,
                __LINE__, SPH_ERROR_ERROR,
                "Need to have at least 2 calibration wavelengths! "
                        "But number of lines was %d", self->number_lines);
        goto ERROR_EXIT;
    }

    self->calib_wavelengths = cpl_vector_new(self->number_lines);
    lambdas = self->calib_wavelengths;

    if (self->wavelength_line1 > 0.0) {
        cpl_vector_set(lambdas, 0, self->wavelength_line1);
    } else {
        sph_error_raise(SPH_IFS_DITHERING_EFFECTS_GENERAL, __FILE__, __func__,
                __LINE__, SPH_ERROR_ERROR,
                "Need to have at least 2 calibration wavelengths! "
                        "But wavelength of line 1 was %lf",
                self->wavelength_line1);
        goto ERROR_EXIT;
    }
    if (self->wavelength_line2 > 0.0) {
        cpl_vector_set(lambdas, 1, self->wavelength_line2);
    } else {
        sph_error_raise(SPH_IFS_DITHERING_EFFECTS_GENERAL, __FILE__, __func__,
                __LINE__, SPH_ERROR_ERROR,
                "Need to have at least 2 valid calibration wavelengths! "
                        "But wavelength of line 2 was %lf",
                self->wavelength_line1);
        goto ERROR_EXIT;
    }
    if (self->wavelength_line3 > 0.0 && self->number_lines > 2) {
        cpl_vector_set(lambdas, 2, self->wavelength_line3);
    }
    if (self->wavelength_line4 > 0.0 && self->number_lines > 3) {
        cpl_vector_set(lambdas, 3, self->wavelength_line4);
    }
    if (self->wavelength_line5 > 0.0 && self->number_lines > 4) {
        cpl_vector_set(lambdas, 4, self->wavelength_line5);
    }

    return CPL_ERROR_NONE;

    ERROR_EXIT: if (self->calib_wavelengths) {
        cpl_vector_delete(self->calib_wavelengths);
        self->calib_wavelengths = NULL;
    }
    return SPH_IFS_DITHERING_EFFECTS_GENERAL;

}

static sph_error_code sph_ifs_dithering_effects_obtain_dither(
        cpl_frameset* rawframes, double* dither_x, double* dither_y) {
    int ff = 0;
    int setx = 0;
    int sety = 0;
    double ddx = 0.0;
    cpl_frame* aframe = NULL;
    cpl_propertylist* plist = NULL;

    if (!rawframes || !dither_x || !dither_y) {
        return SPH_ERROR_RAISE_ERR(CPL_ERROR_NULL_INPUT,"Null input");
    }

    for (ff = 0; ff < cpl_frameset_get_size(rawframes); ++ff) {
        aframe = cpl_frameset_get_position(rawframes, ff);
        plist = sph_keyword_manager_load_properties(
                cpl_frame_get_filename(aframe), 0);
        if (plist) {
            if (cpl_propertylist_has(plist, SPH_COMMON_KEYWORD_SPH_DITHERX)) {
                if (setx == 0) {
                    setx = 1;
                    *dither_x = cpl_propertylist_get_double(plist,
                            SPH_COMMON_KEYWORD_SPH_DITHERX);
                } else {
                    ddx = *dither_x
                            - cpl_propertylist_get_double(plist,
                                    SPH_COMMON_KEYWORD_SPH_DITHERX);
                    if (fabs(ddx)
                            > SPH_IFS_DITHERING_EFFECTS_DITHER_TOLERANCE) {
                        cpl_propertylist_delete(plist);
                        plist = NULL;
                        return CPL_ERROR_INCOMPATIBLE_INPUT;
                    }
                }
                if (sety == 0) {
                    sety = 1;
                    *dither_y = cpl_propertylist_get_double(plist,
                            SPH_COMMON_KEYWORD_SPH_DITHERY);
                } else {
                    ddx = *dither_y
                            - cpl_propertylist_get_double(plist,
                                    SPH_COMMON_KEYWORD_SPH_DITHERY);
                    if (fabs(ddx)
                            > SPH_IFS_DITHERING_EFFECTS_DITHER_TOLERANCE) {
                        cpl_propertylist_delete(plist);
                        plist = NULL;
                        return CPL_ERROR_INCOMPATIBLE_INPUT;
                    }
                }
            }
            cpl_propertylist_delete(plist);
            plist = NULL;
        } else {
            return SPH_ERROR_RAISE_ERR(
                    CPL_ERROR_DATA_NOT_FOUND,
                    "Could not load propertylist");
        }
    }
    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ifs_dithering_effects_run Run the Instrument Flat Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * instrument flat recipe for IFS.
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ifs_dithering_effects.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/
/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list
 @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static sph_pixel_description_table*
sph_ifs_dithering_effects_process_prism(sph_ifs_dithering_effects* self,
        cpl_frameset* rawframes, sph_pixel_description_table* pdt,
        sph_master_frame* iff, sph_master_frame* dark) {
    sph_error_code rerr = CPL_ERROR_NONE;
    sph_master_frame* mframe = NULL;
    sph_spectral_region* reg = NULL;
    sph_spectrum* spec = NULL;
    cpl_vector* wavs = NULL;
    cpl_vector* diffs = NULL;

    int sid = 0;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;

    if (!pdt) {
        SPH_ERR("Absolutely need the PDT to proceed");
        return NULL;
    }
    mframe = sph_framecombination_master_frame_from_cpl_frameset(rawframes,
            self->coll_alg, self->framecomb_parameterlist);
    if (!mframe) {
        SPH_ERR("Could not collapse raws.");
        return NULL;
    }
    if (dark) {
        rerr = sph_ifs_proc_apply_master_dark(mframe, dark);
        if (rerr != CPL_ERROR_NONE) {
            SPH_INFO_MSG("Could not subtract dark.");
            SPH_RAISE_CPL_RESET;
        }
    }
    if (iff) {
        rerr = sph_master_frame_divide_master_frame(mframe, iff);
        if (rerr != CPL_ERROR_NONE) {
            SPH_ERR("Could not divide flat.");
            sph_master_frame_delete(mframe);
            mframe = NULL;
            return NULL;
        }
    }
    for (sid = 0; sid < pdt->nregions; ++sid) {
        reg = sph_pixel_description_table_get_region(pdt, sid + 1);
        if (reg) {
            sph_error_raise(SPH_IFS_DITHERING_EFFECTS_GENERAL, __FILE__,
                    __func__, __LINE__, SPH_ERROR_INFO,
                    "Processing spectrum region %d, specid %d, lensid %d",
                    sid + 1, reg->specid, reg->lensid);
            spec = sph_pixel_description_table_extract_spectrum_mframe(pdt,
                    reg->specid, mframe, NULL);
            if (spec) {
                wavs = sph_spectrum_fit_wavs(spec, self->calib_wavelengths,
                        self->polyfit_order, 4, &diffs, NULL, NULL);
                if (wavs && diffs) {
                    sph_pixel_description_table_set_spectra_wavelengths(pdt,
                            reg->specid, wavs, 0);
                    if (cpl_error_get_code() == CPL_ERROR_INCOMPATIBLE_INPUT) {
                        SPH_RAISE_CPL_RESET;
                    }
                    if (cpl_error_get_code() == CPL_ERROR_SINGULAR_MATRIX) {
                        SPH_RAISE_CPL_RESET;
                    }
                } else {
                    if (cpl_error_get_code() == CPL_ERROR_SINGULAR_MATRIX) {
                        SPH_RAISE_CPL_RESET;
                    }
                    cpl_vector_fill(spec->bads, 1.0);
                    cpl_vector_fill(spec->ncomb, 0.0);
                    cpl_vector_fill(spec->rms, 0.0);
                    cpl_vector_fill(spec->wavelengths, 0.0);
                    cpl_vector_fill(spec->spec, 0.0);
                    sph_pixel_description_table_set_spectra_wavelengths(pdt,
                            reg->specid, spec->wavelengths, 0);
                    if (cpl_error_get_code() == CPL_ERROR_INCOMPATIBLE_INPUT) {
                        SPH_RAISE_CPL_RESET;
                    }
                }
            } else {
                SPH_RAISE_CPL_RESET;
            }
        }
    }

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;

    return pdt;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list
 @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ifs_dithering_effects_run(sph_ifs_dithering_effects* self) {
    sph_pixel_description_table* specpos = NULL;
    sph_pixel_description_table* pdt = NULL;
    sph_master_frame* iff = NULL;
    sph_master_frame* dark = NULL;
    sph_master_frame* mframe = NULL;
    sph_pixel_description_table* shifted = NULL;
    sph_ifs_lenslet_model* model = NULL;
    cpl_image* waveim1 = NULL;
    cpl_image* waveim2 = NULL;
    cpl_propertylist* plist = NULL;
    double dither_x = 0.0;
    double dither_y = 0.0;
    sph_error_code rerr = CPL_ERROR_NONE;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE

    rerr = sph_ifs_dithering_effects_set_calib_wavelengths(self);
    if (rerr != CPL_ERROR_NONE) {
        sph_error_raise(SPH_IFS_DITHERING_EFFECTS_GENERAL, __FILE__, __func__,
                __LINE__, SPH_ERROR_ERROR,
                "Could not process calibration wavelengths. "
                        "Will not process frames.");

        return rerr;
    }
    if (self->master_dark_frame) {
        SPH_INFO_MSG("Loading master frame...");
        dark = sph_master_frame_load(
                cpl_frame_get_filename(self->master_dark_frame), 0);
    }
    if (self->master_iff_frame) {
        SPH_INFO_MSG("Loading master IFF frame...");
        iff = sph_master_frame_load(
                cpl_frame_get_filename(self->master_iff_frame), 0);
    }
    if (self->specpos_frame) {
        SPH_INFO_MSG("Loading specpos frame...");
        specpos = sph_pixel_description_table_load(
                cpl_frame_get_filename(self->specpos_frame));
        plist = sph_keyword_manager_load_properties(
                cpl_frame_get_filename(self->specpos_frame), 0);
        model = sph_ifs_lenslet_model_new_from_propertylist(plist);
        cpl_propertylist_delete(plist);
        plist = NULL;
    }
    rerr = sph_ifs_dithering_effects_obtain_dither(self->rawframes, &dither_x,
            &dither_y);
    if (rerr == CPL_ERROR_INCOMPATIBLE_INPUT) {
        sph_error_raise(
                CPL_ERROR_INCOMPATIBLE_INPUT,
                __FILE__,
                __func__,
                __LINE__,
                SPH_ERROR_ERROR,
                "The raw frames do not all have the same "
                        "dithering positions. I cant process this. "
                        "Please check the headers of the inpput raw frames and only "
                        "use frames as input that have the same value for the"
                        " %s and %s keywords.", SPH_COMMON_KEYWORD_SPH_DITHERX,
                SPH_COMMON_KEYWORD_SPH_DITHERY);
    } else if (rerr == CPL_ERROR_NONE) {
        SPH_ERROR_RAISE_INFO(CPL_ERROR_CONTINUE, "Construction shifted PDT "
        "from raw data with ditherx=%f dithery=%f.", dither_x, dither_y);
        shifted = sph_pixel_description_table_new_shift(specpos, model,
                dither_x, dither_y);
        sph_pixel_description_table_save(shifted, "shifted_pdt0.fits", model);
        pdt = sph_ifs_dithering_effects_process_prism(self, self->rawframes,
                shifted, iff, dark);
        sph_pixel_description_table_save(pdt, "shifted_pdt.fits", model);
        if (pdt) {
            waveim1 = sph_pixel_description_table_get_wavimage(pdt);
            sph_pixel_description_table_delete(specpos);
            specpos = NULL;
            sph_pixel_description_table_delete(shifted);
            shifted = NULL;
            specpos = sph_pixel_description_table_load(
                    cpl_frame_get_filename(self->wave_calib_frame));
            shifted = sph_pixel_description_table_new_shift(specpos, model,
                    dither_x, dither_y);
            waveim2 = sph_pixel_description_table_get_wavimage(shifted);
            mframe = sph_master_frame_new_from_cpl_image(waveim1);
            cpl_image_subtract(mframe->image, waveim2);
            sph_master_frame_save_dfs(mframe, self->dithering_effects_filename,
                    self->inframes, cpl_frameset_get_first(self->rawframes),
                    self->inparams, SPH_IFS_TAG_DITHERING_EFFECTS_CALIB,
                    SPH_RECIPE_NAME_IFS_DITHERING_EFFECTS,
                    SPH_PIPELINE_NAME_IFS, NULL);
            cpl_image_save(waveim1, self->dithering_effects_filename,
                    CPL_TYPE_DOUBLE, NULL, CPL_IO_EXTEND);
            cpl_image_save(waveim2, self->dithering_effects_filename,
                    CPL_TYPE_DOUBLE, NULL, CPL_IO_EXTEND);
            sph_master_frame_delete(mframe);
            mframe = NULL;
            cpl_image_delete(waveim1);
            waveim1 = NULL;
            cpl_image_delete(waveim2);
            waveim2 = NULL;
        }
    }
    sph_pixel_description_table_delete(specpos);
    specpos = NULL;
    sph_pixel_description_table_delete(shifted);
    shifted = NULL;
    sph_master_frame_delete(iff);
    sph_master_frame_delete(dark);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
}

/**@}*/
