/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ifs_keywords.h"
#include "sph_dataset.h"
#include "sph_dataset_stats.h"
#include "sph_ifs_flux_calib.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ifs_tags.h"
#include "sph_ldt.h"
#include "sph_framecombination.h"
#include "sph_framesorter.h"
#include "sph_keyword_manager.h"
#include "sph_create_super_flat.h"
#include "sph_filemanager.h"
#include "sph_ifs_science.h"
#include <strings.h>
/*-----------------------------------------------------------------------------
 Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IFS_FLUX_CALIB_GENERAL;
extern sph_error_code SPH_IFS_FLUX_CALIB_FRAMES_MISSING;

static cpl_frameset*
sph_ifs_flux_calib_extract_frames(cpl_frameset* inframes, const char* keyword,
        const char* keywordval) {
    cpl_frameset* result = NULL;
    cpl_frame* aframe = NULL;
    cpl_propertylist* plist = NULL;
    int ff;
    result = cpl_frameset_new();
    for (ff = 0; ff < cpl_frameset_get_size(inframes); ++ff) {
        aframe = cpl_frameset_get_position(inframes, ff);
        if (aframe) {
            plist = sph_keyword_manager_load_properties(
                    cpl_frame_get_filename(aframe), 0);
            if (cpl_propertylist_has(plist, keyword)) {
                if (strcasecmp(cpl_propertylist_get_string(plist, keyword),
                        keywordval) == 0) {
                    cpl_frameset_insert(result, cpl_frame_duplicate(aframe));
                }
            }
            cpl_propertylist_delete(plist);
            plist = NULL;
            cpl_error_reset();
        }
    }
    if (cpl_frameset_get_size(result) < 1) {
        cpl_frameset_delete(result);
        result = NULL;
    }
    SPH_RAISE_CPL_RESET;
    return result;
}

static
double sph_ifs_flux_calib_get_flux_companion(sph_master_frame* mframe,
        double sigma, double min_radius, double max_radius) {
    cpl_apertures* aps = NULL;
    cpl_image* image = NULL;
    int nobs = 0;
    int nn = 0;
    int ngoodobs = 0;
    int idbrightest = 0;
    double x = 0.0;
    double y = 0.0;
    double cx = 0.0;
    double cy = 0.0;
    double flux = 0.0;
    cpl_vector* idgoods = NULL;

    SPH_RAISE_CPL_RESET
    image = sph_master_frame_extract_image(mframe, 1);
    cpl_image_fill_rejected(image, 0.0);
    cx = cpl_image_get_size_x(image) / 2.0;
    cy = cpl_image_get_size_y(image) / 2.0;
    idgoods = cpl_vector_new(1);
    if (image) {
        aps = cpl_apertures_extract_sigma(image, sigma);
        if (aps)
            nobs = cpl_apertures_get_size(aps);
        if (nobs < 1) {
            SPH_ERROR_RAISE_ERR( CPL_ERROR_DATA_NOT_FOUND, "Found no object "
            "on the reduced image using a sigma of %f.", sigma);
            return 0.0;
        }
        for (nn = 0; nn < nobs; ++nn) {
            x = cpl_apertures_get_centroid_x(aps, nn + 1);
            y = cpl_apertures_get_centroid_y(aps, nn + 1);
            flux = cpl_apertures_get_flux(aps, nn + 1);
            if ((x - cx) * (x - cx) + (y - cy) * (y - cy)
                    > min_radius * min_radius
                    && (x - cx) * (x - cx) + (y - cy) * (y - cy)
                            < max_radius * max_radius) {
                cpl_vector_set(idgoods, ngoodobs, nn + 1);
                ngoodobs++;
                if (idbrightest < 1)
                    idbrightest = nn + 1;
                else if (cpl_apertures_get_flux(aps, idbrightest) < flux) {
                    idbrightest = nn + 1;
                }
                cpl_vector_set_size(idgoods, ngoodobs);
            }
        }
        if (ngoodobs < 1) {
            SPH_ERROR_RAISE_ERR( CPL_ERROR_DATA_NOT_FOUND, "Found no object "
            "on the reduced image using a sigma of %f between"
            " the radii of %f and %f.", sigma, min_radius, max_radius);
            return 0.0;
        }
        if (ngoodobs > 2) {
            SPH_ERROR_RAISE_INFO(
                    CPL_ERROR_INCOMPATIBLE_INPUT,
                    "Found %d onjects "
                    "on the reduced image using a sigma of %f between"
                    " the radii of %f and %f. Am assuming the brightest one "
                    " (at coords %f,%f) "
                    "is the companion to use for flux calibration.", ngoodobs, sigma, min_radius, max_radius, cpl_apertures_get_centroid_x(aps,idbrightest), cpl_apertures_get_centroid_y(aps,idbrightest));
        }
        return cpl_apertures_get_flux(aps, idbrightest);
    }
    SPH_RAISE_CPL;
    return 0.0;
}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ifs_flux_calib_run Run the Instrument Flat Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * instrument flat recipe for IFS.
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ifs_flux_calib.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/
cpl_error_code sph_ifs_flux_calib_run(sph_ifs_flux_calib* self) {
    sph_ldt* ldt = NULL;
    sph_master_frame* result_no_coro = NULL;
    sph_master_frame* result_coro = NULL;
    double flux_coro = 0.0;
    double flux_nocoro = 0.0;
    cpl_frameset* rawframes_no_coro = NULL;
    cpl_frameset* rawframes_coro = NULL;
    sph_ifs_lenslet_model* lensmodel = NULL;
    cpl_propertylist* plist = NULL;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE

    plist = sph_keyword_manager_load_properties(
            cpl_frame_get_filename(self->wave_calib_frame), 0);
    lensmodel = sph_ifs_lenslet_model_new_from_propertylist(plist);

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE

    rawframes_no_coro = sph_ifs_flux_calib_extract_frames(
            self->raw_science_frameset, SPH_IFS_KEYWORD_CORONO_SETTING,
            SPH_IFS_KEYWORD_VALUE_CORONO_SETTING_OUT);
    rawframes_coro = sph_ifs_flux_calib_extract_frames(
            self->raw_science_frameset, SPH_IFS_KEYWORD_CORONO_SETTING,
            SPH_IFS_KEYWORD_VALUE_CORONO_SETTING_IN);

    if (!rawframes_coro) {
        SPH_ERROR_RAISE_ERR(
                CPL_ERROR_ILLEGAL_INPUT,
                "The recipe needs at least one raw frame with tag %s "
                "and keyword %s set to %s. ", SPH_IFS_TAG_FLUX_CALIB_RAW, SPH_IFS_KEYWORD_CORONO_SETTING, SPH_IFS_KEYWORD_VALUE_CORONO_SETTING_IN);
        return CPL_ERROR_ILLEGAL_INPUT;
    }
    if (!rawframes_no_coro) {
        SPH_ERROR_RAISE_ERR(
                CPL_ERROR_ILLEGAL_INPUT,
                "The recipe needs at least one raw frame with tag %s "
                "and keyword %s set to %s. ", SPH_IFS_TAG_FLUX_CALIB_RAW, SPH_IFS_KEYWORD_CORONO_SETTING, SPH_IFS_KEYWORD_VALUE_CORONO_SETTING_OUT);
        return CPL_ERROR_ILLEGAL_INPUT;
    }

    if (self->dither == 0) {
        ldt = sph_ifs_science_reduce_no_dither_no_adi(rawframes_no_coro,
                self->coll_alg, lensmodel, self->master_dark_frame, NULL,
                self->wave_calib_frame, NULL, self->master_ifu_flat_frame,
                self->master_dff_long1_frame, self->master_dff_long2_frame,
                self->master_dff_long3_frame, self->master_dff_long4_frame,
                self->master_dff_longbb_frame, self->master_dff_short_frame);
        result_no_coro = sph_ldt_collapse(ldt, 0.0, -1.0);
        sph_ldt_delete(ldt);
        ldt = NULL;
        ldt = sph_ifs_science_reduce_no_dither_no_adi(rawframes_coro,
                self->coll_alg, lensmodel, self->master_dark_frame, NULL,
                self->wave_calib_frame, NULL, self->master_ifu_flat_frame,
                self->master_dff_long1_frame, self->master_dff_long2_frame,
                self->master_dff_long3_frame, self->master_dff_long4_frame,
                self->master_dff_longbb_frame, self->master_dff_short_frame);
        result_coro = sph_ldt_collapse(ldt, 0.0, -1);
        sph_ldt_delete(ldt);
        ldt = NULL;
    } else {
        ldt = sph_ifs_science_reduce_dither_no_adi(rawframes_no_coro,
                self->coll_alg, lensmodel, self->master_dark_frame, NULL,
                self->wave_calib_frame, NULL, self->master_ifu_flat_frame,
                self->master_dff_long1_frame, self->master_dff_long2_frame,
                self->master_dff_long3_frame, self->master_dff_long4_frame,
                self->master_dff_longbb_frame, self->master_dff_short_frame);
        result_no_coro = sph_ldt_collapse(ldt, 0.0, -1);
        sph_ldt_delete(ldt);
        ldt = NULL;
        ldt = sph_ifs_science_reduce_dither_no_adi(rawframes_coro,
                self->coll_alg, lensmodel, self->master_dark_frame, NULL,
                self->wave_calib_frame, NULL, self->master_ifu_flat_frame,
                self->master_dff_long1_frame, self->master_dff_long2_frame,
                self->master_dff_long3_frame, self->master_dff_long4_frame,
                self->master_dff_longbb_frame, self->master_dff_short_frame);
        result_coro = sph_ldt_collapse(ldt, 0.0, -1);
        sph_ldt_delete(ldt);
        ldt = NULL;
    }
    SPH_RAISE_CPL_RESET
    if (result_coro && result_no_coro) {
        sph_master_frame_save(result_no_coro, "master_frame_no_coro.fits",
                NULL);
        sph_master_frame_save(result_coro, "master_frame_coro.fits", NULL);
        flux_coro = sph_ifs_flux_calib_get_flux_companion(result_coro,
                self->threshold_sigma, self->min_radius, self->max_radius);
        flux_nocoro = sph_ifs_flux_calib_get_flux_companion(result_no_coro,
                self->threshold_sigma, self->min_radius, self->max_radius);
        if (flux_coro > 0.0 && flux_nocoro > 0.0) {
            sph_master_frame_divide_master_frame(result_no_coro, result_coro);
            cpl_propertylist_append_double(result_no_coro->properties,
                    SPH_COMMON_KEYWORD_SPH_FLUX_COMP_NOCORO, flux_nocoro);
            cpl_propertylist_append_double(result_no_coro->properties,
                    SPH_COMMON_KEYWORD_SPH_FLUX_COMP_CORO, flux_coro);
            cpl_propertylist_append_double(result_no_coro->properties,
                    SPH_COMMON_KEYWORD_SPH_FLUX_RATIO, flux_coro / flux_nocoro);

            sph_master_frame_save(result_no_coro, self->flux_calib_filename,
                    NULL);
            sph_utils_create_calib_frame(self->flux_calib_filename,
                    CPL_FRAME_TYPE_IMAGE, CPL_FRAME_LEVEL_FINAL, self->inframes,
                    self->inparams, SPH_IFS_TAG_FLUX_CALIB_CALIB,
                    SPH_RECIPE_NAME_IFS_FLUX_CALIB, SPH_PIPELINE_NAME_IFS,
                    NULL);
        }
        sph_master_frame_delete(result_coro);
        result_coro = NULL;
        sph_master_frame_delete(result_no_coro);
        result_no_coro = NULL;
    }

    cpl_frameset_delete(rawframes_coro);
    cpl_frameset_delete(rawframes_no_coro);
    return cpl_error_get_code();
}
/**@}*/
