/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_keyword_manager.h"
#include "sph_ifs_wave_calib.h"
#include "sph_master_frame.h"
#include "sph_spectral_region.h"
#include "sph_ldt.h"
#include "sph_dataset_stats.h"
#include "sph_dataset.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ifs_keywords.h"
#include "sph_ifs_proc_apply_master_dark.h"
#include "sph_ifs_tags.h"
#include "sph_framecombination.h"

/*-----------------------------------------------------------------------------
 Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IFS_WAVE_CALIB_GENERAL;
extern sph_error_code SPH_IFS_WAVE_CALIB_FRAMES_MISSING;

#define SPH_IFS_WAVE_CALIB_LAMBTOL 3 * model->dispersion
#define SPH_IFS_WAVE_CALIB_LAMBTOL2 2 * model->dispersion
#define SPH_IFS_WAVE_CALIB_LAMBTOL1 1 * model->dispersion

static
int sph_ifs_wave_calib_set_calib_wavelengths(sph_ifs_wave_calib* self) {
    cpl_vector* lambdas = NULL;

    if (self->number_lines < 2) {
        sph_error_raise(SPH_IFS_WAVE_CALIB_GENERAL, __FILE__, __func__,
                __LINE__, SPH_ERROR_ERROR,
                "Need to have at least 2 calibration wavelengths! "
                        "But number of lines was %d", self->number_lines);
        goto ERROR_EXIT;
    }

    self->calib_wavelengths = cpl_vector_new(self->number_lines);
    lambdas = self->calib_wavelengths;

    if (self->wavelength_line1 > 0.0) {
        cpl_vector_set(lambdas, 0, self->wavelength_line1);
    } else {
        sph_error_raise(SPH_IFS_WAVE_CALIB_GENERAL, __FILE__, __func__,
                __LINE__, SPH_ERROR_ERROR,
                "Need to have at least 2 calibration wavelengths! "
                        "But wavelength of line 1 was %lf",
                self->wavelength_line1);
        goto ERROR_EXIT;
    }
    if (self->wavelength_line2 > 0.0) {
        cpl_vector_set(lambdas, 1, self->wavelength_line2);
    } else {
        sph_error_raise(SPH_IFS_WAVE_CALIB_GENERAL, __FILE__, __func__,
                __LINE__, SPH_ERROR_ERROR,
                "Need to have at least 2 valid calibration wavelengths! "
                        "But wavelength of line 2 was %lf",
                self->wavelength_line1);
        goto ERROR_EXIT;
    }
    if (self->wavelength_line3 > 0.0 && self->number_lines > 2) {
        cpl_vector_set(lambdas, 2, self->wavelength_line3);
    }
    if (self->wavelength_line4 > 0.0 && self->number_lines > 3) {
        cpl_vector_set(lambdas, 3, self->wavelength_line4);
    }
    if (self->wavelength_line5 > 0.0 && self->number_lines > 4) {
        cpl_vector_set(lambdas, 4, self->wavelength_line5);
    }

    return CPL_ERROR_NONE;

    ERROR_EXIT: if (self->calib_wavelengths) {
        cpl_vector_delete(self->calib_wavelengths);
        self->calib_wavelengths = NULL;
    }
    return SPH_IFS_WAVE_CALIB_GENERAL;

}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ifs_wave_calib_run Run the Instrument Flat Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * instrument flat recipe for IFS.
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ifs_wave_calib.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list
 @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static sph_pixel_description_table*
sph_ifs_wave_calib_process_prism(sph_ifs_wave_calib* self,
        cpl_frameset* rawframes, sph_pixel_description_table* pdt,
        sph_ifs_lenslet_model* model, sph_master_frame* iff,
        sph_master_frame* dark, cpl_mask* badspecmask, cpl_propertylist* pl) {
    sph_error_code rerr = CPL_ERROR_NONE;
    sph_master_frame* mframe = NULL;
    sph_spectral_region* reg = NULL;
    sph_spectrum* spec = NULL;
    cpl_vector* wavs = NULL;
    cpl_vector* diffs = NULL;
    cpl_vector* minw = NULL;
    cpl_vector* maxw = NULL;
    cpl_vector* deltaw = NULL;
    cpl_vector* resolving_power = NULL;
    cpl_vector* centw = NULL;
    cpl_image* dispim = NULL;
    double dispersion = 0.0;
    double flux = 0.0;
    double flux_stdev = 0.0;
    int sid = 0;
    int badcount = 0;
    int badfitcount = 0;
    int badf = 0;
    int badf1 = 0;
    int badf2 = 0;
    int goodcount = 0;

    cpl_mask* tmpmask = NULL;
    cpl_image* specidimage = NULL;
    double exptime = 1.0;

    if (!pdt || !model) {
        SPH_ERR("Absolutely need the PDT to proceed");
        return NULL;
    }
    mframe = sph_framecombination_master_frame_from_cpl_frameset(rawframes,
            self->coll_alg, self->framecomb_parameterlist);
    if (!mframe) {
        SPH_ERR("Could not collapse raws.");
        return NULL;
    }
    if (dark) {
        rerr = sph_ifs_proc_apply_master_dark(mframe, dark);
        if (rerr != CPL_ERROR_NONE) {
            SPH_INFO_MSG("Could not subtract dark.");
            SPH_RAISE_CPL_RESET;
        }
    }
    if (iff) {
        rerr = sph_master_frame_divide_master_frame(mframe, iff);
        if (rerr != CPL_ERROR_NONE) {
            SPH_ERR("Could not divide flat.");
            sph_master_frame_delete(mframe);
            mframe = NULL;
            return NULL;
        }
    }

    flux = sph_master_frame_get_mean(mframe, &flux_stdev);
    exptime = sph_utils_get_exposure_time(cpl_frameset_get_first(rawframes));

    cpl_propertylist_append_double(pl, SPH_COMMON_KEYWORD_FLAT_LAMP_FLUX,
            flux / exptime);
    cpl_propertylist_append_double(pl, SPH_COMMON_KEYWORD_FLAT_LAMP_FLUX_STDEV,
            flux_stdev / exptime);

    maxw = cpl_vector_new(pdt->nregions);
    minw = cpl_vector_new(pdt->nregions);
    centw = cpl_vector_new(pdt->nregions);
    deltaw = cpl_vector_new(pdt->nregions);
    resolving_power = cpl_vector_new(pdt->nregions);

    cpl_vector_fill(minw, 0.0);
    cpl_vector_fill(maxw, 0.0);
    cpl_vector_fill(centw, 0.0);
    cpl_vector_fill(deltaw, 0.0);
    cpl_vector_fill(resolving_power, 0.0);
    sph_pixel_description_table_setup_regions(pdt);

    specidimage = sph_pixel_description_table_get_specid_image(pdt);
    for (sid = 0; sid < pdt->nregions; ++sid) {
        reg = sph_pixel_description_table_get_region(pdt, sid + 1);
        if (reg) {
            /*sph_error_raise(SPH_IFS_WAVE_CALIB_GENERAL, __FILE__, __func__,
                    __LINE__, SPH_ERROR_INFO,
                    "Processing spectrum region %d/%d, specid %d, lensid %d",
                    sid + 1, pdt->nregions, reg->specid, reg->lensid);*/
            spec = sph_pixel_description_table_extract_spectrum_mframe(pdt,
                    reg->specid, mframe, NULL);
            if (spec) {
                wavs = sph_spectrum_fit_wavs(spec, self->calib_wavelengths,
                        self->polyfit_order, self->fit_window_size, &diffs,
                        NULL, NULL);
                if (wavs && diffs) {
                    if (cpl_error_get_code() == CPL_ERROR_SINGULAR_MATRIX) {
                        badfitcount++;
                        SPH_RAISE_CPL_RESET;
                        if (badspecmask) {
                            tmpmask = cpl_mask_threshold_image_create(
                                    specidimage, spec->specid - 0.1,
                                    spec->specid + 0.1);
                            cpl_mask_or(badspecmask, tmpmask);
                            cpl_mask_delete(tmpmask);
                            tmpmask = NULL;
                        }
                    } else if (fabs(
                            cpl_vector_get(wavs, 0)
                                    - model->minlambda) > SPH_IFS_WAVE_CALIB_LAMBTOL ||
                                    fabs(cpl_vector_get(wavs,cpl_vector_get_size(wavs)-1) - model->maxlambda) > SPH_IFS_WAVE_CALIB_LAMBTOL) {
                        cpl_vector_set(minw, goodcount + badf,
                                cpl_vector_get(spec->wavelengths, 0));
                        cpl_vector_set(
                                maxw,
                                goodcount + badf,
                                cpl_vector_get(
                                        spec->wavelengths,
                                        cpl_vector_get_size(spec->wavelengths)
                                                - 1));
                        cpl_vector_set(
                                centw,
                                goodcount + badf,
                                cpl_vector_get(
                                        spec->wavelengths,
                                        (cpl_vector_get_size(spec->wavelengths)
                                                - 1) / 2));
                        cpl_vector_set(resolving_power, goodcount + badf,
                                sph_spectrum_get_resolving_power(spec));
                        badf++;
                        sph_pixel_description_table_set_spectra_wavelengths(pdt,
                                reg->specid, wavs, 0); //temp fix
                        if (badspecmask) {
                            tmpmask = cpl_mask_threshold_image_create(
                                    specidimage, spec->specid - 0.1,
                                    spec->specid + 0.1);
                            cpl_mask_or(badspecmask, tmpmask);
                            cpl_mask_delete(tmpmask);
                            tmpmask = NULL;
                        }
                    } else {
                        SPH_ERROR_RAISE_INFO(
                                CPL_ERROR_CONTINUE,
                                "Found good spectrum %f, %f", cpl_vector_get(wavs, 0), cpl_vector_get(wavs, cpl_vector_get_size(wavs)-1));
                        sph_pixel_description_table_set_spectra_wavelengths(pdt,
                                reg->specid, wavs, self->no_spline_interpol);
                        cpl_vector_set(minw, goodcount + badf,
                                cpl_vector_get(wavs, 0));
                        cpl_vector_set(
                                maxw,
                                goodcount + badf,
                                cpl_vector_get(wavs,
                                        cpl_vector_get_size(wavs) - 1));
                        cpl_vector_set(
                                centw,
                                goodcount + badf,
                                cpl_vector_get(wavs,
                                        (cpl_vector_get_size(wavs) - 1) / 2));
                        cpl_vector_set(resolving_power, goodcount + badf,
                                sph_spectrum_get_resolving_power(spec));
                        goodcount++;
                    }
                    if (fabs(cpl_vector_get(wavs, 0) - model->minlambda)
                            > SPH_IFS_WAVE_CALIB_LAMBTOL / 3.0
                            || fabs(
                                    cpl_vector_get(wavs,
                                            cpl_vector_get_size(wavs) - 1)
                                            - model->maxlambda)
                                    > SPH_IFS_WAVE_CALIB_LAMBTOL / 3.0) {
                        badf1++;
                    }
                    if (fabs(cpl_vector_get(wavs, 0) - model->minlambda)
                            > 2.0 * SPH_IFS_WAVE_CALIB_LAMBTOL / 3.0
                            || fabs(
                                    cpl_vector_get(wavs,
                                            cpl_vector_get_size(wavs) - 1)
                                            - model->maxlambda)
                                    > 2.0 * SPH_IFS_WAVE_CALIB_LAMBTOL / 3.0) {
                        badf2++;
                    }
                } else {
                    badcount++;
                    if (cpl_error_get_code() == CPL_ERROR_SINGULAR_MATRIX) {
                        SPH_RAISE_CPL_RESET;
                    }
                    cpl_vector_fill(spec->bads, 1.0);
                    cpl_vector_fill(spec->ncomb, 0.0);
                    cpl_vector_fill(spec->rms, 0.0);
                    cpl_vector_fill(spec->wavelengths, 0.0);
                    cpl_vector_fill(spec->spec, 0.0);
                    sph_pixel_description_table_set_spectra_wavelengths(pdt,
                            reg->specid, spec->wavelengths,
                            self->no_spline_interpol);
                    if (cpl_error_get_code() == CPL_ERROR_INCOMPATIBLE_INPUT) {
                        SPH_RAISE_CPL_RESET;
                    }
                    if (badspecmask) {
                        tmpmask = cpl_mask_threshold_image_create(specidimage,
                                reg->specid - 0.1, reg->specid + 0.1);
                        cpl_mask_or(badspecmask, tmpmask);
                        cpl_mask_delete(tmpmask);
                        tmpmask = NULL;
                    }
                }
                cpl_vector_delete(wavs);
                wavs = NULL;
                cpl_vector_delete(diffs);
                diffs = NULL;
                sph_spectrum_delete(spec);
                spec = NULL;
            } else {
                cpl_error_reset();
            }
        }
    }
    dispim = sph_pixel_description_table_get_dispimage(pdt);
    dispersion = cpl_image_get_median(dispim);
    cpl_image_delete(dispim);
    dispim = NULL;
    cpl_vector_set_size(minw, goodcount + badf);
    cpl_vector_set_size(maxw, goodcount + badf);
    cpl_vector_set_size(centw, goodcount + badf);
    cpl_vector_set_size(resolving_power, goodcount + badf);
    cpl_propertylist_append_int(pl, SPH_COMMON_KEYWORD_QC_WAVE_NSINGFITS,
            badfitcount);
    cpl_propertylist_append_int(pl, SPH_COMMON_KEYWORD_QC_WAVE_NOUTOFSPEC,
            badf);
    cpl_propertylist_append_int(pl, SPH_COMMON_KEYWORD_QC_WAVE_N2DISP, badf2);
    cpl_propertylist_append_int(pl, SPH_COMMON_KEYWORD_QC_WAVE_NDISP, badf1);
    cpl_propertylist_append_int(pl, SPH_COMMON_KEYWORD_QC_WAVE_BADSPEC,
            badcount);
    cpl_propertylist_append_double(pl,
            SPH_COMMON_KEYWORD_MEDIAN_RESOLVING_POWER,
            cpl_vector_get_median(resolving_power));
    cpl_propertylist_append_double(pl, SPH_COMMON_KEYWORD_MEDIAN_DISPERSION,
            dispersion);
    cpl_propertylist_append_double(pl, SPH_COMMON_KEYWORD_MEDIAN_MINWAVEL,
            cpl_vector_get_median(minw) - 0.5 * dispersion);
    cpl_propertylist_append_double(pl, SPH_COMMON_KEYWORD_MEDIAN_MAXWAVEL,
            cpl_vector_get_median(maxw) + 0.5 * dispersion);
    cpl_propertylist_append_double(pl, SPH_COMMON_KEYWORD_MEDIAN_CENTWAVEL,
            cpl_vector_get_median(centw));

    cpl_vector_delete(maxw);
    maxw = NULL;
    cpl_vector_delete(minw);
    minw = NULL;
    cpl_vector_delete(centw);
    centw = NULL;
    cpl_vector_delete(deltaw);
    deltaw = NULL;
    cpl_vector_delete(resolving_power);
    resolving_power = NULL;
    cpl_image_delete(specidimage);
    specidimage = NULL;
    sph_error_raise(SPH_IFS_WAVE_CALIB_GENERAL, __FILE__, __func__, __LINE__,
            SPH_ERROR_INFO,
            "Had %d good spectra, with %d inaccurate by more than %f, "
                    "%d inaccurate by more than %f. "
                    "Had %d bad spectra, out of spec fits that were "
                    "inaccurate by more than %f: %d, no fit: %d out of %d",
            pdt->nregions - badcount, badf1, 
            SPH_IFS_WAVE_CALIB_LAMBTOL1, badf2, 
            SPH_IFS_WAVE_CALIB_LAMBTOL2, badcount, 
            SPH_IFS_WAVE_CALIB_LAMBTOL, badf, badfitcount, pdt->nregions);
    sph_master_frame_delete(mframe);
    mframe = NULL;
    return pdt;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list
 @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ifs_wave_calib_run(sph_ifs_wave_calib* self) {
    sph_pixel_description_table* specpos = NULL;
    sph_pixel_description_table* pdt = NULL;
    sph_master_frame* iff = NULL;
    sph_master_frame* dark = NULL;
    cpl_mask* bmask = NULL;
    cpl_image* badim = NULL;
    sph_ifs_lenslet_model* model = NULL;
    cpl_propertylist* plist = NULL;
    cpl_propertylist* pli   = NULL;
    cpl_propertylist* extpl = NULL;

    //
    sph_error_code rerr = CPL_ERROR_NONE;

    /*------------------------------------------------------------------
     -  Process calibration wavelengths
     --------------------------------------------------------------------*/
    rerr = sph_ifs_wave_calib_set_calib_wavelengths(self);
    if (rerr != CPL_ERROR_NONE) {
        sph_error_raise(SPH_IFS_WAVE_CALIB_GENERAL, __FILE__, __func__,
                __LINE__, SPH_ERROR_ERROR,
                "Could not process calibration wavelengths. "
                        "Will not process frames.");

        return rerr;
    }
    if (sph_utils_check_exposure_times_are_same(self->rawframes) == 0) {
        SPH_ERR("The exposure times of the input frames are not the same."
        "Can not process these.");
        return CPL_ERROR_ILLEGAL_INPUT;
    }

    pli = cpl_propertylist_load(cpl_frame_get_filename(cpl_frameset_get_first(self->rawframes)),
            	                    0); // read template property list of first frame to copy singular keys later
    if (self->master_dark_frame) {
        SPH_INFO_MSG("Loading master dark frame...");
        dark = sph_master_frame_load(
                cpl_frame_get_filename(self->master_dark_frame), 0);
    }
    if (self->master_iff_frame) {
        SPH_INFO_MSG("Loading master IFF frame...");
        iff = sph_master_frame_load(
                cpl_frame_get_filename(self->master_iff_frame), 0);
    }
    if (self->specpos_frame) {
        SPH_INFO_MSG("Loading specpos frame...");
        specpos = sph_pixel_description_table_load(
                cpl_frame_get_filename(self->specpos_frame));
        plist = sph_keyword_manager_load_properties(
                cpl_frame_get_filename(self->specpos_frame), 0);
        model = sph_ifs_lenslet_model_new_from_propertylist(plist);
        cpl_propertylist_delete(plist);
        plist = NULL;
    }
    /*------------------------------------------------------------------
     -  Now process all sets of prism frames separately
     --------------------------------------------------------------------*/

    plist = cpl_propertylist_new();
    bmask = cpl_mask_new(specpos->nx, specpos->ny);
    pdt = sph_ifs_wave_calib_process_prism(self, self->rawframes, specpos,
            model, iff, dark, bmask, plist);
    if (pdt) {
    	sph_utils_simple_copy_singular(pli,plist);
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Created final PDT. Measuring dispersion...");
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL, "Now saving...");
        if (sph_pixel_description_table_save_dfs(pdt, self->wave_calib_filename,
                model, self->inframes, NULL, self->inparams,
                SPH_IFS_TAG_WAVE_CALIB_CALIB, SPH_RECIPE_NAME_IFS_WAVE_CALIB,
                SPH_PIPELINE_NAME_IFS, plist) != CPL_ERROR_NONE) {
            SPH_ERROR_RAISE_ERR(CPL_ERROR_FILE_IO,
                    "Could not write main product.");
        } else {
            if (bmask) {
                badim = cpl_image_new_from_mask(bmask);
                extpl=cpl_propertylist_new();
                cpl_propertylist_update_string(extpl,SPH_COMMON_KEYWORD_EXTNAME, SPH_COMMON_KEYWORD_VALUE_PDT_BADIMAGE_EXTNAME);

                cpl_image_save(badim, self->wave_calib_filename,
                        CPL_BPP_16_UNSIGNED, extpl, CPL_IO_EXTEND);
                cpl_image_delete(badim);
                badim = NULL;
                cpl_propertylist_delete(extpl);
                extpl=NULL;
            }
        }SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL, "All done.");
    }
    sph_pixel_description_table_delete(pdt);
    pdt = NULL;
    sph_ifs_lenslet_model_delete(model);
    model = NULL;
    if(pli){
    	cpl_propertylist_delete(pli);
    	pli = NULL;
    }
    cpl_propertylist_delete(plist);
    plist = NULL;
    sph_master_frame_delete(iff);
    iff = NULL;
    sph_master_frame_delete(dark);
    dark = NULL;
    cpl_mask_delete(bmask);
    bmask = NULL;
    cpl_vector_delete(self->calib_wavelengths);
    self->calib_wavelengths = NULL;
    return rerr;
}

/**@}*/
