/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ird_keywords.h"

#include "sph_ird_tags.h"
#include "sph_ird_ins_throughput.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ird_instrument_model.h"
#include "sph_framecombination.h"
#include "sph_ird_common_science.h"
#include "sph_keyword_manager.h"
#include "sph_common_science.h"
/*-----------------------------------------------------------------------------
 Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IRD_INS_THROUGHPUT_GENERAL;
extern sph_error_code SPH_IRD_INS_THROUGHPUT_FRAMES_MISSING;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ird_ins_throughput_run Create Master Dark Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * master dakr
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ird_ins_throughput.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/
/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_ird_ins_throughput recipe. The error
 code returned is always a cpl error code (to allow maximal compatibility with
 esorex, gasgano, etc.) even if during recipe execution an error in the SPHERE
 API is the cause. In this case (and if the underlying error is not a cpl error)
 the cpl error code is set to the cpl_error_code that matches the failure
 reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ird_ins_throughput_run(sph_ird_ins_throughput* self) {
    sph_master_frame* dark = NULL;
    sph_master_frame* flat = NULL;
    sph_master_frame* left = NULL;
    sph_master_frame* right = NULL;
    sph_double_image* result = NULL;
    cpl_propertylist* plist = NULL;
    double count_left = 0.0;
    double count_right = 0.0;
    double flux_left = 0.0;
    double flux_right = 0.0;
    sph_ird_common_science* sci = NULL;
    cpl_frameset* left_frames = NULL;
    cpl_frameset* right_frames = NULL;
    cpl_image* imleft = NULL;
    cpl_image* imright = NULL;
    double exptime = 1.0;
    cpl_propertylist* pl = NULL;

    pl = sph_keyword_manager_load_properties(
            cpl_frame_get_filename(cpl_frameset_get_first(self->rawframes)), 0);
    cpl_ensure_code(pl, cpl_error_get_code());
    if (cpl_propertylist_has(pl, SPH_COMMON_KEYWORD_EXPTIME) == 0) {
        cpl_propertylist_delete(pl);
        pl = NULL;
        cpl_ensure_code(0, CPL_ERROR_DATA_NOT_FOUND);
    }

    exptime = cpl_propertylist_get_double(pl, SPH_COMMON_KEYWORD_EXPTIME);

    cpl_propertylist_delete(pl);
    pl = NULL;
    sci = sph_ird_common_science_new(self->inframes, self->inparams,
            self->dark_frame, self->flat_frame, NULL, NULL, NULL,
            self->flat_frame, SPH_RECIPE_NAME_IRD_INS_THROUGHPUT,
            SPH_IRD_TAG_INS_THROUGHPUT_CALIB);
    sph_ird_common_science_process_cubes(sci, self->rawframes, NULL,
            &left_frames, &right_frames, NULL);

    left = sph_common_science_combine(left_frames, self->coll_alg, 0, 1, -1);
    right = sph_common_science_combine(right_frames, self->coll_alg, 0, 1, -1);
    if (left && right) {
        count_left = cpl_image_get_absflux(imleft =
                sph_master_frame_extract_image(left, 1));
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Measured a total count on left side of %f. ", count_left);
        count_right = cpl_image_get_absflux(imright =
                sph_master_frame_extract_image(right, 1));

        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Measured a total count on right side of %f. ", count_right);
        flux_left = self->input_flux * 1e-10 * exptime / count_left;
        flux_right = self->input_flux * 1e-10 * exptime / count_right;
        SPH_ERROR_RAISE_INFO( SPH_ERROR_GENERAL,
                "Estimated the flux correction to "
                "be %f erg cm^-2 A^-1 per count for left side.", flux_left);
        SPH_ERROR_RAISE_INFO( SPH_ERROR_GENERAL,
                "Estimated the flux correction to "
                "be %f erg cm^-2 A^-1 per count for right side.", flux_right);
        SPH_ERROR_RAISE_WARNING(SPH_ERROR_GENERAL,
                "Please note that in the current version of this recipe"
                "observing filters etc are not taken into account.");

        sph_master_frame_delete(left);
        left = NULL;
        sph_master_frame_delete(right);
        right = NULL;
        cpl_image_delete(imleft);
        imleft = NULL;
        cpl_image_delete(imright);
        imright = NULL;
        result = sph_double_image_new_from_master_frames(left, right);
    }

    if (result) {
        cpl_propertylist_append_double(result->properties,
                SPH_IRD_KEYWORD_FLUX_RIGHT, count_right);
        cpl_propertylist_append_double(result->properties,
                SPH_IRD_KEYWORD_FLUX_LEFT, count_left);
        cpl_propertylist_append_double(result->properties,
                SPH_IRD_KEYWORD_FLCORR_RIGHT, flux_right);
        cpl_propertylist_append_double(result->properties,
                SPH_IRD_KEYWORD_FLCORR_LEFT, flux_left);
        sph_double_image_save_dfs(result, self->ins_throughput_outfilename,
                self->inframes, NULL, self->inparams,
                SPH_IRD_TAG_INS_THROUGHPUT_CALIB,
                SPH_RECIPE_NAME_IRD_INS_THROUGHPUT, SPH_PIPELINE_NAME_IRDIS,
                plist);
        sph_double_image_delete(result);
        result = NULL;
    }

    if (dark) {
        sph_master_frame_delete(dark);
        dark = NULL;
    }
    if (flat) {
        sph_master_frame_delete(flat);
        flat = NULL;
    }
    cpl_frameset_delete(left_frames);
    cpl_frameset_delete(right_frames);
    return (int) cpl_error_get_code();
}

/**@}*/

