/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ird_keywords.h"

#include "sph_ird_tags.h"
#include "sph_ird_tff.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ird_instrument_model.h"
#include "sph_framecombination.h"
#include "sph_common_science.h"
#include "sph_create_flat.h"
#include "sph_ird_common_science.h"
#include "sph_keyword_manager.h"
#include "sph_fitting.h"
#include <gsl_rng.h>
#include <gsl_randist.h>

/*-----------------------------------------------------------------------------
 Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IRD_TFF_GENERAL;
extern sph_error_code SPH_IRD_TFF_FRAMES_MISSING;

static const double SPH_IRD_TFF_SMOOTHLENGTH = 2.0;

static sph_master_frame*
sph_ird_tff_create_polyfit(sph_master_frame* mframe, int order, int nsamples) {
    sph_master_frame* polframe = NULL;
    int nx = 0;
    int ny = 0;
    cpl_vector* xvals = NULL;
    cpl_vector* yvals = NULL;
    cpl_vector* zvals = NULL;
    int xx = 0;
    int yy = 0;
    double val = 0.0;
    gsl_rng* pRNG = NULL;
    int cc = 0;
    int bpix = 0;
    cpl_polynomial* poly = NULL;
    cpl_vector* vv = NULL;

    nx = (int) cpl_image_get_size_x(mframe->image);
    ny = (int) cpl_image_get_size_y(mframe->image);

    pRNG = gsl_rng_alloc(gsl_rng_taus);
    if (!pRNG)
        return NULL;
    polframe = sph_master_frame_duplicate(mframe);
    if (!polframe)
        return NULL;
    xvals = cpl_vector_new(nsamples);
    yvals = cpl_vector_new(nsamples);
    zvals = cpl_vector_new(nsamples);
    if (!xvals || !yvals || !zvals)
        return NULL;

    cpl_image_delete(polframe->ncombmap);
    polframe->ncombmap = cpl_image_new(nx, ny, CPL_TYPE_DOUBLE);
    cpl_image_delete(polframe->badpixelmap);
    polframe->badpixelmap = cpl_image_new(nx, ny, CPL_TYPE_INT);

    xx = (int) gsl_ran_flat(pRNG, 0.0, nx);
    yy = (int) gsl_ran_flat(pRNG, 0.0, ny);

    while (cc < nsamples) {
        xx = (int) gsl_ran_flat(pRNG, 0.0, nx);
        yy = (int) gsl_ran_flat(pRNG, 0.0, ny);
        while (cpl_image_get(mframe->badpixelmap, xx + 1, yy + 1, &bpix) > 0) {
            xx = (int) gsl_ran_flat(pRNG, 0.0, nx);
            yy = (int) gsl_ran_flat(pRNG, 0.0, ny);
        }
        val = cpl_image_get(mframe->image, xx + 1, yy + 1, &bpix);
        cpl_vector_set(xvals, cc, xx);
        cpl_vector_set(yvals, cc, yy);
        cpl_vector_set(zvals, cc, val);
        val = cpl_image_get(polframe->ncombmap, xx + 1, yy + 1, &bpix);
        cpl_image_set(polframe->ncombmap, xx + 1, yy + 1, val + 1);
        cc++;
    }

    poly = sph_fitting_fit_poly2d(xvals, yvals, zvals, NULL, order, order);
    if (poly) {
        vv = cpl_vector_new(2);
        for (xx = 0; xx < nx; ++xx) {
            for (yy = 0; yy < ny; ++yy) {
                cpl_vector_set(vv, 0, xx);
                cpl_vector_set(vv, 1, yy);
                val = cpl_polynomial_eval(poly, vv);
                cpl_image_set(polframe->image, xx + 1, yy + 1, val);
                val = val - cpl_image_get(mframe->image, xx + 1, yy + 1, &bpix);
                val = val * val;
                cpl_image_set(polframe->rmsmap, xx + 1, yy + 1, val);
            }
        }
        cpl_vector_delete(vv);
        vv = NULL;
    }
    SPH_RAISE_CPL_RESET;
    return polframe;
}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ird_tff_run Create Master Dark Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * master dakr
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ird_tff.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

static cpl_propertylist*
sph_ird_tff_get_qc(sph_master_frame* temp_master);
static
void
sph_ird_tff_analyse(sph_master_frame* temp_master, cpl_image* nonlin_image,
        double* mean, double* fpn, double *rms, double* nonlin);

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_ird_tff recipe. The error
 code returned is always a cpl error code (to allow maximal compatibility with
 esorex, gasgano, etc.) even if during recipe execution an error in the SPHERE
 API is the cause. In this case (and if the underlying error is not a cpl error)
 the cpl error code is set to the cpl_error_code that matches the failure
 reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ird_tff_run(sph_ird_tff* self) {
    sph_master_frame* dark = NULL;
    sph_master_frame* flat = NULL;
    sph_master_frame* result = NULL;
    sph_master_frame* fitresult = NULL;
    cpl_propertylist* plist = NULL;
    cpl_propertylist* pli   = NULL;
    sph_ird_common_science* sci = NULL;
    cpl_frameset* total_frames = NULL;
    cpl_image* linbadpix = NULL;
    cpl_vector* lampflux_out = NULL;
    cpl_vector* counts_out = NULL;
    cpl_vector* lampfluxstdev_out = NULL;
    cpl_propertylist* pl = NULL;

    double median = 0.0;


    pli = cpl_propertylist_load(
                        cpl_frame_get_filename(cpl_frameset_get_first(self->rawframes)),
                        0); // read template property list of first frame to copy singular keys later


    if (self->robust_fit) {
        if (cpl_frameset_get_size(self->rawframes) < 5) {
            SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
                    "The robust fitting is enabled. "
                    "For this to work, a minimum of 5 raw frames needs"
                    "to be provided. Either provide more frames or"
                    "switch robust fitting off.");
            return CPL_ERROR_ILLEGAL_INPUT;
        }
    }

    sci = sph_ird_common_science_new(self->inframes, self->inparams,
            self->dark_frame, self->flat_frame, NULL, NULL, NULL,
            self->flat_frame, SPH_RECIPE_NAME_IRD_TEL_POL_OFFSET,
            SPH_IRD_TAG_TEL_POL_OFFSET_CALIB);

    sph_ird_common_science_process_cubes(sci, self->rawframes, NULL, NULL, NULL,
            &total_frames);

    result = sph_create_flat(total_frames, self->robust_fit, NULL, NULL,
            &linbadpix, self->badpix_lowtolerance, self->badpix_uptolerance,
            self->badpix_chisqtolerance, &lampflux_out, &counts_out, 0, &lampfluxstdev_out);

    plist = cpl_propertylist_new();
    if (result) {
        median = sph_master_frame_get_median(result, NULL);
        sph_master_frame_divide_double(result, median);

        sph_utils_simple_copy_singular(pli, plist);
        pl = sph_ird_tff_get_qc(result);
        cpl_propertylist_append(plist, pl);

        sph_master_frame_save_dfs(result, self->tff_outfilename, self->inframes,
                NULL, self->inparams, SPH_IRD_TAG_TFF_CALIB,
                SPH_RECIPE_NAME_IRD_TFF, SPH_PIPELINE_NAME_IRDIS, plist);
        fitresult = sph_ird_tff_create_polyfit(result, self->order, 1000);

        cpl_propertylist_delete(pl);
        pl = NULL;

        if (linbadpix) {
            pl = cpl_propertylist_new();
            cpl_propertylist_append_double(pl,
                    SPH_COMMON_KEYWORD_NUMBER_BADPIXELS,
                    cpl_image_get_flux(linbadpix));
            cpl_propertylist_append_string(pl, SPH_COMMON_KEYWORD_PRO_CATG,
                    SPH_IRD_TAG_NON_LINEAR_PIXELMAP_CALIB);
            cpl_propertylist_append_string(pl, SPH_COMMON_KEYWORD_EXTNAME,
                    "LINEAR BADPIXELMAP");
            cpl_image_save(linbadpix, self->tff_outfilename, CPL_TYPE_DOUBLE,
                    pl, CPL_IO_EXTEND);
            cpl_propertylist_delete(pl);
            pl = NULL;
        }
        sph_master_frame_delete(result);
        result = NULL;
    }
    if (fitresult) {
        sph_master_frame_save_dfs(fitresult, self->lsf_outfilename,
                self->inframes, NULL, self->inparams, SPH_IRD_TAG_LSF_CALIB,
                SPH_RECIPE_NAME_IRD_TFF, SPH_PIPELINE_NAME_IRDIS, plist);

        sph_master_frame_delete(fitresult);
        fitresult = NULL;
    }
    if (dark) {
        sph_master_frame_delete(dark);
        dark = NULL;
    }
    if (flat) {
        sph_master_frame_delete(flat);
        flat = NULL;
    }
    sph_ird_common_science_delete(sci);
    cpl_propertylist_delete(plist);
    cpl_frameset_delete(total_frames);
    cpl_image_delete(linbadpix);
    cpl_vector_delete(counts_out);
    cpl_vector_delete(lampflux_out);
    cpl_vector_delete(lampfluxstdev_out);
    if (pli) {
                cpl_propertylist_delete(pli);
                pli = NULL;
            }
    //sph_filemanager_clean();
    return (int) cpl_error_get_code();
}

static cpl_propertylist*
sph_ird_tff_get_qc(sph_master_frame* temp_master) {
    cpl_propertylist* pl = NULL;
    double mean = 0.0;
    double fpn = 0.0;
    double frms = 0.0;
    double nonlin = 0.0;

    pl = cpl_propertylist_new();
    cpl_propertylist_append_int(pl, SPH_COMMON_KEYWORD_NUMBER_BADPIXELS,
            sph_master_frame_get_nbads(temp_master));
    sph_ird_tff_analyse(temp_master, temp_master->rmsmap, &mean, &fpn, &frms,
            &nonlin);
    cpl_propertylist_update_double(pl, SPH_COMMON_KEYWORD_FLAT_MEAN_COUNT,
            mean);
    cpl_propertylist_update_double(pl, SPH_COMMON_KEYWORD_FLAT_FPN, fpn);
    cpl_propertylist_update_double(pl, SPH_COMMON_KEYWORD_FLAT_RMS, frms);
    cpl_propertylist_update_double(pl, SPH_COMMON_KEYWORD_FLAT_NONLIN_FACTOR,
            nonlin);
    return pl;
}

static
void sph_ird_tff_analyse(sph_master_frame* temp_master, cpl_image* nonlin_image,
        double* mean, double* fpn, double *rms, double* nonlin) {
    cpl_mask* nonlin_mask = NULL;
    cpl_image* smoothed = NULL;
    cpl_image* tmp = NULL;

    *mean = sph_master_frame_get_mean(temp_master, rms);
    if (nonlin_image) {
        nonlin_mask = sph_master_frame_get_badpixelmask(temp_master);
        if (nonlin_mask) {
            cpl_image_reject_from_mask(nonlin_image, nonlin_mask);
        }
        *nonlin = cpl_image_get_mean(nonlin_image);
    }
    if (fpn) {
        tmp = sph_master_frame_extract_image(temp_master, 1);
        smoothed = cpl_image_duplicate(tmp);
        sph_fft_smooth_image(&smoothed, SPH_IRD_TFF_SMOOTHLENGTH);
        cpl_image_subtract(tmp, smoothed);
        cpl_image_delete(smoothed);
        smoothed = NULL;
        *fpn = cpl_image_get_stdev(tmp);
        *fpn = *fpn / *mean; // express FPN as perecentage of mean
        cpl_image_delete(tmp);
        tmp = NULL;
    }
    cpl_mask_delete(nonlin_mask);
    nonlin_mask = NULL;

}

/**@}*/

