/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_ird_xtalk_correction.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ird_keywords.h"
#include "sph_ird_tags.h"
#include "sph_common_keywords.h"
#include "sph_shared_irdifs.h"
#include "sph_frame_validator.h"
#include "sph_xtalk.h"
#include "sph_fits.h"

/*-----------------------------------------------------------------------------
 Error Codes
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ird_master_dark_run Create Master Dark Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * master dark.
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ird_master_dark.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/
/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_ird_master_dark recipe. The error
 code returned is always a cpl error code (to allow maximal compatibility with
 esorex, gasgano, etc.) even if during recipe execution an error in the SPHERE
 API is the cause. In this case (and if the underlying error is not a cpl error)
 the cpl error code is set to the cpl_error_code that matches the failure
 reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ird_xtalk_correction_run(sph_ird_xtalk_correction* self) {
    sph_xtalk* xtalk = NULL;
    cpl_frameset* outframes = NULL;
    int npguess = 0;

    xtalk = sph_xtalk_load_pattern(cpl_frame_get_filename(self->xtalk_frame));
    if (!xtalk) {
        SPH_ERROR_RAISE_ERR(
                CPL_ERROR_FILE_IO,
                "Could not read the cross talk coefficients "
                "from the file %s.", cpl_frame_get_filename( self->xtalk_frame));
        SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
    }
    //cpl_ensure_code(cpl_image_flip(xtalk->coeffs,1) == CPL_ERROR_NONE,cpl_error_get_code());

    npguess = sph_fits_get_nplanes(
            cpl_frame_get_filename(cpl_frameset_get_first(self->rawframes)), 0);
    SPH_ERROR_RAISE_INFO(
            SPH_ERROR_GENERAL,
            "Applying cross talk correction to %d frames."
            "I am guessing they all have %d planes and"
            "if that is true this operation will take about %d"
            "seconds. Please wait....", (int)cpl_frameset_get_size(self->rawframes), npguess, npguess * (int)cpl_frameset_get_size(self->rawframes) * 3);

    outframes = cpl_frameset_new();
    sph_xtalk_apply_frameset(xtalk, self->rawframes, outframes);

    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL, "Done applying cross talk, "
    "created a new set of %d"
    " corrected frames.", (int)cpl_frameset_get_size(outframes));

    cpl_image_save(xtalk->coeffs, "coeffs.fits", CPL_TYPE_DOUBLE, NULL,
            CPL_IO_CREATE);
    cpl_image_save(xtalk->preamp_image, "preamp_image.fits", CPL_TYPE_DOUBLE,
            NULL, CPL_IO_CREATE);
    cpl_image_save(xtalk->derivim, "deriv_image.fits", CPL_TYPE_DOUBLE, NULL,
            CPL_IO_CREATE);
    sph_xtalk_delete(xtalk);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
}

/**@}*/
