/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <cpl.h>
#include "sph_error.h"
#include "sph_distortion_map.h"
#include "sph_point.h"
#include "sph_point_pattern.h"
/*----------------------------------------------------------------------------*/
/**
 * @defgroup A SPHERE API Module
 * @par Synopsis:
 * @code
 * typedef _module_ {
 * } module
 * @endcode
 * @par Desciption:
 *
 * This module provides ...
 */
/*----------------------------------------------------------------------------*/
/**@{*/
sph_distortion_map*
sph_distortion_map_new( sph_point_pattern* pA, sph_point_pattern* pB ) {
    sph_distortion_map* result = NULL;
    sph_point* pointA = NULL;
    sph_point* pointB = NULL;
    sph_distortion_vector* vv = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;
    int ii;
    result = cpl_calloc(sizeof(sph_distortion_map*),1);
    if (result == NULL ) return NULL;
    if ( !pA || !pB ) {
        return result;
    }
    if ( sph_point_pattern_get_size(pA) > 0 ) {
        pointA = sph_point_pattern_get_point(pA,0);
        pointB = sph_point_pattern_find_closest(pB,pointA->x,pointA->y);
        result->distvectlist = sph_distortion_vector_new(pointA->x,pointA->y,pointB->x-pointA->x,pointB->y-pointA->y);
        vv = result->distvectlist;
        sph_point_delete(pointA);pointA = NULL;
        sph_point_delete(pointB);pointB = NULL;
        for ( ii = 1; ii < sph_point_pattern_get_size(pA); ++ii) {
            pointA = sph_point_pattern_get_point(pA,ii);
            if ( pointA ) {
            	pointB = sph_point_pattern_find_closest(pB,pointA->x,pointA->y);
            	if ( pointB ) {
            		rerr = sph_distortion_vector_append_new(vv, pointA->x,pointA->y,pointB->x-pointA->x,pointB->y-pointA->y);
            		vv = vv->next;
                    sph_point_delete(pointB);pointB = NULL;
            	}
            	sph_point_delete(pointA);pointA = NULL;
            }
        }
    }
    return result;
}

// 0 is first element
sph_distortion_vector*
sph_distortion_map_get( sph_distortion_map* self,int n) {
    sph_distortion_vector* v = NULL;
    int cc = 0;
    if ( !self ) return NULL;
    if ( !self->distvectlist ) return NULL;
    v = self->distvectlist;
    while (v && cc < n) {
        v = v->next;
        cc++;
    }
    if ( cc == n ) return v;
    else return NULL;
}
sph_error_code
sph_distortion_map_insert_vec( sph_distortion_map* self,sph_distortion_vector* vec) {
    sph_distortion_vector* v = NULL;
    if ( !self ) return CPL_ERROR_NULL_INPUT;
    if ( !self->distvectlist ) {
        self->distvectlist = vec;
        return CPL_ERROR_NONE;
    }
    v = self->distvectlist;
    while (v->next) {
        v = v->next;
    }
    return sph_distortion_vector_append(v,vec);
}

int
sph_distortion_map_remove_above_length( sph_distortion_map* self, double length ) {
    int cc = 0;
    sph_distortion_vector* v = NULL;
    sph_distortion_vector* next = NULL;
    if ( !self ) return 0;
    v = self->distvectlist;
    while (v) {
        next = v->next;
        if ( v->dx * v->dx + v->dy * v->dy > length * length) {
            if ( v == self->distvectlist ) {
                self->distvectlist = next;
            }
            sph_distortion_vector_delete(v);
            cc++;
        }
        v = next;
    }
    return cc;
}
int
sph_distortion_map_remove_offset( sph_distortion_map* self, double* poffx, double* poffy ) {
    int cc = 0;
    sph_distortion_vector* v = NULL;
    double offx = 0.0;
    double offy = 0.0;
    if ( !self ) return 0;
    v = self->distvectlist;
    while (v) {
        offx+=v->dx;
        offy+=v->dy;
        cc++;
        v = v->next;
    }
    if ( cc ) {
        offx /= cc;
        offy /= cc;
        v = self->distvectlist;
        while (v) {
            v->dx-=offx;
            v->dy-=offy;
            v = v->next;
        }
    }
    if ( poffx ) {
        *poffx = offx;
    }
    if ( poffy ) {
        *poffy = offy;
    }
    return cc;
}
int
sph_distortion_map_get_size( sph_distortion_map* self) {
    int cc = 0;
    sph_distortion_vector* v = NULL;
    if ( !self ) return 0;
    v = self->distvectlist;
    while (v) {
        cc++;
        v = v->next;
    }
    return cc;
}

void
sph_distortion_map_delete( sph_distortion_map* self ) {
    if ( !self ) return;
    if ( self->distvectlist ) {
        sph_distortion_vector_delete_all( self->distvectlist );
    }
    cpl_free(self);
    return;
}


/**@}*/
