#include "sph_extract_angles.h"
#include "sph_time.h"
#include "sph_fits.h"
#include "sph_skycalc.h"
#include "sph_coord.h"
#include "sph_keyword_manager.h"
#include "sph_common_keywords.h"


/*----------------------------------------------------------------------------*/
/**
 * @brief Extract the angles from a raw data cube
 *
 * @return error code of the operation
 *
 * This function extracts parallactic angles for every plane in an input data cube.
 * The angles are returned as doubles and defined in the same way as for
 * skycalc (see the corresponding skycalc documention in sph_skycalc).
 *
 * The frame containing the cube is appssed as input must contain the keywords
 * for RA, DEC MJD-OBS and UTC.
 */
/*----------------------------------------------------------------------------*/
cpl_error_code
sph_extract_angles_from_cube( cpl_frame* cubeframe,
							  double offsetA,
							  double offsetB,
							  cpl_vector** output_angles)
{
	unsigned int 			nplanes = 0;
	cpl_vector* 			timev = NULL;
	double					time = 0.0;
	sph_skycalc_input 		skycal_inp[7];
	sph_skycalc_output 		skycal_out;
	cpl_propertylist*		pl = NULL;

	cpl_ensure_code( cubeframe, CPL_ERROR_NULL_INPUT);
	cpl_ensure_code( output_angles, CPL_ERROR_NULL_INPUT);

	nplanes = (unsigned int)sph_fits_get_nplanes( cpl_frame_get_filename(
									cubeframe), 0 );
	cpl_ensure_code(nplanes > 0, CPL_ERROR_ILLEGAL_INPUT );

    timev = sph_time_extract_times_from_frame( cubeframe );
	cpl_ensure_code(timev, CPL_ERROR_ILLEGAL_INPUT );
	cpl_ensure_code(cpl_vector_get_size(timev) > 0, CPL_ERROR_ILLEGAL_OUTPUT );

    *output_angles = cpl_vector_new(nplanes);

    pl = sph_keyword_manager_load_properties(cpl_frame_get_filename(cubeframe), 0 );

    for ( int i = 0; i < 7; ++i ) {
    	skycal_inp[i].command = ' ';
    	snprintf(skycal_inp[i].param_string,80," ");
    }
	sph_coord_initialise_skycalc( pl,
								  skycal_inp,
								  cpl_vector_get(timev,0),
								  &skycal_out);

    for ( unsigned int ii = 0; ii < nplanes; ++ii ) {
    	time = cpl_vector_get(timev,ii);
    	sph_coord_do_skycalc(skycal_inp,time,&skycal_out);
    	cpl_vector_set(*output_angles,ii,skycal_out.pa + offsetA + offsetB);
    }

    cpl_vector_delete(timev); timev = NULL;
    cpl_propertylist_delete(pl); pl = NULL;

	return CPL_ERROR_NONE;
}
