/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */
#include "sph_error.h"
#include "cpl.h"
#include "sph_line.h"
#include "sph_point.h"
#include <math.h>

/*----------------------------------------------------------------------------*/
/**
 @brief Create a new line

 @param x1        the x-coordinate of the start point
 @param y1        the y-coordinate of the start point
 @param x2        the x-coordinate of the end point
 @param xy2        the y-coordinate of the end point

 @return the new line or NULL

 Description: create a new line with the given endpoints.

 */
/*----------------------------------------------------------------------------*/
sph_line* sph_line_new( double x1, double y1, double x2, double y2 ) {
    sph_line* result = cpl_calloc(sizeof(sph_line),1);
    if ( !result ) {
        return NULL;
    }
    result->p1 = sph_point_new( x1, y1 );
    result->p2 = sph_point_new( x2, y2 );
    return result;
}



/*----------------------------------------------------------------------------*/
/**
 @brief Test if two lines intersect

 @param lineA        the first line
 @param lineB        the second line

 @return 1 if lines intersect otherwise 0

 Description: checks if two lines intersect. Note that lines are
 considered infinite for this test. Also, lines that overlap are
 considered to be intersecting.

 */
/*----------------------------------------------------------------------------*/
short sph_line_intersect_test( sph_line* lineA, sph_line* lineB ) {
    double dx = fabs ( lineB->p1->x - lineA->p1->x );
    double dy = fabs ( lineB->p1->y - lineA->p1->y );
    double dx2 = fabs ( lineB->p2->x - lineA->p1->x );
    double dy2 = fabs ( lineB->p2->y - lineA->p1->y );
    double adx = fabs ( lineA->p2->x - lineA->p1->x );
    double ady = fabs ( lineA->p2->y - lineA->p1->y );
    double bdx = fabs ( lineB->p2->x - lineB->p1->x );
    double bdy = fabs ( lineB->p2->y - lineB->p1->y );
    double dd = adx * dx + ady * dy;
    double ds = sqrt( (adx * adx + ady * ady) * (bdx * bdx + bdy * bdy) );
    if ( dx * dx + dy * dy < DBL_MIN ) {
        return 1;
    }
    if ( dx2 * dx2 + dy2 * dy2 < DBL_MIN ) {
        return 1;
    }
    if ( fabs( dd - ds ) < DBL_MIN ) return 1;
    return( sph_point_ccw_test( *lineA->p1, *lineB->p1, *lineB->p2) != sph_point_ccw_test( *lineA->p2, *lineB->p1, *lineB->p2 ) &&
            sph_point_ccw_test(*lineA->p1,*lineA->p2,*lineB->p1) != sph_point_ccw_test(*lineA->p1,*lineA->p2,*lineB->p2 ) );
}
/*----------------------------------------------------------------------------*/
/**
 @brief Get intersection point of two lines.

 @param lineA        the first line
 @param lineB        the second line

 @return the intersection point or NULL

 Description: returns the intersection point between two lines.
 If the lines do not intersect, NULL is returned.

 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_line_intersect_point(
        sph_line* lineA,
        sph_line* lineB,
        sph_point* result )
{
    double          x           = 0;
    double          y           = 0;
    double          x1          = 0;
    double          x2          = 0;
    double          x3          = 0;
    double          x4          = 0;
    double          y1          = 0;
    double          y2          = 0;
    double          y3          = 0;
    double          y4          = 0;
    double          denom       = 0;

    x1 = lineA->p1->x;
    x2 = lineA->p2->x;
    y1 = lineA->p1->y;
    y2 = lineA->p2->y;
    x3 = lineB->p1->x;
    x4 = lineB->p2->x;
    y3 = lineB->p1->y;
    y4 = lineB->p2->y;

    denom = ( ( x1 - x2 ) * ( y3 - y4 ) - ( y1 - y2 ) * ( x3 - x4 ) );

    if ( fabs( denom ) < DBL_MIN ) {
        return -1;
    }
    x = ( x1 * y2 - y1 * x2 ) * (x3 - x4) -
           ( x1 - x2 ) * ( x3 * y4 - y3 * x4 );

    y = ( x1 * y2 - y1 * x2 ) * (y3 - y4) -
           ( y1 - y2 ) * ( x3 * y4 - y3 * x4 );

    result->x = x / denom;
    result->y = y / denom;
    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
 @brief Delete the line and free space

 @param self    the line to delete

 @return nothing

 Description: Does just that.

 */
/*----------------------------------------------------------------------------*/
void sph_line_delete( sph_line* self ) {
    if ( self ) {
        if ( self->p1 ) {
            sph_point_delete( self->p1 );
        }
        if ( self->p2 ) {
            sph_point_delete( self->p2 );
        }
        cpl_free(self);
    }
}
