/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#include "sph_error.h"
#include "sph_common_keywords.h"
#include "sph_pixel_polyfit_table.h"
#include "sph_cube.h"
#include "sph_master_frame.h"
#include <math.h>
#include <cpl.h>
/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_pixel_polyfit_table Pixel Polynomial Fit Table
 *
 * @par Synopsis:
 * @code
typedef struct _sph_pixel_polyfit_table_
{
    int                     nx;  // number pixel x
    int                     ny;  // number pixel y
    int                     order; // Order of the polynomials used for fitting
    cpl_table*              table; // The table storing fit parameters
} sph_pixel_polyfit_table;
 * @endcode
 * @par Descirption:
 * This module provides functions to define and operate on a table
 * describing function fits to pixel behaviours.
 * It is essentially just a table, which, contains one row for every pixel.
 * The columns contrain the different fitting coefficients of the polynomial
 * fit.
 * The table can be saved and loaded as/from FITS table extensions.
 * The columns of the table are labelled: pixel-x, pixel-y for the
 * integer pixel positions, Red. Chi sfor the reduced chi squared for the
 * fit and then Order-0, Order-1, ... for all the polynomial coefficients.
 *
 * TODO: at some point this should be changed to a FITS image format, where
 * each coefficient is in one plane or extension of a FITS image file instead
 * of a FITS binary table.
 *
 */
/*----------------------------------------------------------------------------*/
/**@{*/
sph_error_code SPH_PIXEL_POLYFIT_TABLE_GENERAL      = SPH_PIXEL_POLYFIT_TABLE_ERR_START + 0;

/*-----------------------------------------------------------------------------
 Function definitions
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @brief Create a new pixel polyfit table thats empty.
 *
 * @param nx   The number of pixels in x
 * @param ny   The number of pixels in y
 * @param order The order of the polynomials used for fitting
 *
 * @return Pointer to new table or null in case of error.
 *
 * A new pixel polyfit table is created with nx * ny entries.
 *
 */
/*----------------------------------------------------------------------------*/
sph_pixel_polyfit_table*
sph_pixel_polyfit_table_new(int nx, int ny, int order ) {
    sph_pixel_polyfit_table*        self   = NULL;
    int                             ii      = 0;
    char                            czOrder[256];
    if ( nx < 0 || ny < 0 ) {
        sph_error_raise(CPL_ERROR_ILLEGAL_INPUT, __FILE__, __func__, __LINE__,
                        SPH_ERROR_ERROR, "Cant create pixel"
                                " polyfit table with"
                                " negative number of pixels.\n");
        return NULL;
    }

    self = cpl_calloc( 1, sizeof(sph_pixel_polyfit_table) );

    if ( !self ) {
        sph_error_raise(SPH_ERROR_MEMORY_FAIL, __FILE__, __func__, __LINE__,
                        SPH_ERROR_ERROR, "Cant create pixel"
                                " polyfit table "
                                " -- cant allocate space.\n");
        return NULL;
    }

    self->table = cpl_table_new( nx * ny );

    if ( !self->table ) {
        sph_error_raise(SPH_ERROR_MEMORY_FAIL, __FILE__, __func__, __LINE__,
                        SPH_ERROR_ERROR, "Cant create pixel"
                                " polyfit table "
                                " -- cant allocate space.\n");
        cpl_free(self);
        return NULL;
    }

    cpl_table_new_column( self->table, "pixel-x" , CPL_TYPE_INT );
    cpl_table_new_column( self->table, "pixel-y" , CPL_TYPE_INT );
    cpl_table_new_column( self->table, "Red. Chi" , CPL_TYPE_DOUBLE );

    for (ii = 0; ii <= order; ++ii) {
        sprintf( czOrder, "Order-%d", ii);
        cpl_table_new_column( self->table, czOrder , CPL_TYPE_DOUBLE );
    }

    self->nx = nx;
    self->ny = ny;
    self->order = order;
    return self;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Create a new pixel polyfit table using the image list
 *
 * @param imagelist the imagelist containing the coefficient images
 * @param errimage  optional image with error of fit
 *
 * @return Pointer to new table or null in case of error.
 *
 * A new pixel polyfit table is created from the input imagelist.
 *
 */
/*----------------------------------------------------------------------------*/
sph_pixel_polyfit_table*
sph_pixel_polyfit_table_new_from_imlist( cpl_imagelist* imlist, cpl_image* errimage ) {
    sph_pixel_polyfit_table*        self   = NULL;
    int                             ii      = 0;
    char                            czOrder[256];
    int                             nx       = 0;
    int                             ny       = 0;
    int                             ncoeff  = 0;
    double                          coeff   = 0.0;
    cpl_image*                      imp = NULL;
    int                             xx = 0;
    int                             yy = 0;
    int                             bpix = 0;
    double                          mse = 0.0;
    cpl_ensure(imlist,CPL_ERROR_NULL_INPUT,NULL);
    ncoeff = cpl_imagelist_get_size(imlist);
    cpl_ensure(ncoeff > 0 ,CPL_ERROR_ILLEGAL_INPUT,NULL);

    nx = cpl_image_get_size_x(cpl_imagelist_get(imlist,0));
    ny = cpl_image_get_size_y(cpl_imagelist_get(imlist,0));

    self = sph_pixel_polyfit_table_new(nx,ny,ncoeff-1);

    if ( !self ) {
        sph_error_raise(SPH_ERROR_MEMORY_FAIL, __FILE__, __func__, __LINE__,
                        SPH_ERROR_ERROR, "Cant create pixel"
                                " polyfit table "
                                " -- cant allocate space.\n");
        return NULL;
    }
    for (ii = 0; ii < ncoeff; ++ii) {
        sprintf( czOrder, "Order-%d", ii);
        imp = cpl_imagelist_get(imlist,ii);
        for (xx = 0; xx < nx; ++xx) {
            for (yy = 0; yy < ny; ++yy) {
                if (errimage) {
                    mse = cpl_image_get(errimage,xx+1,yy+1,&bpix);
                }
                coeff = cpl_image_get(imp,xx+1,yy+1,&bpix);
                if ( bpix == 0 ) {
                    cpl_table_set( self->table, czOrder, yy * ( self->nx) + xx, coeff );
                    cpl_table_set( self->table, "pixel-x", yy * ( self->nx) + xx, xx );
                    cpl_table_set( self->table, "pixel-y", yy * ( self->nx) + xx, yy );
                    cpl_table_set( self->table, "Red. Chi", yy * ( self->nx) + xx, mse );
                }
            }
        }
    }

    return self;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Set pixel as invalid.
 *
 * @param nx   The pixel in x
 * @param ny   The pixel in y
 *
 * @return ErrorCode of the operation.
 *
 * Sets the  a pixel in the table as invalid.
 *
 */
/*----------------------------------------------------------------------------*/
int
sph_pixel_polyfit_table_set_invalid(sph_pixel_polyfit_table* self,
                            int nx, int ny )
{
    int                             ii      = 0;
    char                            czOrder[256];
    int                             err         = CPL_ERROR_NONE;
    if ( !self ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }

    if ( !self->table ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }
    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
        SPH_RAISE_CPL
        return cpl_error_get_code();
    }
    for (ii = 0; ii <= self->order; ++ii) {
        sprintf( czOrder, "Order-%d", ii);
        cpl_table_set_invalid( self->table, czOrder, ny * ( self->nx) + nx );
    }
    err |= cpl_table_set_invalid( self->table, "pixel-x", ny * ( self->nx) + nx );
    err |= cpl_table_set_invalid( self->table, "pixel-y", ny * ( self->nx) + nx );
    err |= cpl_table_set_invalid( self->table, "Red. Chi", ny * ( self->nx) + nx );
    if ( err != CPL_ERROR_NONE ) {
        SPH_RAISE_CPL
    }
    return err;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief Set pixel fitting coefficients.
 *
 * @param nx   The pixel in x
 * @param ny   The pixel in y
 * @param mse  The red. chi squared of the fit
 * @param poly The polynomial to set coefficients from
 *
 * @return ErrorCode of the operation.
 *
 * Sets the fitting coefficients for a pixel in the table from an input
 * polynomial. The polynomial is assumed to have been created using the
 * cpl_polynomial_fit_1d_create() function and has to have the same order as
 * the table itself.
 *
 */
/*----------------------------------------------------------------------------*/
int
sph_pixel_polyfit_table_set(sph_pixel_polyfit_table* self,
                            int nx, int ny,
                            double mse, cpl_polynomial* poly )
{
    int                             ii      = 0;
    double                          coeff   = 0.0;
    char                            czOrder[256];
    cpl_size                             pows[1];
    int                             poldegree = 0;
    int                             err         = CPL_ERROR_NONE;
    if ( !self ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }

    if ( !poly ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }

    if ( !self->table ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }
    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
        SPH_RAISE_CPL
        return cpl_error_get_code();
    }
    poldegree = cpl_polynomial_get_degree( poly );
    if ( self->order <  poldegree ) {
        sph_error_raise( CPL_ERROR_ILLEGAL_INPUT,
                         __FILE__,
                         __func__,
                         __LINE__,
                         SPH_ERROR_ERROR,
                         "The polynomial and the pixel fit table "
                         "have different order/degree ( %d for table, "
                         "%d for polynomial). Cant proceed.", self->order,
                         poldegree );
        return CPL_ERROR_ILLEGAL_INPUT;
    }
    for (ii = 0; ii <= self->order; ++ii) {
        sprintf( czOrder, "Order-%d", ii);
        pows[0] = ii;
        if ( ii <= poldegree ) {
            coeff = cpl_polynomial_get_coeff( poly, pows );
            cpl_table_set( self->table, czOrder, ny * ( self->nx) + nx, coeff );
        }
    }
    err |= cpl_table_set( self->table, "pixel-x", ny * ( self->nx) + nx, nx );
    err |= cpl_table_set( self->table, "pixel-y", ny * ( self->nx) + nx, ny );
    err |= cpl_table_set( self->table, "Red. Chi", ny * ( self->nx) + nx, mse );
    if ( err != CPL_ERROR_NONE ) {
        SPH_RAISE_CPL
    }
    return err;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief Get fitting coefficients as sph_master_frame.
 *
 * @param self  The sph_pixel_polyfit_table
 * @param order The order of the coefficent (starting with 0)
 *
 * @return     The polynomial coefficients as a sph_master_frame or NULL in
 *             case of error.
 *
 * Gets the fitting polynomial coefficients for all pixels as an image. The
 * order starts with 0, so that requesting the 0th order gives the offset,
 * order 1 gives the linear slope etc.
 *
 * In case an error occurs a null is returned.
 *
 */
/*----------------------------------------------------------------------------*/
sph_master_frame* sph_pixel_polyfit_table_get_coeff_as_image( sph_pixel_polyfit_table* self,
                                                              int order )
{
    int                             xx      = 0;
    int                             yy      = 0;
    double                          coeff   = 0.0;
    int                             badval  = 0;
    char                            czOrder[256];
    cpl_size                             pows[1];
    double                          mse     = 0;
    sph_master_frame*               result  = NULL;


    if ( !self ) {
        SPH_NULL_ERROR;
        return NULL;
    }

    if ( ! self->table ) {
        SPH_NULL_ERROR;
        return NULL;
    }

    if ( order > self->order || order < 0 ) {
        sph_error_raise( SPH_PIXEL_POLYFIT_TABLE_GENERAL,
                         __FILE__, __func__, __LINE__,
                         SPH_ERROR_ERROR,
                         "Invalid order given. "
                         "Can only retrieve order in the range "
                         "0 to % d", self->order );
        return NULL;
    }
    result = sph_master_frame_new( self->nx, self->ny );

    if ( result == NULL ) {
        sph_error_raise( SPH_PIXEL_POLYFIT_TABLE_GENERAL,
                         __FILE__, __func__, __LINE__, SPH_ERROR_ERROR,
                         "Could not create master frame to store coefficient." );
        return NULL;
    }

    pows[0] = order;
    sprintf( czOrder, "Order-%d", order);

    for (xx = 0; xx < self->nx; ++xx) {
        for (yy = 0; yy < self->ny; ++yy) {
            badval = 0;
            if ( cpl_table_is_valid(self->table,czOrder,yy * ( self->nx ) + xx)) {
                coeff = cpl_table_get( self->table, czOrder, yy * ( self->nx ) + xx, &badval );
                cpl_image_set( result->image, xx + 1, yy + 1, coeff );
                cpl_image_set( result->badpixelmap, xx + 1, yy + 1, badval );
                mse = cpl_table_get( self->table, "Red. Chi", yy * ( self->nx ) + xx, &badval );
                cpl_image_set( result->rmsmap, xx + 1, yy + 1, mse );
            }
            else {
                mse = -1.0;
            }
            if ( mse < 0 ) {
                cpl_image_set( result->badpixelmap, xx + 1, yy + 1, 1 );
            }
        }
    }
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Get a mask of badpixels
 *
 * @param self The pixel polyfit table
 * @param badpix_lowtolerance The lower tolerance bound for "good" pixels
 * @param badpix_uptolerance The upper tolerance bound for "good" pixels
 * @param badpix_chisqtolerance The red. chi squared tolerance bound for "good" pixels
 *
 * @return a new cpl_mask with the mask set to 1 for those pixels that are bad.
 *
 * This method creates a new cpl_mask from the sph_pixel_polyfit_table using the
 * given tolerance values as criteria to decide if a pixel is good or bad.
 * The two values lowtolerance and uptolerance give the lower and upper bound
 * on the slope of the polynomial fit (1st coefficient) and the chisqtolerance
 * gives the maximum chi squared thats tolerated for a good pixel unless it is
 * set to zero in which case it is ignored.
 *
 * For example, to select all pixels as good that have a slope (1st polynomial
 * coefficient) between -2.0 and 5.0 and a chi-squared fit below 3.0 call
 * <code>cpl_mask* badpixmask = sph_pixel_polyfit_table_create_badpixel_
 * mask( pStruct, -2.0, 5.0, 3.0);</code>
 */
/*----------------------------------------------------------------------------*/
cpl_mask*
sph_pixel_polyfit_table_create_badpixel_mask( sph_pixel_polyfit_table* self,
                                              double badpix_lowtolerance,
                                              double badpix_uptolerance,
                                              double badpix_chisqtolerance)
{
    int                             xx      = 0;
    int                             yy      = 0;
    double                          coeff   = 0.0;
    int                             badval  = 0;
    char                            czOrder[256];
    cpl_size                             pows[1];
    double                          mse     = 0;
    cpl_mask*                       result  = NULL;


    if ( !self ) {
        SPH_NULL_ERROR;
        return NULL;
    }

    if ( !self->table ) {
        SPH_NULL_ERROR;
        return NULL;
    }

    if ( self->order < 1 ) {
        sph_error_raise( SPH_PIXEL_POLYFIT_TABLE_GENERAL,
                         __FILE__, __func__, __LINE__,
                         SPH_ERROR_ERROR,
                         "Invalid order . "
                         "Can only create badpix mask if order > 0 "
                         );
        return NULL;
    }

    result = cpl_mask_new( self->nx, self->ny );

    if ( result == NULL ) {
        sph_error_raise( SPH_PIXEL_POLYFIT_TABLE_GENERAL,
                         __FILE__, __func__, __LINE__, SPH_ERROR_ERROR,
                         "Could not create cpl_mask to store mask." );
        return NULL;
    }

    pows[0] = 1;
    sprintf( czOrder, "Order-1" );

    for (xx = 0; xx < self->nx; ++xx) {
        for (yy = 0; yy < self->ny; ++yy) {
            badval = 0;

            if ( cpl_table_is_valid(self->table,czOrder,yy * ( self->nx ) + xx)) {
                coeff = cpl_table_get( self->table, czOrder, yy * ( self->nx ) + xx, &badval );
                if ( coeff < badpix_uptolerance && coeff > badpix_lowtolerance ) {
                    cpl_mask_set( result, xx + 1, yy + 1 , 0 );
                }
                else {
                    cpl_mask_set( result, xx + 1, yy + 1 , 1 );
                }
                mse = cpl_table_get( self->table, "Red. Chi", yy * ( self->nx ) + xx, &badval );
                if ( mse > badpix_chisqtolerance ) {
                    cpl_mask_set( result, xx + 1, yy + 1 , 1 );
                }
            }
            else {
                cpl_mask_set( result, xx + 1, yy + 1 , 1 );
            }
        }
    }

    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Get pixel fitting coefficients.
 *
 * @param nx   The pixel in x
 * @param ny   The pixel in y
 * @param mse  The red. chi squared of the fit (returned)
 *             or NULL if not required
 * @return     The polynomial fit of that pixel or NULL on error.
 *
 * Gets the fitting polynomial for a pixel in the table
 * The polynomial is of the same type as returned by the
 * cpl_polynomial_fit_1d_create() function
 * and has to have the same order as
 * the table itself. If mse is provided, it is filled with the
 * reduced chi squared value.
 *
 */
/*----------------------------------------------------------------------------*/

cpl_polynomial*
sph_pixel_polyfit_table_get( sph_pixel_polyfit_table* self,
                             int nx, int ny,
                             double* mse )
{
    int                             ii      = 0;
    cpl_polynomial*                 result  = NULL;
    double                          coeff   = 0.0;
    int                             badval  = 0;
    char                            czOrder[256];
    cpl_size                             pows[1];

    if ( !self ) {
        SPH_NULL_ERROR;
        return NULL;
    }

    if ( ! self->table ) {
        SPH_NULL_ERROR;
        return NULL;
    }

    result = cpl_polynomial_new( 1 );
    for (ii = 0; ii <= self->order; ++ii) {
        sprintf( czOrder, "Order-%d", ii);
        if ( cpl_table_is_valid(self->table,czOrder,ny * ( self->nx ) + nx)) {
            pows[0] = ii;
            coeff = cpl_table_get( self->table, czOrder, ny * ( self->nx ) + nx, &badval );
            cpl_polynomial_set_coeff( result, pows, coeff);
            *mse = cpl_table_get( self->table, "Red. Chi", ny * ( self->nx ) + nx, &badval );
        }
    }

    return result;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Save the pixel polyfit table.
 *
 * @param self The pixel polyfit table to save
 * @param fitsfilename. The filename to save to
 * @return Error code.
 *
 * Saves the table in FITS format. The routines saves
 * the internal cpl_table to the FITS file specified.
 * The file is created ( if it
 * does not yet exist) or overwritten (if it does).
 */
/*----------------------------------------------------------------------------*/
int
sph_pixel_polyfit_table_save(sph_pixel_polyfit_table* self,
                                 const char* fitsfilename)
{
    int                 rerr        = CPL_ERROR_NONE;

    if ( !self ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }

    rerr |= cpl_table_save( self->table, NULL, NULL, fitsfilename, CPL_IO_DEFAULT);
    if ( rerr ) {
        sph_error_raise( (int)cpl_error_get_code(), __FILE__, __func__, __LINE__,
                        SPH_ERROR_ERROR, "Problem saving table. "
                                "CPL says: %s", cpl_error_get_message() );
        return (int)cpl_error_get_code();
    }

    return rerr;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Save the pixel polyfit table as a FITS cube.
 *
 * @param self The pixel polyfit table to save
 * @param fitsfilename. The filename to save to
 * @return Error code.
 *
 * Saves the table in FITS image format, according to the
 * sph_cube specifications.
 */
/*----------------------------------------------------------------------------*/
int
sph_pixel_polyfit_table_save_cube(sph_pixel_polyfit_table* self,
                                 const char* fitsfilename)
{
    int                 rerr        = CPL_ERROR_NONE;
    sph_master_frame*    mf            = NULL;
    sph_cube*            cube        = NULL;
    int                    ii            = 0;

    if ( !self ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }

    cube = sph_cube_new( fitsfilename );

    if ( !cube ) {
        return SPH_PIXEL_POLYFIT_TABLE_GENERAL;
    }

    for ( ii = 0;  ii <= self->order; ++ ii) {
        mf = sph_pixel_polyfit_table_get_coeff_as_image( self, ii );
        if ( mf ) {
            sph_cube_append_master( cube, mf, ii * 1.0 );
            sph_master_frame_delete( mf );
        }
        else {
            rerr = SPH_PIXEL_POLYFIT_TABLE_GENERAL;
            sph_master_frame_delete( mf );
            break;
        }
    }
    sph_cube_finalise_file(cube->filename);
    sph_cube_delete(cube); cube = NULL;
    return rerr;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Load a pixel polyfit table.
 *
 * @param fitsfilename The filename to load from
 * @return Pointer to the loaded table.
 *
 * Loads a sph_pixel_polyfit_table from a FITS filename. The routine
 * loads the cpl_table and then creates and returns a sph_pixel_polyfit_table
 * structure.
 */
/*----------------------------------------------------------------------------*/
sph_pixel_polyfit_table*
sph_pixel_polyfit_table_load(  const char* fitsfilename ) {
    sph_pixel_polyfit_table*    self   = NULL;
    int                             nx      = 0;
    int                             ny      = 0;
    cpl_table*                      ctab    = NULL;
    int                             order   = 0;

    ctab = cpl_table_load( fitsfilename, 1, 0);

    if ( !ctab ) {
        sph_error_raise(cpl_error_get_code(),
                        __FILE__,
                        __func__,
                        __LINE__,
                        SPH_ERROR_ERROR,"%s",
                        (char*)cpl_error_get_message() );
        return NULL;
    }

    nx = (int)sqrt( cpl_table_get_nrow(ctab) );
    ny = nx;
    order = cpl_table_get_ncol( ctab ) - 4;

    if ( order < 0 ) {
        return NULL;
    }

    self = sph_pixel_polyfit_table_new( nx, ny , order );

    if ( ! self ) {
        return NULL;
    }

    self->table = ctab;

    return self;

}
/*----------------------------------------------------------------------------*/
/**
 * @brief Delete the poly tabe
 *
 * @return Error code.
 *
 * Frees up the space and deletes the polyfit table
 */
/*----------------------------------------------------------------------------*/

sph_error_code
sph_pixel_polyfit_table_delete( sph_pixel_polyfit_table* self ) {
    if ( self == NULL ) {
        SPH_NULL_ERROR
        return CPL_ERROR_NULL_INPUT;
    }

    if ( self->table ) {
        cpl_table_delete( self->table );
    }
    cpl_free(self);
    return CPL_ERROR_NONE;
}

/**@}*/
