/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "sph_polygon.h"

#include "sph_error.h"

#include <cpl.h>

#include <string.h>
#include <math.h>
#include <limits.h>


/**@{*/

/*-----------------------------------------------------------------------------
                        Defines
 -----------------------------------------------------------------------------*/

#ifndef SPH_POLYGON_INISIZE
#define SPH_POLYGON_INISIZE 8
#endif

/*-----------------------------------------------------------------------------
                        Private function prototypes
 -----------------------------------------------------------------------------*/
static
void sph_poly_add_point_( sph_polygon* newpol,
                          const sph_point* p) CPL_ATTR_NONNULL;

static
void sph_line_intersect_point_xy_( sph_point* ppoint,
                                   const sph_line* lineA,
                                   const sph_line* lineB ) CPL_ATTR_NONNULL;

static
int sph_polygon_get_edge_inplace_( const sph_polygon* self,
                                   int edge,
                                   sph_line* inplace ) CPL_ATTR_NONNULL;

static
sph_polygon* sph_polygon_new_(size_t nsize) CPL_ATTR_ALLOC;

static
void sph_polygon_grow_(sph_polygon* self, size_t addsize ) CPL_ATTR_NONNULL;

static
void sph_polygon_copy_(sph_polygon* self, const sph_polygon* other)
    CPL_ATTR_NONNULL;
 

/*-----------------------------------------------------------------------------
                              Function definitions
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 @brief Create a new empty polygon

 @param nothing

 @return the pointer to new polygon.

 Description: create a new empty polygon.

 */
/*----------------------------------------------------------------------------*/
sph_polygon* sph_polygon_new( void ) {

    return sph_polygon_new_( SPH_POLYGON_INISIZE );

}
/*----------------------------------------------------------------------------*/
/**
 @brief Create a new duplicate polygon

 @param self The polygon to copy

 @return the pointer to new polygon.

 Description: create a new duplicate polygon.

 */
/*----------------------------------------------------------------------------*/
sph_polygon* sph_polygon_duplicate( sph_polygon* self ) {
    sph_polygon*        result      = NULL;

    if ( self == NULL ) {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_NULL_INPUT,"Null input poly");
    } else {
        result = sph_polygon_new_( 2 * self->npoints );
        sph_polygon_copy_(result, self);
    }

    return result;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Add a new point to an existing polygon

 @param self        the polygon to add the point to
 @param x            the x-coordinate of point to add
 @param y            the y-coordinate of point to add

 @return error code of the operation.

 Description: Adds a new point to the polygon.

 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_polygon_add_point( sph_polygon* self, double x, double y ) {

    const sph_point p = {x, y};

    sph_poly_add_point_( self, &p );

    return CPL_ERROR_NONE;

}

double sph_polygon_get_top(const sph_polygon* self) {
    double result            = -DBL_MAX;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        if ( self->points[ii]->y > result) result = self->points[ii]->y;
    }
    return result;
}
double sph_polygon_get_bottom(const sph_polygon* self) {
    double result            = DBL_MAX;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        if ( self->points[ii]->y < result) result = self->points[ii]->y;
    }
    return result;
}
double sph_polygon_get_left(const sph_polygon* self) {
    double result            = DBL_MAX;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        if ( self->points[ii]->x < result) result = self->points[ii]->x;
    }
    return result;
}
double sph_polygon_get_right(const sph_polygon* self) {
    double result            = -DBL_MAX;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        if ( self->points[ii]->x > result) result = self->points[ii]->x;
    }
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Return the coordinate of the geometric centre.

 @param self        the polygon
 @param pmid        The centre coordinate, or undefined on error

 @return CPL_ERROR_NONE on succes or the relevant CPL error
 @note If the polygon is empty, the point is unmodified

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_polygon_get_mid_xy(sph_point * pmid,
                                      const sph_polygon* self)
{
    if (self == NULL) {
        return cpl_error_set(cpl_func, CPL_ERROR_NULL_INPUT);
    } else if (pmid == NULL) {
        return cpl_error_set(cpl_func, CPL_ERROR_NULL_INPUT);
    } else if (self->npoints > 0) {
        double x = 0.0;
        double y = 0.0;
        size_t ii;

        for ( ii = 0;  ii < self->npoints; ii++) {
            x += self->points[ii]->x;
            y += self->points[ii]->y;
        }

        pmid->x = x / (double)self->npoints;
        pmid->y = y / (double)self->npoints;
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Return the centre x coordinate

 @param self        the polygon

 @return the x coordinate of the geometric centre

 Description: returns the x coordinate of the geometric centre.

 */
/*----------------------------------------------------------------------------*/
double sph_polygon_get_midx(const sph_polygon* self ) {
    double result            = 0.0;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        result = result + self->points[ii]->x;
    }
    if ( self->npoints > 0 )
        result = result / self->npoints;
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 @brief enlarge the polygon by a factor

 @param self        the polygon
 @param    factor        the factor to enlarge by (lower 1 means shrink)

 @return error code

 Description: enlarge the polygon by a factor. If the factor is less than 1, it
 corresponds to a shrink. The factor must be larger 0.
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_polygon_enlarge( sph_polygon* self, double fac ) {
    double x            = 0.0;
    double y            = 0.0;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }
    if ( fac <= DBL_MIN ) {
        SPH_NULL_ERROR
        return CPL_ERROR_NULL_INPUT;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        x += self->points[ii]->x;
        y += self->points[ii]->y;
    }
    if ( self->npoints > 0 ) {
        x /= self->npoints;
        y /= self->npoints;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        self->points[ii]->x = x + fac * ( self->points[ii]->x - x );
        self->points[ii]->y = y + fac * ( self->points[ii]->y - y );
    }
    return CPL_ERROR_NONE;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Return the centre y coordinate

 @param self        the polygon

 @return the y coordinate of the geometric centre

 Description: returns the y coordinate of the geometric centre.

 */
/*----------------------------------------------------------------------------*/
double sph_polygon_get_midy( const sph_polygon* self ) {
    double result            = 0.0;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        result = result + self->points[ii]->y;
    }
    if ( self->npoints > 0 )
        result = result / self->npoints;
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Return the edge of a polygon as sph_line

 @param self        the polygon the edge belongs to
 @param edge        the id of the edge (see below)

 @return the edge as a new sph_line

 Description: returns the edge of a polygon as a new sph_line
 (the pointer to the sph_line is returned). The edge is identified
 by an integer id, between 1 and the number of edges of the polygon.
 The first edge is the one between points 0 and point 1. The last
 edge is the one between the past point and point 0.

 */
/*----------------------------------------------------------------------------*/
sph_line* sph_polygon_get_edge( const sph_polygon* self, int edge ) {

    sph_line*  result = NULL;

    if (self != NULL && 0 < edge && edge <= self->npoints) {

        result = sph_line_new(0.0, 0.0, 1.0, 1.0);

        (void)sph_polygon_get_edge_inplace_( self, edge, result );
    }

    return result;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Clip a polygon to a target polygon.

 @param self        the polygon to clip
 @param target        the polygon to clip to

 @return error code of the operation

 Description: This clips the polygon self, so that it afterward
 corresponds to an area that is the union between the area of self and
 the target polygon.

 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_polygon_clip_to_target(sph_polygon* self,
                                          const sph_polygon* target,
                                          sph_polygon_workspace* workspace)
{
    sph_polygon*        oldpol      = self;
    sph_polygon*        newpol      = NULL;
    int                 ii          = 0;
    int                 nextp       = 0;
    sph_point*          intersect   = NULL;
    sph_line*           targetedge  = NULL;
    sph_line*           sourceedge  = NULL;
    int                 ee          = 0;
    sph_error_code      rerr        = CPL_ERROR_NONE;
    sph_line*           workline    = NULL;

    if ( !self || !target ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }

    intersect  = workspace->point1;
    newpol     = workspace->poly1;
    workline   = workspace->line1;
    sourceedge = workspace->line2;
    targetedge = workspace->line3;

    for ( ee = 1; ee <= target->npoints; ++ee ) {
        newpol->npoints = 0;

        (void)sph_polygon_get_edge_inplace_( target, ee, targetedge );
        for ( ii = 0; ii < oldpol->npoints; ++ii ) {
            nextp = ii + 1;
            if ( nextp == oldpol->npoints ) nextp = 0;
            if ( sph_polygon_test_same_side( target, ee, oldpol->points[ii],
                                             workline)) {
                if ( sph_polygon_test_same_side(target, ee, oldpol->points[nextp],
                                                workline)) {
                    sph_poly_add_point_( newpol, oldpol->points[nextp]);
                } else {
                    sourceedge->p1->x = oldpol->points[ii]->x;
                    sourceedge->p1->y = oldpol->points[ii]->y;
                    sourceedge->p2->x = oldpol->points[nextp]->x;
                    sourceedge->p2->y = oldpol->points[nextp]->y;

                    sph_line_intersect_point_xy_(intersect, targetedge,
                                                 sourceedge );

                    sph_poly_add_point_( newpol, intersect);
                }
            } else if ( sph_polygon_test_same_side( target, ee,
                                                    oldpol->points[nextp],
                                                    workline)) {
                sourceedge->p1->x = oldpol->points[ii]->x;
                sourceedge->p1->y = oldpol->points[ii]->y;
                sourceedge->p2->x = oldpol->points[nextp]->x;
                sourceedge->p2->y = oldpol->points[nextp]->y;

                sph_line_intersect_point_xy_(intersect, targetedge,
                                             sourceedge );
  
                sph_poly_add_point_( newpol, intersect);
                sph_poly_add_point_( newpol, oldpol->points[nextp]);
            }
        }

        /* FIXME: Change the API to (avoid the copy and) the aliasing */

        /* Copy newpol to self */
        if ( self->nalloc < newpol->npoints ) {
            self->npoints = 0; /* Will overwrite any point so no copy */
            sph_polygon_grow_( self, newpol->npoints );
        }

        sph_polygon_copy_(self, newpol);
    }

    return rerr;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Calculate the overlap are of the polygon with another.

 @param self     The first polygon
 @param target   The second polygon to overlap.
 @param ws       Workspace with temporary storage

 @return area of overlap or negative on error

 Description: The calculates the area of overlap.

 */
/*----------------------------------------------------------------------------*/
double sph_polygon_calculate_overlap(const sph_polygon* self,
                                     const sph_polygon* target,
                                     sph_polygon_workspace* ws )
{
    double result = 0.0;

    /* Make sure poly2 is long ennough. FIXME: Anticipate subsequent growth ? */
    if (ws->poly2->nalloc < (size_t)(self->npoints + target->npoints) ) {
        ws->poly2->npoints = 0; /* Will overwrite any point so no copy */
        sph_polygon_grow_(ws->poly2, (size_t)(self->npoints + target->npoints));
    }

    sph_polygon_copy_(ws->poly2, self);

    /* FIXME: poly2 is aliased */
    if (sph_polygon_clip_to_target( ws->poly2, target, ws )) {
        (void)cpl_error_set_where(cpl_func);
    } else {
        result = sph_polygon_area( ws->poly2 );
    }

    return result;
}
/*----------------------------------------------------------------------------*/
/**
 @brief NOT IMPLEMENTED YET!

 @param

 @return

 Description:

 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_polygon_sort( sph_polygon* self ) {
    if ( !self ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }
    return CPL_ERROR_UNSPECIFIED;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Test if a point is on the same side of edge as all polygon points

 @param self  The polygon the edge belongs to
 @param ee    The id of the edge (see sph_polygon_get_edge)
 @param p     The (x,y)-coord. of point to test.
 @param edge  The edge line. Temporary storage ?

 @return 1 if point is on same side, 0 if not and -1 on error.

 Description: This function tests of a point is on the same side of an edge
 as all polygon points of the polygon self.

 */
/*----------------------------------------------------------------------------*/
short sph_polygon_test_same_side(const sph_polygon* self,
                                 int ee,
                                 const sph_point* p,
                                 sph_line* edge)
{
    const int pt = (ee + 1) % self->npoints;

    sph_polygon_get_edge_inplace_( self, ee, edge );

    return sph_point_ccw_test(*edge->p1, *self->points[pt], *edge->p2)
        == sph_point_ccw_test(*edge->p1, *p,                *edge->p2) ? 1 : 0;

}
/*----------------------------------------------------------------------------*/
/**
 @brief Test if a point is inside a polygon

 @param self         the polygon
 @param x            the x-coordinate of point to test
 @param y            the y-coordinate of point to test

 @return 0 if not inside, 1 if inside, -1 on error.

 Description: This function tests if a point is inside the
 given polygon. If an error occures (e.g. the polygon has 2 or less points)
 -1 is returned.

 */
/*----------------------------------------------------------------------------*/
short sph_polygon_test_inside( sph_polygon* self, double x, double y ) {
    double              minx        = 0.0;
    double              miny        = 0.0;
    sph_line*           edge        = NULL;
    sph_line*           toout       = NULL;
    int                 ii          = 0;
    int                 intersectc  = 0;

    if ( !self ) {
        return -1;
    }
    if ( self->npoints < 3 ) {
        return -1;
    }
    minx         = self->points[ 0 ]->x;
    miny         = self->points[ 0 ]->y;
    for ( ii = 1; ii < self->npoints; ++ii ) {
        if ( self->points[ii]->x < minx ) {
            minx = self->points[ii]->x;
        }
        if ( self->points[ii]->y < miny ) {
            miny = self->points[ii]->y;
        }
    }

    toout = sph_line_new( x , y , minx - 0.1, miny - 0.1 );
    for ( ii = 1; ii <= self->npoints; ++ii ) {
        edge = sph_polygon_get_edge( self, ii );
        if ( sph_line_intersect_test( edge, toout) ) {
            intersectc++;
        }
        sph_line_delete( edge );
    }
    sph_line_delete( toout );
    if ( intersectc % 2 ) {
        return 1;
    }
    return 0;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Shift a polygon.

 @param self        the polygon to shift
 @param dx            the x amount to shift by
 @param dy            the y amount to shift by

 @return error code of the operation

 Description: Shift a polygon.

 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_polygon_shift( sph_polygon* self,
                   double dx, double dy )
{
    int                 ii          = 0;
    sph_error_code      rerr        = CPL_ERROR_NONE;

    if ( !self ) {
        return CPL_ERROR_NULL_INPUT;
    }

    for ( ii = 0; ii < self->npoints; ++ii ) {
        rerr |= sph_point_shift( self->points[ii], dx, dy );
    }
    return rerr;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Rotate a polygon around a point and shift afterwards.

 @param self        the polygon to rotate and shift
 @param cosangle    the cosine of the angle to rotate by
 @param    sinangle    the sine of the angle to rotate by
 @param dx            the x amount to shift by
 @param dy            the y amount to shift by

 @return error code of the operation

 Description: Rotate and afterwards shift a polygon. The rotation is
 specified by the sine and cosine of the rotation angle (measured counter
 clockwise). This input format is for performance reasons.
 The shift is performed after the rotation.

 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_polygon_rotate_around( sph_polygon* self,
                                          double x, double y,
                                          double cosangle,
                                          double sinangle,
                                          double dx, double dy )
{
    int                 ii          = 0;
    sph_error_code      rerr        = CPL_ERROR_NONE;

    if ( !self ) {
        return CPL_ERROR_NULL_INPUT;
    }

    for ( ii = 0; ii < self->npoints; ++ii ) {
        rerr |= sph_point_rotate_around( self->points[ii],
                                 x, y,
                                 cosangle,
                                 sinangle,
                                 dx, dy );
    }
    return rerr;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Return a point(vertex) of the polygon

 @param  self     the polygon
 @param  point  the point (vertext number) to return

 @return pointer to the vertex or NULL if vertex does not exist.

 Description: This returns a pointer to a vertex.
 */
/*----------------------------------------------------------------------------*/
sph_point*
sph_polygon_get_point( sph_polygon* self, int p ) {
    if ( !self ) return NULL;
    if ( p<0 || p>=self->npoints ) return NULL;
    return self->points[p];
}

/*----------------------------------------------------------------------------*/
/**
 @brief Calculate the are of the polygon

 @param  self     the polygon

 @return area of the polygon

 Description: This calculates the are of all clokwise or
 counter clockwise sorted polygons.

 */
/*----------------------------------------------------------------------------*/
double sph_polygon_area( const sph_polygon* self) {

      double  area = 0.0;
      int     ii, jj =0;

      for ( ii = 0; ii < self->npoints; ii++) {
        jj++; if ( jj == self->npoints) jj=0;
        area += (self->points[ii]->x + self->points[jj]->x) *
        ( self->points[ii]->y - self->points[jj]->y );
      }

      return fabs(area)*.5;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Delete the polygon and free space

 @param self        the polygon to delete

 @return nothing

 Description: Does just that.

 */
/*----------------------------------------------------------------------------*/
void sph_polygon_delete( sph_polygon* self ) {

    if ( self != NULL ) {
        cpl_free(self->points_);
        cpl_free(self->points);
        cpl_free(self);
    }
}

 /*----------------------------------------------------------------------------*/
 /**
 @brief Free all space allocated for this module
 */
/*----------------------------------------------------------------------------*/
void sph_polygon_free_all( void ) {
}
/*----------------------------------------------------------------------------*/
/**
 @brief Free the workspace
 */
/*----------------------------------------------------------------------------*/
void sph_polygon_workspace_delete( sph_polygon_workspace* self ) {
    if ( self ) {
        sph_point_delete(self->point1);
        sph_point_delete(self->point2);
        sph_point_delete(self->point3);
        sph_line_delete(self->line1);
        sph_line_delete(self->line2);
        sph_line_delete(self->line3);
        sph_polygon_delete(self->poly1);
        sph_polygon_delete(self->poly2);
        cpl_free(self);
    }
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Allocate a new polygon workspace for use in the clip function
 * @return  pointer to new sph_polygon_workspace or NULL
 */
/*----------------------------------------------------------------------------*/
sph_polygon_workspace*
sph_polygon_workspace_new( void ) {
    sph_polygon_workspace*      self = NULL;

    self = cpl_calloc( 1, sizeof( sph_polygon_workspace ) );
    self->line1 = sph_line_new(0.0,0.0,0.0,0.0);
    self->line2 = sph_line_new(0.0,0.0,0.0,0.0);
    self->line3 = sph_line_new(0.0,0.0,0.0,0.0);
    self->point1 = sph_point_new(0.0,0.0);
    self->point2 = sph_point_new(0.0,0.0);
    self->point3 = sph_point_new(0.0,0.0);
    self->poly1 = sph_polygon_new();
    self->poly2 = sph_polygon_new();
    return self;
}


/**@}*/
/*----------------------------------------------------------------------------*/
//          STATIC FUNCTIONS
/*----------------------------------------------------------------------------*/
static
void
sph_poly_add_point_( sph_polygon* self,
                     const sph_point* p)
{
    if (self->npoints == self->nalloc) {
        sph_polygon_grow_( self, self->nalloc );
    }

    self->points[self->npoints] = self->points_ + self->npoints;
    self->points[self->npoints]->x = p->x;
    self->points[self->npoints]->y = p->y;

    self->npoints++;
}

static
void sph_line_intersect_point_xy_( sph_point* ppoint,
                                   const sph_line* lineA,
                                   const sph_line* lineB )
{
    const double x1 = lineA->p1->x;
    const double x2 = lineA->p2->x;
    const double y1 = lineA->p1->y;
    const double y2 = lineA->p2->y;
    const double x3 = lineB->p1->x;
    const double x4 = lineB->p2->x;
    const double y3 = lineB->p1->y;
    const double y4 = lineB->p2->y;

    const double denom
        = ( ( x1 - x2 ) * ( y3 - y4 ) - ( y1 - y2 ) * ( x3 - x4 ) );

    if (fabs( denom ) < DBL_MIN) {
        ppoint->x = ppoint->y = 0.0;
    } else {
        ppoint->x = (( x1 * y2 - y1 * x2 ) * (x3 - x4) -
                     ( x1 - x2 ) * ( x3 * y4 - y3 * x4 )) / denom;
        ppoint->y = (( x1 * y2 - y1 * x2 ) * (y3 - y4) -
                     ( y1 - y2 ) * ( x3 * y4 - y3 * x4 )) / denom;
    }
}

/*----------------------------------------------------------------------------*/
/**
 @internal
 @brief Get edge ?

 @param self The polygon to process
 @param edge The point to process 
 @param inplace The line to process. Temporary work space ?

 @return Zero on success, non-zero otherwise
 @see  sph_polygon_get_edge()
 @note No error checking in this static function.

 FIXME: If called without a cast to void, then the error checking is missing!

 */
/*----------------------------------------------------------------------------*/

static
int sph_polygon_get_edge_inplace_( const sph_polygon* self,
                                   int edge,
                                   sph_line* inplace )
{
    const int p1 = edge - 1;
    const int p2 = edge == self->npoints ? 0 : edge;

    inplace->p1->x = self->points[ p1 ]->x;
    inplace->p1->y = self->points[ p1 ]->y;
    inplace->p2->x = self->points[ p2 ]->x;
    inplace->p2->y = self->points[ p2 ]->y;

    return 0;
}


/*----------------------------------------------------------------------------*/
/**
 @internal
 @brief Create a new empty polygon with a pre-reserved size

 @param nsize Pre-reserved size
 @return the pointer to new polygon.

 Description: create a new empty polygon.

 */
/*----------------------------------------------------------------------------*/
static sph_polygon* sph_polygon_new_( size_t nsize ) {
    sph_polygon*   self = (sph_polygon*)cpl_malloc( sizeof( *self ) );

    self->nalloc  = nsize; /* Reserve space for this many */
    self->points_ = (sph_point*)cpl_malloc(sizeof(*self->points_) *
                                           self->nalloc );
    /* Need to be able to access (for no good reason) the pointer to each point
       thus increasing the memory usage by 50 % */
    self->points =  (sph_point**)cpl_malloc( sizeof(*self->points) *
                                             self->nalloc );

    self->npoints = 0; /* Initially empty */

    return self;
}

/*----------------------------------------------------------------------------*/
/**
 @internal
 @brief Copy the polygon from one to another, space allowing
 @param self  The polygon to copy from
 @param other The polygon to copy to
 @return void
 @note The copying of points stops when the number of source points have been
   copied, or when the size of the destination has been reached, whichever comes
   first

 */
/*----------------------------------------------------------------------------*/
static
void sph_polygon_copy_(sph_polygon* self, const sph_polygon* other) {

    const size_t copysize = other->npoints <= self->nalloc
        ? other->npoints : self->nalloc;

    if (copysize > 0) {
        size_t i;

        (void)memcpy(self->points_, other->points_, sizeof(*other->points_) *
                     copysize);

        /* If there are more points then extend the list of pointers */
        for (i = (size_t)self->npoints; i < copysize; i++) {
            self->points[i] = self->points_ + i;
        }
    }
    self->npoints = copysize;
}
 

/*----------------------------------------------------------------------------*/
/**
 @internal
 @brief Grow the pre-reserved size
 @param self    The polygon to grow
 @param addsize The size to grow with, a (good) fraction of the current size
 @return void

 */
/*----------------------------------------------------------------------------*/
static
void sph_polygon_grow_( sph_polygon* self, size_t addsize ) {

    void * new;
    void * old = (void*)self->points_;
    size_t i;

    self->nalloc += addsize; /* Increase the size */

    new = cpl_malloc( sizeof(*self->points_) * self->nalloc );

    /* Copy the points in use */
    self->points_ = (sph_point*)memcpy(new, old, sizeof(*self->points_) *
                                       (size_t)self->npoints );
    cpl_free(old);

    /* Need to generate a new list of pointers to the points in use */
    cpl_free(self->points);

    self->points = cpl_malloc( sizeof(*self->points) * self->nalloc );

    for (i = 0; i < (size_t)self->npoints; i++) {
        self->points[i] = self->points_ + i;
    }
}
