/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <cpl.h>
#include "sph_error.h"
#include "sph_keyword_manager.h"
#include "sph_distortion_map.h"
#include "sph_polynomial_distortion_model.h"
#include "sph_double_image.h"
#include "sph_fitting.h"
#include "sph_common_keywords.h"
/*----------------------------------------------------------------------------*/
/**
 * @defgroup A SPHERE API Module
 * @par Synopsis:
 * @code
 * typedef _module_ {
 * } module
 * @endcode
 * @par Desciption:
 *
 * This module provides ...
 */
/*----------------------------------------------------------------------------*/
/**@{*/
sph_polynomial_distortion_model*
sph_polynomial_distortion_model_new( cpl_polynomial* polyx, cpl_polynomial* polyy ) {
    sph_polynomial_distortion_model* model = cpl_calloc(sizeof(sph_polynomial_distortion_model),1);
    if ( polyx ) {
        model->polyx = cpl_polynomial_duplicate( polyx);
    }
    else {
        model->polyx = cpl_polynomial_new(2);
    }
    if ( polyy ) {
        model->polyy = cpl_polynomial_duplicate( polyy);
    }
    else {
        model->polyy = cpl_polynomial_new(2);
    }
    return model;
}

sph_error_code
sph_polynomial_distortion_model_set_coeff_x(
        sph_polynomial_distortion_model* self,
        int order_x,
        int order_y,
        double value)
{
    cpl_size         pows[2];

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    pows[0] = order_x;
    pows[1] = order_y;

    cpl_polynomial_set_coeff(self->polyx,pows,value);

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

sph_point_pattern*
sph_polynomial_distortion_model_apply_pp(
        sph_polynomial_distortion_model* self,
        sph_point_pattern* pp )
{
    cpl_vector* profile = NULL;
    cpl_polynomial* polyx = NULL;
    cpl_polynomial* polyy = NULL;
    cpl_image*      dxim = NULL;
    cpl_image*      dyim = NULL;
    sph_point_pattern*  result = NULL;
    sph_point*      p = NULL;
    int             nn   = 0;
    cpl_vector*     posv = NULL;
    double          dx = 0.0;
    double          dy = 0.0;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    cpl_ensure(pp,CPL_ERROR_NULL_INPUT,NULL);

    polyx = cpl_polynomial_duplicate(self->polyx);
    polyy = cpl_polynomial_duplicate(self->polyy);
    cpl_ensure(polyx,CPL_ERROR_ILLEGAL_OUTPUT,NULL);
    cpl_ensure(polyy,CPL_ERROR_ILLEGAL_OUTPUT,NULL);

    result = sph_point_pattern_new();
    posv = cpl_vector_new(2);
    for (nn = 0; nn < sph_point_pattern_get_size(pp); ++nn) {
        p = sph_point_pattern_get_point(pp,nn);
        cpl_vector_set(posv,0,p->x);
        cpl_vector_set(posv,1,p->y);
        dx = cpl_polynomial_eval(polyx,posv);
        dy = cpl_polynomial_eval(polyy,posv);

        sph_point_pattern_add_point(result,p->x + dx, p->y + dy);

        sph_point_delete(p); p = NULL;
    }
    cpl_vector_delete(posv);
    cpl_polynomial_delete(polyx);
    cpl_polynomial_delete(polyy);
    cpl_vector_delete(profile);
    cpl_image_delete(dxim);
    cpl_image_delete(dyim);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}

sph_error_code
sph_polynomial_distortion_model_set_coeff_y(
        sph_polynomial_distortion_model* self,
        int order_x,
        int order_y,
        double value)
{
    cpl_size         pows[2];

    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);

    pows[0] = order_x;
    pows[1] = order_y;

    cpl_polynomial_set_coeff(self->polyy,pows,value);

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

sph_polynomial_distortion_model*
sph_polynomial_distortion_model_new_from_proplist( const cpl_propertylist* proplist ) {
    sph_polynomial_distortion_model* result = NULL;
    cpl_size pows_x[2] = {0,0};
    cpl_size pows_y[2] = {0,0};
    int xx,yy;
    double coeffx = 0.0;
    double coeffy = 0.0;
    int maxdegree = 0;
    char str[256];

    sprintf( str, "%s %d_%d",SPH_COMMON_KEYWORD_DISTMAP_COEFFX,
            maxdegree,maxdegree);
    while ( cpl_propertylist_has(proplist,str) ) {
        maxdegree++;
        sprintf( str, "%s %d_%d",SPH_COMMON_KEYWORD_DISTMAP_COEFFX,
                maxdegree,maxdegree);
    }
    result = sph_polynomial_distortion_model_new(NULL,NULL);
    for (xx = 0; xx < maxdegree; ++xx) {
        for (yy = 0; yy < maxdegree; ++yy) {
            pows_x[0] = pows_y[0] = xx;
            pows_x[1] = pows_y[1] = yy;
            sprintf( str, "%s %d_%d",
                    SPH_COMMON_KEYWORD_DISTMAP_COEFFX,xx,yy);
            coeffx = cpl_propertylist_get_double(proplist, str);
//            if ( coeffx == 0 ) coeffx = 1.0e-20;
            cpl_polynomial_set_coeff(result->polyx,pows_x,coeffx);
            sprintf( str, "%s %d_%d",
                    SPH_COMMON_KEYWORD_DISTMAP_COEFFY,xx,yy);
            coeffy = cpl_propertylist_get_double(proplist, str);
//            if ( coeffy == 0 ) coeffy = 1.0e-20;
            cpl_polynomial_set_coeff(result->polyy,pows_y,coeffy);
        }
    }
    return result;
}
sph_polynomial_distortion_model*
sph_polynomial_distortion_model_load( const char* filename, int ext ) {
    sph_polynomial_distortion_model* result = NULL;
    cpl_propertylist* proplist = NULL;

    proplist = cpl_propertylist_load(filename,ext);
    result = sph_polynomial_distortion_model_new_from_proplist(proplist);
    cpl_propertylist_delete(proplist);
    return result;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief Create a polynomial model of the distortion from image of grid
 * @param distim        the image containing gridpoints
 * @param pp            the point pattern of ideal points
 * @param thresh        the sigma of threshold for point detection
 * @param polyfit_order the polynomial order of the fit in x and y
 * @param qcpl          the quality control parameterlist (changed on output)
 *
 * @return a new polyonomial distortion model
 *
 *
 */
/*----------------------------------------------------------------------------*/
sph_polynomial_distortion_model*
sph_polynomial_distortion_model_from_grid_image(
        sph_master_frame* mframe,
        sph_point_pattern* pp,
        int polyfit_order,
        double threshold_sigma,
        double threshold_length,
        cpl_propertylist* qclist,
        double * centx,
        double * centy,
        sph_point_pattern** obs_pp)
{
    sph_polynomial_distortion_model*    result = NULL;
    sph_distortion_map*                    dmap = NULL;
    sph_distortion_map*                    dmap_new = NULL;
    cpl_polynomial*                        polyx = NULL;
    cpl_polynomial*                        polyy = NULL;
    cpl_image*                            testim = NULL;
    sph_point_pattern*                      pp2 = NULL;
    int                                    nremoved = 0;
    double                                length_threshold = 100.0;
    double                              midx = 0.0;
    double                              midy = 0.0;
    double                              doffx0 = 0.0;
    double                              doffy0 = 0.0;
    double 								chiX = 0.0;
    double 								chiY = 0.0;
    sph_point*                          point = NULL;
    sph_point*                          optical_axis = NULL;

    cpl_ensure(mframe!=NULL && pp != NULL,
            CPL_ERROR_NULL_INPUT,NULL);

    testim = sph_master_frame_extract_image(mframe,1);

    cpl_ensure(testim != NULL,cpl_error_get_code(),NULL);


    pp2 = sph_point_pattern_new_from_image(
            testim,
            NULL,
            threshold_sigma,
            3.0,3,3);

    if ( obs_pp ) *obs_pp = pp2;

    cpl_image_delete(testim); testim = NULL;

    cpl_ensure(pp2 != NULL,cpl_error_get_code(),NULL);

    if ( centx == NULL || centy == NULL ) {
        optical_axis = sph_point_pattern_get_point(pp2,
                sph_point_pattern_get_central_point(pp2,
                        &midx,&midy) );
    }
    else {
        if ( *centx > 0 && *centy > 0 ) {
            optical_axis = sph_point_new( *centx, *centy);
        }
        else {
            optical_axis = sph_point_pattern_get_point(pp2,
                    sph_point_pattern_get_central_point(pp2,
                            &midx,&midy) );
        }
    }
    SPH_ERROR_RAISE_INFO(
            SPH_ERROR_GENERAL,
            "Central point in observed pattern is at %f, %f."
            "This is taken to correspond to the "
            "centre of the opitcal axis.",
            optical_axis->x, optical_axis->y);

    point = sph_point_pattern_get_point(
            pp,
            sph_point_pattern_get_central_point(pp,
                    &doffx0,
                    &doffy0) );

    SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
            "Central point in input pattern is at %f, %f",
            point->x, point->y);
    sph_point_pattern_add_offset(
            pp,
            optical_axis->x - point->x,
            optical_axis->y - point->y);

    sph_point_pattern_add_offset(pp2,
            -1.0 * optical_axis->x,
            -1.0 * optical_axis->y);

    sph_point_pattern_add_offset(pp,
            -1.0 * optical_axis->x,
            -1.0 * optical_axis->y);


    dmap = sph_distortion_map_new(pp2,pp);

    length_threshold = threshold_length;

    sph_distortion_map_remove_offset(dmap,NULL,NULL);
    nremoved = sph_distortion_map_remove_above_length(dmap,
            length_threshold);

    sph_error_raise(SPH_ERROR_GENERAL,
            __FILE__,__func__,__LINE__,
            SPH_ERROR_INFO,
            "Removed %d vectors above distortion %f "
            "from distortion map from a total of %d "
            "initially.",
            nremoved, length_threshold,
            sph_distortion_map_get_size(dmap) + nremoved);

    if ( qclist )
        cpl_propertylist_update_int(qclist,
            SPH_COMMON_KEYWORD_DISTMAP_NREMOVED,
            nremoved);

    result = sph_polynomial_distortion_model_new( polyx, polyy);
    if ( !result ) {
        sph_distortion_map_delete(dmap);dmap = NULL;
        sph_point_pattern_delete(pp2);pp2=NULL;
        return NULL;
    }
    cpl_image_delete(testim); testim = NULL;
    dmap_new = sph_polynomial_distortion_model_fit_distortion_map(
            result,
            dmap,
            polyfit_order,
            &chiX,
            &chiY);
    sph_polynomial_distortion_set_offset(result,
            (cpl_image_get_size_x(mframe->image) - 1) / 2,
            (cpl_image_get_size_y(mframe->image) - 1) / 2);
    sph_distortion_map_delete(dmap); dmap = NULL;
    sph_distortion_map_delete(dmap_new); dmap_new = NULL;

    sph_point_pattern_add_offset(pp,point->x,point->y);

    if ( centx && centy ) {
        *centx = optical_axis->x;
        *centy = optical_axis->y;
    }
    sph_point_delete(point); point = NULL;
    sph_point_delete(optical_axis); optical_axis = NULL;

    if ( obs_pp == NULL )
        sph_point_pattern_delete(pp2);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}

//{
//    sph_polynomial_distortion_model*       result = NULL;
//    sph_distortion_map*                    dmap = NULL;
//    cpl_polynomial*                        polyx = NULL;
//    cpl_polynomial*                        polyy = NULL;
//    sph_point_pattern*                      pp2 = NULL;
//    cpl_image*                            testim = NULL;
//    int                                    nremoved = 0;
//
//    pp2 = sph_point_pattern_new_from_image(distim,threshold_sigma);
//    if ( !pp2 ) return NULL;
//
//    //Do NOT centralise since this can cause problem if some points are missing in pp2.
//    //sph_point_pattern_centralise(pp2);
//    //sph_point_pattern_centralise(pp);
//
//    testim = sph_point_pattern_create_image(pp2,cpl_image_get_size_x(distim),
//            cpl_image_get_size_y(distim),1.5);
//    cpl_image_save(testim,"testim.fits",CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
//    cpl_image_save(distim,"testmframe.fits",CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
//    dmap = sph_distortion_map_new(pp,pp2);
//
//    if ( !dmap ) {
//        sph_point_pattern_delete(pp2);pp2=NULL;
//        return NULL;
//    }
//
//    sph_distortion_map_remove_offset(dmap,NULL,NULL);
//    nremoved = sph_distortion_map_remove_above_length(dmap,threshold_length);
//    sph_error_raise(SPH_ERROR_GENERAL,
//            __FILE__,__func__,__LINE__,
//            SPH_ERROR_INFO,
//            "Removed %d vectors above distortion %f "
//            "from distortion map from a total of %d "
//            "initially.",
//            nremoved, threshold_length,
//            sph_distortion_map_get_size(dmap)+nremoved);
//
//    result = sph_polynomial_distortion_model_new( polyx, polyy);
//
//    if ( !result ) {
//        sph_distortion_map_delete(dmap);dmap = NULL;
//        sph_point_pattern_delete(pp2);pp2=NULL;
//        return NULL;
//    }
//
//    sph_polynomial_distortion_model_fit_distortion_map(result,dmap,polyfit_order);
//    return result;
//}

sph_error_code
sph_polynomial_distortion_model_save_dfs(
        sph_polynomial_distortion_model* self,
        const char* outfilename,
        cpl_frameset* allframes,
        cpl_frame* template_frame,
        cpl_parameterlist* params,
        const char* tag,
        const char* recipe,
        const char* pipename,
        cpl_propertylist* qclist,
        int nx,
        int ny,
        double pixscale,
        int change_scale)
{
    cpl_propertylist* proplist = NULL;
    sph_double_image* double_image = NULL;
    sph_polynomial_distortion_model*        dmap_duplicate = NULL;
    proplist = sph_polynomial_distortion_model_get_proplist(self);


    if ( change_scale ) {
        self = dmap_duplicate = sph_polynomial_distortion_model_new_from_proplist(
                proplist);
        sph_polynomial_distortion_model_scale_units(
                self,
                1.0/pixscale);
        cpl_propertylist_update_double( proplist,
                SPH_COMMON_KEYWORD_DISTMAP_PIXSCALE, pixscale);
    }
    double_image = sph_double_image_new(nx,ny);
    cpl_image_delete(double_image->iframe->image);
    double_image->iframe->image = NULL;
    cpl_image_delete(double_image->pframe->image);
    double_image->pframe->image = NULL;
    double_image->iframe->image = sph_polynomial_distortion_model_create_image_x(
            self,nx, ny);
    double_image->pframe->image = sph_polynomial_distortion_model_create_image_y(
            self,nx, ny);

    if ( qclist ) {
        if ( double_image->qclist )
            cpl_propertylist_append(double_image->qclist,qclist);
        else
            double_image->qclist = cpl_propertylist_duplicate(qclist);
    }

    sph_double_image_save_dfs(double_image,
    		outfilename,allframes,
    		template_frame,
    		params,
    		tag,
    		recipe,
    		pipename,
    		proplist);

    sph_double_image_delete(double_image);
    cpl_propertylist_delete(proplist);
    sph_polynomial_distortion_model_delete(dmap_duplicate);
    return CPL_ERROR_NONE;
}
sph_error_code sph_polynomial_distortion_model_save(
        sph_polynomial_distortion_model* self,
        const char* filename,
        int nx,
        int ny, unsigned mode,
        cpl_propertylist* pl)
{
    cpl_propertylist* proplist = NULL;
    sph_double_image* double_image = NULL;

    proplist = sph_polynomial_distortion_model_get_proplist(self);
    double_image = sph_double_image_new(nx,ny);
    cpl_image_delete(double_image->iframe->image);
    double_image->iframe->image = NULL;
    cpl_image_delete(double_image->pframe->image);
    double_image->pframe->image = NULL;
    double_image->iframe->image = sph_polynomial_distortion_model_create_image_x(self,nx, ny);
    double_image->pframe->image = sph_polynomial_distortion_model_create_image_y(self,nx, ny);
    if ( pl ) {
        if ( double_image->qclist )
            cpl_propertylist_append(double_image->qclist,pl);
        else
            double_image->qclist = cpl_propertylist_duplicate(pl);
    }
    sph_double_image_save(double_image,filename,proplist,mode);
    sph_double_image_delete(double_image);
    cpl_propertylist_delete(proplist);
    return CPL_ERROR_NONE;
}
sph_point_pattern*
sph_polynomial_distortion_model_create_point_pattern( sph_polynomial_distortion_model* self,
                                                            int nx, int ny, int np)
{
    double dx = (nx * 1.0)/(np+1.0);
    double dy = (ny * 1.0)/(np+1.0);
    double  offsetx = 0.0;
    double  offsety = 0.0;
    double xpos,ypos;
    double distx,disty;
    int xx,yy;
    sph_point_pattern* pp = sph_point_pattern_new();
    cpl_vector* vv = cpl_vector_new(2);

    offsetx = self->offx ;
    offsety = self->offy ;

    for (xx = 0; xx < np; ++xx) {
        for (yy = 0; yy < np; ++yy) {
            // adding 0.5 below to go to FITS (CPL) pixel coords.
            // (note: offsetx should also be in CPL coords.)
            xpos = dx * ( xx + 1.0 ) + 0.5 - offsetx;

            ypos = dy * ( yy + 1.0 ) + 0.5 - offsety;
            cpl_vector_set(vv,0,xpos);
            cpl_vector_set(vv,1,ypos);
            if ( self->polyx != NULL )
                distx = cpl_polynomial_eval( self->polyx, vv);
            else
                distx = 0.0;
            if (self->polyy != NULL)
                disty = cpl_polynomial_eval( self->polyy, vv);
            else
                disty = 0.0;
            xpos = xpos + distx;
            ypos = ypos + disty;
            sph_point_pattern_add_point(pp,xpos, ypos);
        }
    }

    cpl_vector_delete(vv);
    return pp;
}
cpl_polynomial*
sph_polynomial_distortion_model_get_poly_x( sph_polynomial_distortion_model* self ) {
    return self->polyx;
}

cpl_polynomial*
sph_polynomial_distortion_model_get_poly_y( sph_polynomial_distortion_model* self ) {
    return self->polyy;
}

cpl_propertylist*
sph_polynomial_distortion_model_get_proplist( sph_polynomial_distortion_model* self ) {
    cpl_propertylist* proplist = cpl_propertylist_new();
     cpl_size pows_x[2] = {0,0};
     cpl_size pows_y[2] = {0,0};
     int xx,yy;
     int maxdegree = 0;
     int maxdegree_new = 0;
     double deg = 0.0;

     maxdegree = cpl_polynomial_get_degree(self->polyx);
     deg = cpl_polynomial_get_degree(self->polyy);
     if ( deg > maxdegree ) maxdegree = deg;
     maxdegree++;
     for (xx = 0; xx < maxdegree; ++xx) {
         for (yy = 0; yy < maxdegree; ++yy) {
             pows_x[0] = pows_y[0] = xx;
             pows_x[1] = pows_y[1] = yy;
             sph_keyword_manager_update_double2(
                     proplist,
                     SPH_COMMON_KEYWORD_DISTMAP_COEFFX,
                     xx,yy,
                     cpl_polynomial_get_coeff(self->polyx, pows_x));
             sph_keyword_manager_update_double2(
                     proplist,
                     SPH_COMMON_KEYWORD_DISTMAP_COEFFY,
                     xx,yy,
                     cpl_polynomial_get_coeff(self->polyy, pows_y));
         }
     }
     return proplist;
}
double
sph_polynomial_distortion_model_get_dist_x( sph_polynomial_distortion_model* self,
        double xpos, double ypos) {
    double distx;
    cpl_vector* vv = cpl_vector_new(2);
    cpl_vector_set(vv,0,xpos);
    cpl_vector_set(vv,1,ypos);
    if (self->polyx != NULL)
        distx = cpl_polynomial_eval( self->polyx, vv);
    else
        distx = 0.0;
    cpl_vector_delete(vv); vv = NULL;
    return distx;
}
double
sph_polynomial_distortion_model_get_dist_y( sph_polynomial_distortion_model* self,
        double xpos, double ypos) {
    double distx;
    cpl_vector* vv = cpl_vector_new(2);
    cpl_vector_set(vv,0,xpos);
    cpl_vector_set(vv,1,ypos);
    if (self->polyy != NULL)
        distx = cpl_polynomial_eval( self->polyy, vv);
    else
        distx = 0.0;

    cpl_vector_delete(vv); vv = NULL;

    return distx;
}

cpl_image*
sph_polynomial_distortion_model_create_image_x( sph_polynomial_distortion_model* self,
        int nx, int ny) {
    int xx,yy;
    double xpos,ypos,distx;
    cpl_vector* vv = cpl_vector_new(2);
    cpl_image* result = cpl_image_new(nx,ny,CPL_TYPE_DOUBLE);
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            xpos = xx  - self->offx;
            ypos = yy  - self->offy;
            cpl_vector_set(vv,0,xpos);
            cpl_vector_set(vv,1,ypos);
            distx = sph_polynomial_distortion_model_get_dist_x(self,xpos,ypos);
            cpl_image_set(result,xx+1,yy+1,distx);
        }
    }

    cpl_vector_delete(vv);
    return result;
}
sph_error_code
sph_polynomial_distortion_set_offset(
        sph_polynomial_distortion_model* self,
        double offx,
        double offy) {
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    self->offx = offx;
    self->offy = offy;
    return CPL_ERROR_NONE;
}
cpl_image*
sph_polynomial_distortion_model_create_image_y( sph_polynomial_distortion_model* self,
        int nx, int ny) {
    int xx,yy;
    double xpos,ypos,distx;
    cpl_vector* vv = cpl_vector_new(2);
    cpl_image* result = cpl_image_new(nx,ny,CPL_TYPE_DOUBLE);
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            xpos = xx  - self->offx;
            ypos = yy  - self->offy;
            distx = sph_polynomial_distortion_model_get_dist_y(self,xpos,ypos);
            cpl_image_set(result,xx+1,yy+1,distx);
        }
    }

    cpl_vector_delete(vv);
    return result;
}

sph_distortion_map*
sph_polynomial_distortion_model_fit_distortion_map(
        sph_polynomial_distortion_model* self,
        sph_distortion_map* dmap,
        int maxdegree,
        double* chiX,
        double* chiY) {
    cpl_vector* xvals = cpl_vector_new(sph_distortion_map_get_size(dmap));
    cpl_vector* yvals = cpl_vector_new(sph_distortion_map_get_size(dmap));
    cpl_vector* zvalsY = cpl_vector_new(sph_distortion_map_get_size(dmap));
    cpl_vector* zvalsX = cpl_vector_new(sph_distortion_map_get_size(dmap));
    int cc = 0;
    sph_distortion_vector* dv = NULL;
    sph_distortion_vector* di = NULL;
    sph_distortion_map* newdistmap = NULL;
    cpl_vector* vv = NULL;
    double diffx;
    double diffy;

    for (cc = 0; cc < sph_distortion_map_get_size(dmap); ++cc) {
        dv = sph_distortion_map_get( dmap, cc );
        cpl_vector_set(xvals,cc,dv->x);
        cpl_vector_set(yvals,cc,dv->y);
        cpl_vector_set(zvalsX,cc,dv->dx);
        cpl_vector_set(zvalsY,cc,dv->dy);
    }
    if ( self->polyx != NULL ) {
        cpl_polynomial_delete(self->polyx);
        self->polyx = NULL;
    }
    if ( self->polyy != NULL ) {
        cpl_polynomial_delete(self->polyy);
        self->polyy = NULL;
    }

    self->polyy = sph_fitting_fit_poly2d(xvals, yvals, zvalsY, chiX, maxdegree,maxdegree);
    self->polyx = sph_fitting_fit_poly2d(xvals, yvals, zvalsX, chiY, maxdegree,maxdegree);

    newdistmap = sph_distortion_map_new(NULL,NULL);
    for (cc = 0; cc < sph_distortion_map_get_size(dmap); ++cc) {
        vv = cpl_vector_new(2);
        dv = sph_distortion_map_get( dmap, cc );
        cpl_vector_set(vv,0,dv->x);
        cpl_vector_set(vv,1,dv->y);
        diffx = cpl_polynomial_eval(self->polyx,vv) - dv->dx;
        diffy = cpl_polynomial_eval(self->polyy,vv) - dv->dy;
        cpl_vector_delete(vv);
        di = sph_distortion_vector_new(dv->x, dv->y, diffx, diffy);
        sph_distortion_map_insert_vec(newdistmap,di);
    }

    cpl_vector_delete(xvals);xvals = NULL;
    cpl_vector_delete(yvals);yvals = NULL;
    cpl_vector_delete(zvalsX);zvalsX = NULL;
    cpl_vector_delete(zvalsY);zvalsY = NULL;
    return newdistmap;
}

cpl_image*
sph_polynomial_distortion_model_apply(
        sph_polynomial_distortion_model* self,
        cpl_image* image,
        cpl_kernel kerneltype,
        double radin )
{
    cpl_image* result = NULL;
    cpl_vector* profile = NULL;
    double      rad = 10.0;
    double      coeff = 0.0;
    cpl_size pows[2] = {0,0};
    cpl_polynomial* polyx = NULL;
    cpl_polynomial* polyy = NULL;
    int          nx = 0;
    int          ny = 0;
    cpl_image*      dxim = NULL;
    cpl_image*      dyim = NULL;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    cpl_ensure(radin < 2048,CPL_ERROR_ILLEGAL_INPUT,NULL);

    if ( radin > 0.0 ) rad = radin;

    profile = cpl_vector_new( CPL_KERNEL_DEF_SAMPLES * (int)radin);
    cpl_ensure(profile,CPL_ERROR_NULL_INPUT,NULL);


    polyx = cpl_polynomial_duplicate(self->polyx);
    polyy = cpl_polynomial_duplicate(self->polyy);
    cpl_ensure(polyx,CPL_ERROR_ILLEGAL_OUTPUT,NULL);
    cpl_ensure(polyy,CPL_ERROR_ILLEGAL_OUTPUT,NULL);

    nx = cpl_image_get_size_x(image);
    ny = cpl_image_get_size_y(image);

    dxim = sph_polynomial_distortion_model_create_image_x(self,nx,ny);
    dyim = sph_polynomial_distortion_model_create_image_y(self,nx,ny);

    pows[0] = 0; pows[1] = 0;
    coeff = cpl_polynomial_get_coeff(polyx,pows);
    cpl_polynomial_set_coeff(polyx,pows,-coeff);
    pows[0] = 1; pows[1] = 0;
    coeff = cpl_polynomial_get_coeff(polyx,pows);
    cpl_polynomial_set_coeff(polyx,pows,-coeff);
    pows[0] = 0; pows[1] = 1;
    coeff = cpl_polynomial_get_coeff(polyx,pows);
    cpl_polynomial_set_coeff(polyx,pows,-coeff);
    pows[0] = 1; pows[1] = 1;
    coeff = cpl_polynomial_get_coeff(polyx,pows);
    cpl_polynomial_set_coeff(polyx,pows,-coeff);
    pows[0] = 0; pows[1] = 0;
    coeff = cpl_polynomial_get_coeff(polyy,pows);
    cpl_polynomial_set_coeff(polyy,pows,-coeff);
    pows[0] = 1; pows[1] = 0;
    coeff = cpl_polynomial_get_coeff(polyy,pows);
    cpl_polynomial_set_coeff(polyy,pows,-coeff);
    pows[0] = 0; pows[1] = 1;
    coeff = cpl_polynomial_get_coeff(polyy,pows);
    cpl_polynomial_set_coeff(polyy,pows,-coeff);
    pows[0] = 1; pows[1] = 1;
    coeff = cpl_polynomial_get_coeff(polyy,pows);
    cpl_polynomial_set_coeff(polyy,pows,-coeff);

    //cpl_polynomial_shift_1d(polyx,0,-self->offx);
    //cpl_polynomial_shift_1d(polyx,1,-self->offy);
    //cpl_polynomial_shift_1d(polyy,0,-self->offx);
    //cpl_polynomial_shift_1d(polyy,1,-self->offy);

    pows[0] = 1; pows[1] = 0;
    coeff = cpl_polynomial_get_coeff(polyx,pows);
    cpl_polynomial_set_coeff(polyx,pows,coeff + 1.0000);
    pows[0] = 0; pows[1] = 1;
    coeff = cpl_polynomial_get_coeff(polyy,pows);
    cpl_polynomial_set_coeff(polyy,pows,coeff + 1.0000);

    cpl_vector_fill_kernel_profile(profile,kerneltype,radin);

    result = cpl_image_new(cpl_image_get_size_x(image),
            cpl_image_get_size_y(image), cpl_image_get_type(image) );


    //cpl_image_warp_polynomial( result, image, polyx, polyy, profile,rad,profile,rad );
    cpl_image_warp( result, image, dxim, dyim, profile, rad, profile,rad );
    cpl_polynomial_delete(polyx);
    cpl_polynomial_delete(polyy);
    cpl_vector_delete(profile);
    cpl_image_delete(dxim);
    cpl_image_delete(dyim);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}

sph_error_code
sph_polynomial_distortion_model_apply_poly(
        sph_polynomial_distortion_model* self,
        sph_polygon* poly )
{
    cpl_polynomial* polyx = NULL;
    cpl_polynomial* polyy = NULL;
    double          dx = 0.0;
    double          dy = 0.0;
    cpl_vector*     vpos = NULL;
    int             np = 0;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    polyx = cpl_polynomial_duplicate(self->polyx);
    polyy = cpl_polynomial_duplicate(self->polyy);
    cpl_ensure_code(polyx,CPL_ERROR_ILLEGAL_OUTPUT);
    cpl_ensure_code(polyy,CPL_ERROR_ILLEGAL_OUTPUT);

    for (np = 0; np < poly->npoints; ++np) {
        vpos = cpl_vector_new(2);
        cpl_vector_set(vpos,0,poly->points[np]->x);
        cpl_vector_set(vpos,1,poly->points[np]->y);
        dx = cpl_polynomial_eval(polyx,vpos);
        dy = cpl_polynomial_eval(polyy,vpos);
        poly->points[np]->x += dx;
        poly->points[np]->y += dy;
        cpl_vector_delete(vpos); vpos = NULL;
    }

    cpl_polynomial_delete(polyx);
    cpl_polynomial_delete(polyy);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Scale the units of the distortion map
 * @param self
 * @param scale  the factor to scale the units by
 *
 * @return error code
 *
 * This will scale the UNITS of the distortion map by the given
 * factor. So for example if the units of the distortion were in pixels
 * and a factor of 0.5 is applied, the distortion is then in units of 0.5 pixels.
 *
 */
/*----------------------------------------------------------------------------*/

sph_error_code
sph_polynomial_distortion_model_scale_units(
        sph_polynomial_distortion_model* self,
        double scale)
{
    cpl_size           pows[2] = {0,0};
    int                power_x  = 0;
    int                power_y  = 0;
    double             coeff     = 0;
    double              factor   = 0.0;

    cpl_ensure_code( self, CPL_ERROR_NULL_INPUT );
    cpl_ensure_code( scale > 0.0, CPL_ERROR_ILLEGAL_INPUT );

    for (power_x = 0;
            power_x <= cpl_polynomial_get_degree(self->polyx);
            ++power_x)
    {
        for (power_y = 0;
                power_y <= cpl_polynomial_get_degree(self->polyx);
                ++power_y) {
            pows[0] = power_x; pows[1] = power_y;
            // This is due to scaling of axis on unit change
            factor = pow(scale,power_x + power_y);
            coeff = cpl_polynomial_get_coeff( self->polyx, pows);
            // This is due to scaling of value axis on unit change
            coeff = coeff * scale;
            cpl_polynomial_set_coeff( self->polyx, pows,
                    coeff / factor);
            coeff = cpl_polynomial_get_coeff( self->polyy, pows);
            // This is due to scaling of value axis on unit change
            coeff = coeff * scale;
            cpl_polynomial_set_coeff( self->polyy, pows,
                    coeff / factor);

        }
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

void
sph_polynomial_distortion_model_delete(sph_polynomial_distortion_model* self) {
    if ( !self ) return;
    if ( self->polyx) cpl_polynomial_delete( self->polyx );
    if ( self->polyy) cpl_polynomial_delete( self->polyy );
    cpl_free(self);
}


/**@}*/
