/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifndef SPH_NO_USE_HDRL
/* Use HDRL by default, but provide an option to turn it off */
#define HDRL_USE_EXPERIMENTAL
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "sph_strehl.h"
#include "irplib_strehl.h"
#include "irplib_plugin.h"

#ifdef HDRL_USE_EXPERIMENTAL
#include "hdrl.h"
#endif

#include <assert.h>
#include <string.h>

/*-----------------------------------------------------------------------------
                                   Define
 -----------------------------------------------------------------------------*/

#define SPH_STREHL_QC_STREHL          "ESO QC STREHL%c"
#define SPH_STREHL_QC_STREHL_ERR      "ESO QC STREHL%c ERROR"
#define SPH_STREHL_QC_STREHL_POSX     "ESO QC STREHL%c POSX"
#define SPH_STREHL_QC_STREHL_POSY     "ESO QC STREHL%c POSY"
#define SPH_STREHL_QC_STREHL_SIGMA    "ESO QC STREHL%c SIGMA"
#define SPH_STREHL_QC_STREHL_FLUX     "ESO QC STREHL%c FLUX"
#define SPH_STREHL_QC_STREHL_PEAK     "ESO QC STREHL%c PEAK"
#define SPH_STREHL_QC_STREHL_BKG      "ESO QC STREHL%c BACKGROUND"
#define SPH_STREHL_QC_STREHL_BKGNOISE "ESO QC STREHL%c BACKGROUND NOISE"

#ifndef SPH_STREHL_COLUMN_KEY
#define SPH_STREHL_COLUMN_KEY "filter_key"
#endif

#ifndef SPH_STREHL_COLUMN_ID
#define SPH_STREHL_COLUMN_ID "filter_id"
#endif

#ifndef SPH_STREHL_COLUMN_LAMBDA_LEFT
#define SPH_STREHL_COLUMN_LAMBDA_LEFT "lambda_left"
#endif

#ifndef SPH_STREHL_COLUMN_LAMBDA_RIGHT
#define SPH_STREHL_COLUMN_LAMBDA_RIGHT "lambda_right"
#endif

#ifndef SPH_STREHL_COLUMN_WIDTH_LEFT
#define SPH_STREHL_COLUMN_WIDTH_LEFT "width_left"
#endif

#ifndef SPH_STREHL_COLUMN_WIDTH_RIGHT
#define SPH_STREHL_COLUMN_WIDTH_RIGHT "width_right"
#endif

#ifndef SPH_STREHL_COLUMN_LAMBDA_CAM1
#define SPH_STREHL_COLUMN_LAMBDA_CAM1 "lambda_cam1"
#endif

#ifndef SPH_STREHL_COLUMN_LAMBDA_CAM2
#define SPH_STREHL_COLUMN_LAMBDA_CAM2 "lambda_cam2"
#endif

#ifndef SPH_STREHL_COLUMN_WIDTH_CAM1
#define SPH_STREHL_COLUMN_WIDTH_CAM1 "width_cam1"
#endif

#ifndef SPH_STREHL_COLUMN_WIDTH_CAM2
#define SPH_STREHL_COLUMN_WIDTH_CAM2 "width_cam2"
#endif

#ifndef SPH_STREHL_BORDER
#define SPH_STREHL_BORDER 256
#endif


/* Modified from fitsio.h */
#define SPH_FLEN_KEYWORD 72  /* max length of a keyword (HIERARCH convention) */

/*-----------------------------------------------------------------------------
                                   Data Types
 -----------------------------------------------------------------------------*/

typedef enum {
    SPH_ARM_UNKNOWN,
    SPH_ARM_ZIMPOL1,
    SPH_ARM_ZIMPOL2,
    SPH_ARM_IRDIS
} sph_strehl_arm;


/*-----------------------------------------------------------------------------
                                   Function declarations
 -----------------------------------------------------------------------------*/

static cpl_error_code sph_strehl_compute(cpl_propertylist *,
                                         char,
                                         sph_strehl_arm,
                                         const cpl_image  *,
                                         const cpl_propertylist *,
                                         const char       *,
                                         const char       *,
                                         const cpl_frame  *,
                                         const char       *,
                                         const cpl_parameterlist *);

static cpl_error_code sph_strehl_compute_two(cpl_propertylist        *,
                                             const cpl_image         *,
                                             const cpl_image         *,
                                             sph_strehl_arm           ,
                                             sph_strehl_arm           ,
                                             const char              *,
                                             const cpl_frame         *,
                                             const cpl_frame         *,
                                             const char              *,
                                             const char              *,
                                             const char              *,
                                             const char              *,
                                             const char              *,
                                             const cpl_parameterlist *);


/*-----------------------------------------------------------------------------
                                   Function definitions
 -----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
 @brief Compute the Strehl ratios of the 2 brightest objects in the IRDIS image
 @param self        Propertylist to append Strehl numbers to
 @param image_left  Left image to process
 @param image_right Right image to process
 @param rawframe    The input raw frame
 @param filterframe An input calibration frame with the filter wavelengths
 @param recipe      Recipe name (for parameter access)
 @param parlist     Recipe parameters
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_strehl_irdis(cpl_propertylist        * self,
                                const cpl_image         * image_left,
                                const cpl_image         * image_right,
                                const cpl_frame         * rawframe,
                                const cpl_frame         * filterframe,
                                const char              * recipe,
                                const cpl_parameterlist * parlist) {

    const cpl_error_code code =
        sph_strehl_compute_two(self, image_left, image_right,
                               SPH_ARM_IRDIS, SPH_ARM_IRDIS,
                               SPH_STREHL_PSCALE      "|"
                               SPH_STREHL_FILTER_IRD1 "|"
                               SPH_STREHL_FILTER_IRD2,
                               rawframe, filterframe,
                               SPH_STREHL_COLUMN_LAMBDA_LEFT,
                               SPH_STREHL_COLUMN_WIDTH_LEFT,
                               SPH_STREHL_COLUMN_LAMBDA_RIGHT,
                               SPH_STREHL_COLUMN_WIDTH_RIGHT,
                               recipe, parlist);

    return code ? cpl_error_set_where(cpl_func) : CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Compute the Strehl ratio of the brightest object in the 2 ZIMPOL images
 @param self        Propertylist to append Strehl numbers to
 @param image_cam1  Image to process from camera 1
 @param image_cam2  Image to process from camera 2
 @param rawframe    The input raw frame
 @param filterframe An input calibration frame with the filter wavelengths
 @param recipe      Recipe name (for parameter access)
 @param parlist     Recipe parameters
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_strehl_zimpol(cpl_propertylist        * self,
                                 const cpl_image         * image_cam1,
                                 const cpl_image         * image_cam2,
                                 const cpl_frame         * rawframe,
                                 const cpl_frame         * filterframe,
                                 const char              * recipe,
                                 const cpl_parameterlist * parlist) {

    const cpl_error_code code =
        sph_strehl_compute_two(self, image_cam1, image_cam2,
                               SPH_ARM_ZIMPOL1, SPH_ARM_ZIMPOL2,
                               SPH_STREHL_PSCALE      "|"
                               SPH_STREHL_FILTER_ZPL0 "|"
                               SPH_STREHL_FILTER_ZPL1 "|"
                               SPH_STREHL_FILTER_ZPL2,
                               rawframe, filterframe,
                               SPH_STREHL_COLUMN_LAMBDA_CAM1,
                               SPH_STREHL_COLUMN_WIDTH_CAM1,
                               SPH_STREHL_COLUMN_LAMBDA_CAM2,
                               SPH_STREHL_COLUMN_WIDTH_CAM2,
                               recipe, parlist);

    return code ? cpl_error_set_where(cpl_func) : CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
 @internal
 @brief Compute the Strehl ratio of the brightest object in the 2 ZIMPOL images
 @param self          Propertylist to append Strehl numbers to
 @param image1        First image to process
 @param image2        Second image to process
 @param arm1          The SPHERE arm (IRDIS or ZIMPOL1/2)
 @param arm2          The SPHERE arm (IRDIS or ZIMPOL1/2)
 @param loadregexp    The FITS cards to load
 @param rawframe      The input raw frame
 @param filterframe   An input calibration frame with the filter wavelengths
 @param label1_lambda The column label of the filter lambda for image 1
 @param label1_width  The column label of the filter width for image 1
 @param label2_lambda The column label of the filter lambda for image 2
 @param label2_width  The column label of the filter width for image 2
 @param recipe        Recipe name (for parameter access)
 @param parlist       Recipe parameters
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code sph_strehl_compute_two(cpl_propertylist        * self,
                                      const cpl_image         * image1,
                                      const cpl_image         * image2,
                                      sph_strehl_arm            arm1,
                                      sph_strehl_arm            arm2,
                                      const char              * loadregexp,
                                      const cpl_frame         * rawframe,
                                      const cpl_frame         * filterframe,
                                      const char              * label1_lambda,
                                      const char              * label1_width,
                                      const char              * label2_lambda,
                                      const char              * label2_width,
                                      const char              * recipe,
                                      const cpl_parameterlist * parlist) {


    cpl_errorstate     okstate  = cpl_errorstate_get();
    const char       * filename = cpl_frame_get_filename(rawframe);
    cpl_propertylist * plist;

    cpl_ensure_code(self          != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(image1        != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(image2        != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(loadregexp    != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(parlist       != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(filename      != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(label1_lambda != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(label1_width  != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(label2_lambda != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(label2_width  != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(recipe        != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(parlist       != NULL, CPL_ERROR_NULL_INPUT);
    if (filterframe == NULL) {
        /* If the SOF does not contain this optional frame,
           the filterframe is NULL */
        return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                     "Missing filter frame with tag "
                                     IRPLIB_STRINGIFY
                                     (SPH_IRD_TAG_FILTER_TABLE_CALIB));
    }

    plist = cpl_propertylist_load_regexp(filename, 0, loadregexp, CPL_FALSE);

    if (!cpl_errorstate_is_equal(okstate)) return cpl_error_set_where(cpl_func);

    assert( plist != NULL);

    if (sph_strehl_compute(self, '1', arm1, image1, plist,
                           label1_lambda, label1_width,
                           filterframe, recipe, parlist)) {
        cpl_propertylist_delete(plist);
        return cpl_error_set_where(cpl_func);
    }

    if (sph_strehl_compute(self, '2', arm2, image2, plist,
                           label1_lambda, label1_width,
                           filterframe, recipe, parlist)) {
        cpl_propertylist_delete(plist);
        return cpl_error_set_where(cpl_func);
    }

    cpl_propertylist_delete(plist);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Find (approximately) the star to analyze
 @param self         The image to process
 @param ppos_x       The X-position on success otherwise undefined
 @param ppos_y       The Y-position on success otherwise undefined
 @param psigma       The detection level
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

*/
/*----------------------------------------------------------------------------*/
static cpl_error_code sph_strehl_find_star(const cpl_image * self,
                                           double          * ppos_x,
                                           double          * ppos_y,
                                           double          * psigma)
{

    const cpl_size nx = cpl_image_get_size_x(self);
    const cpl_size ny = cpl_image_get_size_y(self);

    /* Sigma-levels for detection of a bright star. */
    double          psigmas[] = {5.0, 2.0, 1.0, 0.5};
    const int       nsigmas = (int)(sizeof(psigmas)/sizeof(double));
    cpl_size        isigma;
    cpl_vector    * sigmas = NULL;
    cpl_apertures * apert = NULL;
    int             iflux;

    cpl_ensure_code(self   != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(ppos_x != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(ppos_y != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(psigma != NULL, CPL_ERROR_NULL_INPUT);

    /* Detect a bright star around the center */
    cpl_msg_info(cpl_func, "---> Detecting a bright star using %d sigma-levels "
                 "ranging from %g down to %g", nsigmas, psigmas[0],
                 psigmas[nsigmas-1]);
    sigmas = cpl_vector_wrap(nsigmas, psigmas);
    apert = cpl_apertures_extract_window(self, sigmas,
                                         1  + SPH_STREHL_BORDER,
                                         1  + SPH_STREHL_BORDER,
                                         nx - SPH_STREHL_BORDER,
                                         ny - SPH_STREHL_BORDER,
                                         &isigma);
    cpl_vector_unwrap(sigmas);

    if (apert == NULL) {
        return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                     "No star in %u X %u with "
                                     IRPLIB_STRINGIFY(SPH_STREHL_BORDER)
                                     " border at sigma=%g", (unsigned)nx,
                                     (unsigned)ny, psigmas[nsigmas-1]);
    }

    if (irplib_apertures_find_max_flux(apert, &iflux, 1)) {
        cpl_apertures_delete(apert);
        return cpl_error_set_where(cpl_func);
    }

    *ppos_x = cpl_apertures_get_centroid_x(apert, iflux);
    *ppos_y = cpl_apertures_get_centroid_y(apert, iflux);
    cpl_apertures_delete(apert);

    *psigma = psigmas[isigma];

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Read filter values from filter frame
 @param self         Filter frame to read from
 @param label_lambda The column label of the filter lambda
 @param label_width  The column label of the filter width
 @param key          The key name
 @param filter       The filter name
 @param plam         Central wavelength [micron]
 @param pdlam        Filter bandwidth [micron]
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

*/
/*----------------------------------------------------------------------------*/
static cpl_error_code sph_strehl_read_filter(const cpl_frame * self,
                                             const char      * label_lambda,
                                             const char      * label_width,
                                             const char      * key,
                                             const char      * filter,
                                             double          * plam,
                                             double          * pdlam)
{

    cpl_errorstate okstate  = cpl_errorstate_get();
    const char   * filename = cpl_frame_get_filename(self);
    cpl_table    * table    = cpl_table_load(filename, 1, 0);
    const cpl_size nrow     = cpl_table_get_nrow(table);
    cpl_size       nsel;


    cpl_ensure_code(self         != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(label_lambda != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(label_width  != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(key          != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(filter       != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(plam         != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pdlam        != NULL, CPL_ERROR_NULL_INPUT);

    if (table == NULL) return cpl_error_set_where(cpl_func);

    /* Select rows not matching the filter key (initially all selected) */
    nsel = cpl_table_and_selected_string(table, SPH_STREHL_COLUMN_KEY,
                                         CPL_NOT_EQUAL_TO, key);
    if (nsel < 0) return cpl_error_set_where(cpl_func);

    /* Select rows not matching the filter id */
    nsel = cpl_table_or_selected_string(table, SPH_STREHL_COLUMN_ID,
                                         CPL_NOT_EQUAL_TO, filter);
    if (nsel < 0) return cpl_error_set_where(cpl_func);

    nsel = nrow - nsel; /* Matching rows */

    if (nsel < 1) return
        cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND, "file=%s. "
                              "key=%s. filter=%s. rows=%u", filename, key,
                              filter, (unsigned)nrow);
    if (nsel > 1) return
        cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "file=%s. key="
                              "%s. filter=%s: %u of %u rows match", filename,
                              key, filter, (unsigned)nsel, (unsigned)nrow);

    if (cpl_table_erase_selected(table)) return cpl_error_set_where(cpl_func);

    *plam  = cpl_table_get_double(table, label_lambda, 0, NULL);
    *pdlam = cpl_table_get_double(table, label_width,  0, NULL);

    cpl_table_delete(table);

    if (!cpl_errorstate_is_equal(okstate)) return cpl_error_set_where(cpl_func);

    if (*plam <= 0.0) return
        cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "file=%s. key="
                              "%s. filter=%s. %u row(s). lambda=%g <= 0. width="
                              "%g", filename, key, filter, (unsigned)nrow,
                              *plam, *pdlam);
    if (*pdlam <= 0.0) return
        cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "file=%s. key="
                              "%s. filter=%s. %u row(s). lambda=%g. width=%g <="
                              " 0", filename, key, filter, (unsigned)nrow,
                              *plam, *pdlam);

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
 @brief Read header values from raw frame
 @param self    Propertylist of raw input frame
 @param arm     The SPHERE arm (IRDIS or ZIMPOL)
 @param ppscale The pixel scale
 @param pkey    The filter key name
 @param pfilter The filter name
 @return   CPL_ERROR_NONE or the relevant CPL error code on error
 @note The filter name is only valid while the propertylist exists

 For IRDIS the light passes through first a broad band filter, with the
 actual filter indicated by 'ESO INS1 FILT NAME'-card. Next, it passes
 though a narrow band filter, with the actual filter indicated by
 'ESO INS1 OPTI2 NAME'-card. The card may have the value 'CLEAR', indicating
 that the second aperture is clear, with no filter. In this case the actual
 filter is the broad band filter.

 For ZIMPOL there is also a shared filter and for each of the two cameras a
 camera specific filter. For ZIMPOL we first check the camera specific filter
 and if it is OPEN, we use the characteristics of the shared filter.

*/
/*----------------------------------------------------------------------------*/
static cpl_error_code sph_strehl_read_header(const cpl_propertylist * self,
                                             sph_strehl_arm           arm,
                                             double                 * ppscale,
                                             const char            ** pkey,
                                             const char            ** pfilter)
{
    cpl_errorstate okstate = cpl_errorstate_get();
    const char * strval;

    cpl_ensure_code(self    != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(ppscale != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pkey    != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pfilter != NULL, CPL_ERROR_NULL_INPUT);

    *ppscale = cpl_propertylist_get_double(self, SPH_STREHL_PSCALE);
    if (!cpl_errorstate_is_equal(okstate)) {
        if (arm == SPH_ARM_IRDIS) {
            *ppscale = SPH_STREHL_PSCALE_IRDIS;

            cpl_msg_warning(cpl_func, "Using default IRDIS pixel scale "
                            "[Arcseconds/pixel]: "
                            IRPLIB_STRINGIFY(SPH_STREHL_PSCALE_IRDIS));
            cpl_errorstate_dump(okstate, CPL_FALSE,
                                cpl_errorstate_dump_one_warning);
            cpl_errorstate_set(okstate);
        } else if (arm == SPH_ARM_ZIMPOL1 || arm == SPH_ARM_ZIMPOL2) {
            *ppscale = SPH_STREHL_PSCALE_ZIMPOL;

            cpl_msg_warning(cpl_func, "Using default ZIMPOL pixel scale "
                            "[Arcseconds/pixel]: "
                            IRPLIB_STRINGIFY(SPH_STREHL_PSCALE_ZIMPOL));
            cpl_errorstate_dump(okstate, CPL_FALSE,
                                cpl_errorstate_dump_one_warning);
            cpl_errorstate_set(okstate);
        } else {
            return cpl_error_set_where(cpl_func);
        }
    }

    if (arm == SPH_ARM_IRDIS) {

        strval = cpl_propertylist_get_string(self, SPH_STREHL_FILTER_IRD2);
        if (!cpl_errorstate_is_equal(okstate)) {
            return cpl_error_set_where(cpl_func);
        }

        assert(strval != NULL);

        if (strcmp(strval, "CLEAR") == 0) {
            strval = cpl_propertylist_get_string(self,
                                                 SPH_STREHL_FILTER_IRD1);
            if (!cpl_errorstate_is_equal(okstate)) {
                return cpl_error_set_where(cpl_func);
            }

            assert(strval != NULL);

            *pkey = SPH_STREHL_FILTER_KEY_IRD1;
        } else {
            *pkey = SPH_STREHL_FILTER_KEY_IRD2;
        }

    } else if (arm == SPH_ARM_ZIMPOL1 || arm == SPH_ARM_ZIMPOL2) {

        strval = cpl_propertylist_get_string(self, SPH_STREHL_FILTER_ZPL0);
        if (!cpl_errorstate_is_equal(okstate)) {
            return cpl_error_set_where(cpl_func);
        }

        assert(strval != NULL);

        if (strcmp(strval, "OPEN") == 0) {
            const char * key = arm == SPH_ARM_ZIMPOL1 ? SPH_STREHL_FILTER_ZPL1
                                                      : SPH_STREHL_FILTER_ZPL2;

            strval = cpl_propertylist_get_string(self, key);
            if (!cpl_errorstate_is_equal(okstate)) 
                return cpl_error_set_where(cpl_func);

            assert(strval != NULL);

            *pkey = arm == SPH_ARM_ZIMPOL1 ? SPH_STREHL_FILTER_KEY_ZPL1
                                           : SPH_STREHL_FILTER_KEY_ZPL2;
        } else {
            *pkey = SPH_STREHL_FILTER_KEY_ZPL0;
        }
    } else {
        return cpl_error_set_message(cpl_func, CPL_ERROR_UNSUPPORTED_MODE,
                                     "Unknown arm: %d", arm);
    }

    *pfilter = strval;

    return CPL_ERROR_NONE;

}

/*----------------------------------------------------------------------------*/
/**
 @brief Insert a double parameter into a parameterlist
 @param self    Parameterlist to append  parameter to
 @param recipe  Recipe name (for parameter access)
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

 */
/*----------------------------------------------------------------------------*/
static
cpl_error_code sph_strehl_fill_parlist_double(cpl_parameterlist * self,
                                              const char        * recipe,
                                              const char        * longname,
                                              const char        * shortname,
                                              const char        * mantext,
                                              double              defaultvalue)
{

    char * context = cpl_sprintf(PACKAGE ".%s", recipe);
    char * paramname = cpl_sprintf(PACKAGE ".%s.%s", recipe, longname);
    cpl_parameter * p =
        cpl_parameter_new_value(paramname, CPL_TYPE_DOUBLE, mantext,
                                context, defaultvalue);
    cpl_free(context);
    cpl_free(paramname);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, shortname);
    cpl_parameterlist_append(self, p);

    return CPL_ERROR_NONE;

}

/*----------------------------------------------------------------------------*/
/**
 @brief Read parameters from parameter list
 @param self    Parameterlist to read parameters from
 @param recipe  Recipe name (for parameter access)
 @param pstar_r Star radius [arcsecond]
 @param pbg_r1  Background internal radius [arcsecond]
 @param pbg_r2  Background external radius [arcsecond]
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

*/
/*----------------------------------------------------------------------------*/
static cpl_error_code sph_strehl_read_parlist(const cpl_parameterlist * self,
                                              const char * recipe,
                                              double * pstar_r,
                                              double * pbg_r1,
                                              double * pbg_r2)
{
    cpl_errorstate okstate = cpl_errorstate_get();


    cpl_ensure_code(self    != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pstar_r != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pbg_r1  != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(pbg_r2  != NULL, CPL_ERROR_NULL_INPUT);

    *pstar_r = irplib_parameterlist_get_double(self, PACKAGE, recipe, "star_r");
    if (!cpl_errorstate_is_equal(okstate)) {
        *pstar_r = IRPLIB_STREHL_STAR_RADIUS;
        cpl_msg_warning(cpl_func, "Recipe parameter 'star_r' missing, using def"
                        "ault: " IRPLIB_STRINGIFY(IRPLIB_STREHL_STAR_RADIUS));
        cpl_errorstate_set(okstate);
    }

    *pbg_r1 = irplib_parameterlist_get_double(self, PACKAGE, recipe, "bg_r1");
    if (!cpl_errorstate_is_equal(okstate)) {
        *pbg_r1 = IRPLIB_STREHL_BACKGROUND_R1;
        cpl_msg_warning(cpl_func, "Recipe parameter 'bg_r1' missing, using def"
                        "ault: "IRPLIB_STRINGIFY(IRPLIB_STREHL_BACKGROUND_R1));
        cpl_errorstate_set(okstate);
    }

    *pbg_r2 = irplib_parameterlist_get_double(self, PACKAGE, recipe, "bg_r2");
    if (!cpl_errorstate_is_equal(okstate)) {
       *pbg_r2 = IRPLIB_STREHL_BACKGROUND_R2;
        cpl_msg_warning(cpl_func, "Recipe parameter 'bg_r2' missing, using def"
                        "ault: "IRPLIB_STRINGIFY(IRPLIB_STREHL_BACKGROUND_R2));
        cpl_errorstate_set(okstate);
    }

    return CPL_ERROR_NONE;

}

/*----------------------------------------------------------------------------*/
/**
 @brief Insert the QC parameters into the propertylist
 @param self       Propertylist to append QC parameters to
 @param idx        The QC index ('1' for first)
 @param strehl     The QC parameter
 @param strehl_err The QC parameter
 @param pos_x      The QC parameter
 @param pos_y      The QC parameter
 @param sigma      The QC parameter
 @param star_flux  The QC parameter
 @param star_peak  The QC parameter
 @param star_bg    The QC parameter
 @param bg_noise   The QC parameter
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

 FIXME: Need 2 sets of strehl keys

 */
/*----------------------------------------------------------------------------*/
static cpl_error_code sph_strehl_fill_qc(cpl_propertylist * self,
                                         char   idx,
                                         double strehl,
                                         double strehl_err,
                                         double pos_x,
                                         double pos_y,
                                         double sigma,
                                         double star_flux,
                                         double star_peak,
                                         double star_bg,
                                         double bg_noise)
{

    char key[SPH_FLEN_KEYWORD+1];
    int psize;

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, strehl))
        return cpl_error_set_where(cpl_func);

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL_ERR, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, strehl_err))
        return cpl_error_set_where(cpl_func);

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL_POSX, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, pos_x))
        return cpl_error_set_where(cpl_func);

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL_POSY, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, pos_y))
        return cpl_error_set_where(cpl_func);

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL_SIGMA, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, sigma))
        return cpl_error_set_where(cpl_func);

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL_FLUX, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, star_flux))
        return cpl_error_set_where(cpl_func);

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL_PEAK, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, star_peak))
        return cpl_error_set_where(cpl_func);

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL_BKG, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, star_bg))
        return cpl_error_set_where(cpl_func);

    psize = snprintf(key, SPH_FLEN_KEYWORD, SPH_STREHL_QC_STREHL_BKGNOISE, idx);
    assert(psize <= SPH_FLEN_KEYWORD);
    if (cpl_propertylist_append_double(self, key, bg_noise))
        return cpl_error_set_where(cpl_func);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Insert the Strehl parameters to a parameterlist
 @param self    Parameterlist to append Strehl parameters to
 @param recipe  Recipe name (for parameter access)
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_strehl_fill_parlist(cpl_parameterlist * self,
                                       const char       * recipe)
{

    if (sph_strehl_fill_parlist_double(self, recipe, "star_r", "star_r",
                                       "The star radius [arcsecond]",
                                       IRPLIB_STREHL_STAR_RADIUS))
        return cpl_error_set_where(cpl_func);


    if (sph_strehl_fill_parlist_double(self, recipe, "bg_r1", "bg_r1",
                                       "The internal radius of the background "
                                       "[arcsecond]",
                                       IRPLIB_STREHL_BACKGROUND_R1))
        return cpl_error_set_where(cpl_func);


    if (sph_strehl_fill_parlist_double(self, recipe, "bg_r2", "bg_r2",
                                       "The external radius of the background "
                                       "[arcsecond]",
                                       IRPLIB_STREHL_BACKGROUND_R2))
        return cpl_error_set_where(cpl_func);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Compute the Strehl ratio of the brightest object in the image
 @param self         Propertylist to append Strehl number to
 @param idx          The QC index ('1' for first)
 @param arm          The SPHERE arm (IRDIS or ZIMPOL)
 @param image        Image to process
 @param rawplist     The propertylist of the input raw frame
 @param label_lambda The column label of the filter lambda
 @param label_width  The column label of the filter width
 @param filterframe  An input calibration frame with the filter wavelengths
 @param recipe       Recipe name (for parameter access)
 @param parlist      Recipe parameters
 @return   CPL_ERROR_NONE or the relevant CPL error code on error

 */
/*----------------------------------------------------------------------------*/
static cpl_error_code sph_strehl_compute(cpl_propertylist        * self,
                                         char                      idx,
                                         sph_strehl_arm            arm,
                                         const cpl_image         * image,
                                         const cpl_propertylist  * rawplist,
                                         const char              * label_lambda,
                                         const char              * label_width,
                                         const cpl_frame         * filterframe,
                                         const char              * recipe,
                                         const cpl_parameterlist * parlist)
{

    double star_r = 0.0, bg_r1 = 0.0, bg_r2 = 0.0;
    double strehl, strehl_err;
    double star_bg, star_peak, star_flux, psf_peak, psf_flux, bg_noise;
    double lam = 0.0, dlam = 0.0, pscale = 0.0, pos_x = 0.0, pos_y = 0.0;
    double sigma = 0.0;
    const char * key    = NULL;
    const char * filter = NULL;
    cpl_image * imagecast = cpl_image_get_type(image) == CPL_TYPE_FLOAT
        ? NULL : cpl_image_cast(image, CPL_TYPE_FLOAT);
    const cpl_image * imageuse = imagecast == NULL ? image : imagecast;
    cpl_error_code    code;

    cpl_ensure_code(self        != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(image       != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(rawplist    != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(filterframe != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(recipe      != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(parlist     != NULL, CPL_ERROR_NULL_INPUT);

    if (sph_strehl_read_parlist(parlist, recipe, &star_r, &bg_r1, &bg_r2))
        return cpl_error_set_where(cpl_func);

    if (sph_strehl_read_header(rawplist, arm, &pscale, &key, &filter))
        return cpl_error_set_where(cpl_func);

    if (sph_strehl_read_filter(filterframe, label_lambda, label_width, key,
                               filter, &lam, &dlam))
        return cpl_error_set_where(cpl_func);

    if (sph_strehl_find_star(image, &pos_x, &pos_y, &sigma))
        return cpl_error_set_where(cpl_func);

    code = irplib_strehl_compute(imageuse,
                                 IRPLIB_STREHL_M1, IRPLIB_STREHL_M2, lam, dlam,
                                 pscale, IRPLIB_STREHL_BOX_SIZE, pos_x, pos_y,
                                 star_r, bg_r1, bg_r2,
                                 -1, -1,
                                 &strehl,
                                 &strehl_err,
                                 &star_bg,
                                 &star_peak,
                                 &star_flux,
                                 &psf_peak,
                                 &psf_flux,
                                 &bg_noise);

#ifdef HDRL_USE_EXPERIMENTAL
    if (!code) {
        cpl_errorstate cleanstate = cpl_errorstate_get();
        hdrl_strehl_result hdrl_strehl;
        hdrl_image       * hdrl_img = hdrl_image_create(image, NULL);
        /* HDRL uses SI (m) and the mirror dimensions are given as radii */
        hdrl_parameter   * hdrl_param =
            hdrl_strehl_parameter_create(lam * 1e-6,
                                         IRPLIB_STREHL_M1 * 0.5,
                                         IRPLIB_STREHL_M2 * 0.5,
                                         pscale, pscale,
                                         star_r, bg_r1, bg_r2);

        if (hdrl_param != NULL)
            hdrl_strehl = hdrl_strehl_compute(hdrl_img, hdrl_param);

        if (!cpl_errorstate_is_equal(cleanstate)) {
            irplib_error_recover(cleanstate, "HDRL Strehl computation failed "
                                 "(keeping %g)", strehl);
        } else if (0.0 < strehl && strehl < 1.0 &&
                   (hdrl_strehl.strehl_value.data <= 0.0 ||
                    hdrl_strehl.strehl_value.data >= 1.0)) {
            cpl_msg_warning(cpl_func, "Ignoring suspicious HDRL Strehl: %g "
                            "(keeping %g)", hdrl_strehl.strehl_value.data,
                            strehl);
        } else {
            cpl_msg_info(cpl_func, "Changing Strehl: %g -> %g",
                         strehl, hdrl_strehl.strehl_value.data);
            strehl     = hdrl_strehl.strehl_value.data;
        }

        hdrl_parameter_delete(hdrl_param);
        hdrl_image_delete(hdrl_img);
    }
#endif

    cpl_image_delete(imagecast);

    if (code) return cpl_error_set_where(cpl_func);

    if (sph_strehl_fill_qc(self, idx, strehl, strehl_err, pos_x, pos_y, sigma,
                           star_flux, star_peak, star_bg, bg_noise))
        return cpl_error_set_where(cpl_func);

    return CPL_ERROR_NONE;
}
