/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#include "sph_utils.h"

#include "sph_common_keywords.h"
#include "sph_error.h"
#include "sph_double_image.h"
#include "sph_fits.h"
#include "sph_keyword_manager.h"
#include "sph_master_frame.h"
#include "sph_cube.h"
#include "sph_filemanager.h"

#include <math.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>

#include <fitsio.h>

#define DIT_EXPTIME_ACCURACY 0.001

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_utils  Sphere utility functions.
 *
 * @par Descirption:
 * This module contains routines to support the functionality in other
 * sphere DRH modules. Several utility functions are included, many of them
 * extended the functionality of the CPL.
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/
sph_error_code SPH_UTILS_GENERAL            = SPH_UTILS_ERR_START + 0;
sph_error_code SPH_UTILS_FRAMES_MISSING     = SPH_UTILS_ERR_START + 1;
sph_error_code SPH_UTILS_TOO_FINE_DITHER    = SPH_UTILS_ERR_START + 2;


/*----------------------------------------------------------------------------*/
/**
 @brief    Return the license

 @return   a character pointer to the license text.

 Returns the license text for the SPHERE data reduction library.
 */
/*----------------------------------------------------------------------------*/
const char* sph_get_license( void ) {
    char* text = NULL;

    return text;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Initialise all sphere and CPL systems.

 @return   error code

 Does some standard debug initialisation of error reporting systems for
 the Sphere API and CPL -- and initialised CPL as well. If there is an error
 returns SPH_ERROR_GENERAL. In particular the routine does the following:
 <ol>
     <li> call cpl_init() </li>
     <li> set component on </li>
     <li> set domain to Testing and on </li>
     <li> set CPL messages to console to OFF </li>
     <li> set CPL messages to file to CPL_MSG_DEBUG </li>
     <li> call sph_error_new <li>
 </ol>
 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_init_erex(void) {

    sph_error_new();
    sph_error_set_reportlevel( 0 );
    sph_keyword_manager_new();


    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Initialise all sphere and CPL systems.

 @return   error code

 Does some standard debug initialisation of error reporting systems for
 the Sphere API and CPL -- and initialised CPL as well. If there is an error
 returns SPH_ERROR_GENERAL. In particular the routine does the following:
 <ol>
     <li> call cpl_init() </li>
     <li> set component on </li>
     <li> set domain to Testing and on </li>
     <li> set CPL messages to console to OFF </li>
     <li> set CPL messages to file to CPL_MSG_DEBUG </li>
     <li> call sph_error_new <li>
 </ol>
 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_init(void) {

#if defined CPL_VERSION_CODE && CPL_VERSION_CODE >= CPL_VERSION(4, 0, 0)
    cpl_init(CPL_INIT_DEFAULT);
#else
    cpl_init();
#endif

    cpl_msg_set_time_on();
    cpl_msg_set_component_on();
    cpl_msg_set_domain("Testing");
    cpl_msg_set_domain_on();
    cpl_msg_set_level( CPL_MSG_OFF );
    cpl_msg_set_log_level( CPL_MSG_ERROR );

    sph_error_new();
    sph_error_set_reportlevel( 3 );
    sph_keyword_manager_new();


    return CPL_ERROR_NONE;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Initialise all sphere and CPL systems for testing.

 @return   error code

 Does some standard debug initialisation of error reporting systems for
 the Sphere API and CPL -- and initialised CPL as well. If there is an error
 returns SPH_ERROR_GENERAL. In particular the routine does the following:
 <ol>
     <li> call cpl_init() </li>
     <li> set component on </li>
     <li> set domain to Testing and on </li>
     <li> set CPL messages to console to OFF </li>
     <li> set CPL messages to file to CPL_MSG_DEBUG </li>
     <li> call sph_error_new <li>
 </ol>
 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_init_test(const char* filename, const char* report, cpl_msg_severity lvl) {

    cpl_test_init_macro(filename,report,lvl);
    cpl_msg_set_level( CPL_MSG_OFF );
    sph_error_new();
    sph_error_set_reportlevel( 0 );
    sph_keyword_manager_new();

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Update a keyword which is numbered
 * @param proplist          pointer to the propertylist
 * @param name              the keyword name, with an "i" in it
 * @param type              the type of the keyword
 * @param i                 the i index
 * @param j                 the j index (give -1 if not needed)
 * @param value             the value as a double
 * @return error code
 *
 * This function updates the properlist keyword with the given name
 * to the given value. The name must contain an "i" and may
 * contain a "j". These are then replaced by the corresponding
 * i and j values given (using sprintf).
 *
 * The format of the value must be BOOL, INT, FLOAT or DOUBLE.
 * An explicit cast is made from the double value passed to the type
 * specified.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_utils_update_key( cpl_propertylist* p,
        const char* name,
        cpl_type type,
        int i, int j,
        double value )
{
    int         ii      = 0;
    int         ipos    = -1;
    int         jpos    = -1;
    char        name1[256];
    char        name2[256];
    char        name3[256];
    char        finalname[256];
    int         endpos = 0;
    cpl_ensure_code(p,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(name,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(type == CPL_TYPE_INT ||
            type == CPL_TYPE_BOOL ||
            type == CPL_TYPE_DOUBLE ||
            type == CPL_TYPE_FLOAT,
            CPL_ERROR_TYPE_MISMATCH);
    cpl_ensure_code(i>=0,CPL_ERROR_ILLEGAL_INPUT);

    for (ii = 0; ii < strlen(name); ++ii) {
        if (name[ii] == 'i') ipos = ii;
        if (name[ii] == 'j') jpos = ii;
    }

    if ( ipos < 0 ) {
        cpl_ensure_code(0,CPL_ERROR_ILLEGAL_INPUT);
    }

    if ( jpos >= 0 && j < 0 ) {
        cpl_ensure_code(0,CPL_ERROR_INCOMPATIBLE_INPUT);
    }
    if ( jpos < 0 && j >= 0 ) {
        cpl_ensure_code(0,CPL_ERROR_ILLEGAL_INPUT);
    }
    strncpy(name1,name,ipos);
    name1[ipos]='\0';
    if ( jpos > 0 ) {
        endpos = jpos;
    }
    else {
        endpos = strlen(name);
    }

    if ( ipos < strlen(name) ) {
        strncpy(name2,&name[ipos+1],endpos - ipos - 1);
        name2[endpos-ipos-1] = '\0';
    }
    else {
        name2[0]='\0';
    }
    if ( jpos > 0 ) {
        if ( jpos < strlen(name) ) {
            strncpy(name3,&name[jpos+1],strlen(name) - jpos - 1);
            name3[strlen(name)-jpos-1] = '\0';
        }
        else {
            name3[0]='\0';
        }
        sprintf(finalname,"%s%d%s%d%s",name1,i,name2,j,name3);
    }
    else {
        sprintf(finalname,"%s%d%s",name1,i,name2);
    }
    if ( type == CPL_TYPE_BOOL ) {
        cpl_propertylist_update_bool(p,finalname,(int)value);
    }
    if ( type == CPL_TYPE_INT ) {
        cpl_propertylist_update_int(p,finalname,(int)value);
    }
    if ( type == CPL_TYPE_FLOAT ) {
        cpl_propertylist_update_float(p,finalname,(float)value);
    }
    if ( type == CPL_TYPE_DOUBLE ) {
        cpl_propertylist_update_double(p,finalname,value);
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief copy singular keys from one propertylist to another
 * @param proplist_in          pointer to the input propertylist
 * @param proplist_out         pointer to the output propertylist
 * @return error code
 *
 * This function simply copies all singular (i.e. non-HIERARCH) header keys
 * from the input to the output
 *
 * This is a workaround because certain singular keys are not copied by
 * cpl_save_dfs by default!
 *
 * Overwrites only if the target property is not already present in the
 * destination property list!
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_utils_simple_copy_singular( cpl_propertylist* pin, cpl_propertylist* pout)
{

	cpl_propertylist*  ptmp = NULL;
	const char*        ppty = NULL;
	int                ndel = 0;
	int                nsin = 0;
	int                pos  = 0;
	//
	//
	cpl_ensure_code(pin, CPL_ERROR_NULL_INPUT);
	cpl_ensure_code(pout, CPL_ERROR_NULL_INPUT);
	ptmp = cpl_propertylist_duplicate(pin);
	ndel = cpl_propertylist_erase_regexp(ptmp,"^ESO",0);
	nsin = cpl_propertylist_get_size(ptmp);


	if (nsin > 0){
		for (pos=0;pos<nsin;pos++){
			ppty = cpl_property_get_name (cpl_propertylist_get(ptmp,pos));
			if (ppty){
				if (!cpl_propertylist_has(pout,ppty)){
					cpl_propertylist_copy_property (pout,ptmp,ppty);
				}
			}
		}
	}

	ppty = NULL;
	cpl_propertylist_delete(ptmp); ptmp=NULL;
	SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief determine EXPTIME from raw frames and put into a provided property list
 * @param rawframes            pointer to raw framset
 * @param proplist_out         pointer to the output propertylist
 * @return error code
 *
 * Currently this function simply takes the EXPTIME keyword from the first
 * * raw frame and copies it into the output plist
 *
 * Overwrites only if the target property is not already present in the
 * destination property list!
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_utils_get_exptime_from_raw( cpl_frameset* rawframes, cpl_propertylist* pout)
{


	cpl_propertylist*  ptmp = NULL;
	const char*        ppty = NULL;
	//
	//
	ptmp = cpl_propertylist_load(cpl_frame_get_filename(cpl_frameset_get_first(rawframes)),
		        	                    0);
	ppty = cpl_property_get_name (cpl_propertylist_get_property(ptmp, SPH_COMMON_KEYWORD_EXPTIME2));
	if (ppty){
		SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
				"Copying EXPTIME of %f seconds to result header!", cpl_propertylist_get_float(ptmp,"EXPTIME"));
		cpl_propertylist_copy_property (pout,ptmp,ppty);
	}
	ppty = NULL;
	cpl_propertylist_delete(ptmp); ptmp=NULL;
	SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}



/*----------------------------------------------------------------------------*/
/**
 * @brief Update a list of keyword to vector values
 * @param proplist          pointer to the propertylist
 * @param name              the keyword name, with an "i" in it
 * @param values            the values as a vector
 * @return error code
 *
 * This function updates the propertylist keyword with the given name
 * to the given vector values. The name must contain an "i".
 * A keyword is updated for each entry in the vector.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_utils_update_key_vector( cpl_propertylist* p,
        const char* name,
        cpl_vector* values )
{
    int         ii   = 0;
    sph_error_code rerr = CPL_ERROR_NONE;
    cpl_ensure_code(p,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(name,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(values,CPL_ERROR_NULL_INPUT);

    for (ii = 0; ii < cpl_vector_get_size(values); ++ii) {
       rerr = sph_utils_update_key(p,name,
               CPL_TYPE_DOUBLE,ii + 1,-1,
               cpl_vector_get(values,ii) );

       if ( rerr != CPL_ERROR_NONE ) {
           return rerr;
       }
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Initialise all sphere and CPL systems verbosely.

 @return   error code

 Does some standard debug initialisation of error reporting systems for
 the Sphere API and CPL -- and initialised CPL as well. If there is an error
 returns SPH_ERROR_GENERAL. In particular the routine does the following:
 <ol>
     <li> call cpl_init() </li>
     <li> set component on </li>
     <li> set domain to Testing and on </li>
     <li> set CPL messages to console to OFF </li>
     <li> set CPL messages to file to CPL_MSG_DEBUG </li>
     <li> call sph_error_new <li>
 </ol>
 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_init_debug_verbose(void) {

    cpl_init(CPL_INIT_DEFAULT);

    cpl_msg_set_time_on();
    cpl_msg_set_component_on();
    cpl_msg_set_domain("Testing");
    cpl_msg_set_domain_on();
    cpl_msg_set_level( CPL_MSG_DEBUG );
    cpl_msg_set_log_level( CPL_MSG_DEBUG );

    sph_error_new();
    sph_error_set_reportlevel( 0 );
    sph_keyword_manager_new();

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Initialise all sphere and CPL systems.

 @return   error code

 Does some standard debug initialisation of error reporting systems for
 the Sphere API and CPL -- and initialised CPL as well. If there is an error
 returns SPH_ERROR_GENERAL. In particular the routine does the following:
 <ol>
     <li> call cpl_init() </li>
     <li> set component on </li>
     <li> set domain to Testing and on </li>
     <li> set CPL messages to console to OFF </li>
     <li> set CPL messages to file to CPL_MSG_DEBUG </li>
     <li> call sph_error_new <li>
 </ol>
 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_init_debug(void) {

#if defined CPL_VERSION_CODE && CPL_VERSION_CODE >= CPL_VERSION(4, 0, 0)
    cpl_init(CPL_INIT_DEFAULT);
#else
    cpl_init();
#endif

    cpl_msg_set_time_on();
    cpl_msg_set_component_on();
    cpl_msg_set_domain("Testing");
    cpl_msg_set_domain_on();
    cpl_msg_set_level( CPL_MSG_OFF );
    cpl_msg_set_log_level( CPL_MSG_DEBUG );

    sph_error_new();
    sph_error_set_reportlevel( 0 );
    sph_keyword_manager_new();


    return CPL_ERROR_NONE;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    End all sphere and CPL systems.

 @return   error code

 Closes the Sphere API and CPL system.
 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_end_erex(void)
{
  sph_filemanager_delete(0);
  sph_keyword_manager_delete();
  sph_error_delete();
  sph_polygon_free_all();
  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    End all sphere and CPL systems.

 @return   error code

 Closes the Sphere API and CPL system.
 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_end(void)
{
  sph_keyword_manager_delete();
  sph_error_delete();
  sph_polygon_free_all();
  cpl_end();
  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    End all sphere and CPL systems.

 @return   error code

 Closes the Sphere API and CPL system.
 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_end_test(void)
{
  sph_filemanager_delete(1);
  sph_keyword_manager_delete();
  sph_error_delete();
  sph_polygon_free_all();
  return cpl_test_end(0);
}



/*----------------------------------------------------------------------------*/
/**
 @brief    Append a frameset to another one
 @param    frameset   The input frameset to append to
 @param    appendset  The input frameset to append

 @return   error code.

 Append all frames from appendset to frameset. The appendset not affected.
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_utils_append_frames( cpl_frameset* frameset, const cpl_frameset* appendframes ) {
    int    ii    = 0;
    cpl_frame*    aframe = NULL;

    cpl_ensure_code( frameset, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code( appendframes, CPL_ERROR_NULL_INPUT);

    for (ii = 0; ii < cpl_frameset_get_size( appendframes ); ++ii) {
        aframe = cpl_frame_duplicate( cpl_frameset_get_position_const( appendframes, ii ) );
        cpl_frameset_insert( frameset, aframe );
    }

    return CPL_ERROR_NONE;

}

/*----------------------------------------------------------------------------*/
/**
 @brief    Delete files from the hard disk
 @param    frameset   The input frameset which files to be deleted from
 @return   error code.

Delete files from the hard disk
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_utils_frames_unlink( cpl_frameset* frameset ) {
    int    ii    = 0;

    cpl_ensure_code( frameset, CPL_ERROR_NULL_INPUT);

    for (ii = 0; ii < cpl_frameset_get_size( frameset ); ++ii) {
        unlink( cpl_frame_get_filename(
                        cpl_frameset_get_frame(frameset, ii)));
    }
    return CPL_ERROR_NONE;
}



/*----------------------------------------------------------------------------*/
/**
 @brief    Extract  frames with a specific keyword setting from a frameset
 @param    frames   The input frames
 @param    tag       A string identifying the  tag of the frames to extract

 @return   new cpl_frameset containing the extracted frames.

 Extract a set of frames from an input frameset which share a common tag.
 */
/*----------------------------------------------------------------------------*/
cpl_frameset*
sph_utils_extract_frames( cpl_frameset* frameset, const char* tag ) {
    cpl_frameset*   RawFlatList     = NULL;
    cpl_frame*      aRawFlat        = NULL;
    int             rerr            = 0;

    if (cpl_error_get_code() != CPL_ERROR_NONE) {
         sph_error_raise( SPH_UTILS_GENERAL,
                          __FILE__, __func__, __LINE__ ,
                          SPH_ERROR_WARNING,
                          "Attention cpl error has already been set!  The CPL message: %s\n",
                          cpl_error_get_message() );
         cpl_error_reset();
     }


    if ( !frameset ) {
        SPH_NO_SELF;
        return NULL;
    }
    if ( ! tag ) {
        SPH_NULL_ERROR;
        return NULL;
    }

    RawFlatList = cpl_frameset_new();
    aRawFlat = cpl_frame_duplicate( cpl_frameset_find(frameset, tag) );
    if ( ! aRawFlat )
    {
        if ( RawFlatList ) {
            cpl_frameset_delete( RawFlatList );
        }
        cpl_error_reset();
        return NULL;
    }

    rerr = cpl_error_get_code();
    while ( aRawFlat )
    {
        cpl_frameset_insert( RawFlatList, aRawFlat );
        aRawFlat = cpl_frame_duplicate( cpl_frameset_find( frameset, NULL ) );
    }
    rerr = cpl_error_get_code();
    if ( rerr == CPL_ERROR_NULL_INPUT ) {
        cpl_error_reset();
    }
    return RawFlatList;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Extract  frames with a specific group setting from a frameset
 @param    frames   The input frames
 @param    tag       A string identifying the  group of the frames to extract

 @return   new cpl_frameset containing the extracted frames.

 Extract a set of frames from an input frameset which share a common group.
 */
/*----------------------------------------------------------------------------*/
cpl_frameset*
sph_utils_extract_frames_group( cpl_frameset* frameset, cpl_frame_group group ) {
    cpl_frameset*   RawFlatList     = NULL;
    cpl_frame*      aRawFlat        = NULL;
    int             rerr            = 0;

    if (cpl_error_get_code() != CPL_ERROR_NONE) {
         sph_error_raise( SPH_UTILS_GENERAL,
                          __FILE__, __func__, __LINE__ ,
                          SPH_ERROR_WARNING,
                          "Attention cpl error has already been set!  The CPL message: %s\n",
                          cpl_error_get_message() );
         cpl_error_reset();
     }


    if ( !frameset ) {
        SPH_NO_SELF;
        return NULL;
    }

    RawFlatList = cpl_frameset_new();
    aRawFlat = cpl_frame_duplicate( cpl_frameset_get_first(frameset ) );
    if ( ! aRawFlat )
    {
        if ( RawFlatList ) {
            cpl_frameset_delete( RawFlatList );
        }
        cpl_error_reset();
        return NULL;
    }

    rerr = cpl_error_get_code();
    while ( aRawFlat )
    {
        if ( cpl_frame_get_group(aRawFlat) == group ) {
            cpl_frameset_insert( RawFlatList, aRawFlat );
        }
        else {
            cpl_frame_delete( aRawFlat ); aRawFlat = NULL;
        }
        aRawFlat = cpl_frame_duplicate( cpl_frameset_get_next( frameset) );
    }
    rerr = cpl_error_get_code();
    if ( rerr == CPL_ERROR_NULL_INPUT ) {
        cpl_error_reset();
    }
    return RawFlatList;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Extract  frame with a specific keyword setting from a frameset
 @param    frames   The input frames
 @param    tag       A string identifying the  tag of the frame to extract

 @return   new cpl_frame containing of the extracted frame.

 Extract a frame from an input frameset which share a common tag.
 */
/*----------------------------------------------------------------------------*/
cpl_frame*
sph_utils_extract_frame( cpl_frameset* frameset, const char* tag ) {
    cpl_frame*      aRawFlat        = NULL;
    int             rerr            = 0;

    if (cpl_error_get_code() != CPL_ERROR_NONE) {
         sph_error_raise( SPH_UTILS_GENERAL,
                          __FILE__, __func__, __LINE__ ,
                          SPH_ERROR_WARNING,
                          "Attention cpl error has already been set!  The CPL message: %s\n",
                          cpl_error_get_message() );
         cpl_error_reset();
     }


    if ( !frameset ) {
        SPH_NO_SELF;
        return NULL;
    }
    if ( ! tag ) {
        SPH_NULL_ERROR;
        return NULL;
    }

    aRawFlat = cpl_frame_duplicate( cpl_frameset_find(frameset, tag) );
    if ( ! aRawFlat )
    {
        sph_error_raise( SPH_UTILS_FRAMES_MISSING,
                         __FILE__, __func__, __LINE__ ,
                         SPH_ERROR_WARNING,
                         "No %s frames found!\n"
                         "Check that they have the %s tag.\n",
                         tag, tag );
        rerr = CPL_ERROR_NULL_INPUT;
        cpl_error_reset();
        return NULL;
    }

    return aRawFlat;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Create a standard calibration product propertylist

 @return   new cpl_propertylist or NULL on error.

 Create a new standard calibration product propertylist, with IFS specific
 properties.

 */
/*----------------------------------------------------------------------------*/
cpl_propertylist*
sph_utils_create_calib_propertylist( void ) {
    cpl_propertylist*       proplist            = NULL;
    int                     timeref             = 0;
    int                     status              = 0;
    char                    timestamp[256];

    proplist = cpl_propertylist_new();

    fits_get_system_time( timestamp, &timeref, &status );

    cpl_propertylist_append_string( proplist, SPH_COMMON_KEYWORD_PRO_REDLEVEL, SPH_COMMON_KEYWORD_VALUE_PRO_REDLEVEL_FULL );
    cpl_propertylist_append_string( proplist, SPH_COMMON_KEYWORD_PRO_REC1_START, timestamp );
    cpl_propertylist_append_string( proplist, SPH_COMMON_KEYWORD_PRO_REC1_STOP, "UNDEFINED" );

    return proplist;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Create a standard calibration product frame

 @param     outframename     The name of the new frame
 @param     frametype        The type of the new frame
 @param     framelevel       The reduction level of the new frame
 @param     inframes         The frameset of the input frames used to create
                             the new frame
 @param     recid            The name of the recipe creating this frame
 @param     pipelineid       The name/id of the pipeline
 @param     dicid            The ID of the sph_dictionary used

 @return   new cpl_frame or NULL on error.

 Create a new standard calibration product propertylist, with IFS specific
 properties.

 */
/*----------------------------------------------------------------------------*/
cpl_frame*
sph_utils_create_calib_frame( const char* outframename, cpl_frame_type frametype,
                              cpl_frame_level framelevel, cpl_frameset* inframes,
                              const cpl_parameterlist* parlist, const char* prod_tag,
                              const char* recid, const char* pipelineid, cpl_frame* inherit_frame )
{
    cpl_frame*              result      = NULL;
    cpl_propertylist*       proplist    = NULL;
    cpl_frameset*           fset        = NULL;
    int                     ii          = 0;
    cpl_error_code          rerr        = CPL_ERROR_NONE;

    result = cpl_frame_new();
    rerr |= cpl_frame_set_filename( result, outframename );
    rerr |= cpl_frame_set_group( result, CPL_FRAME_GROUP_PRODUCT );
    rerr |= cpl_frame_set_level(result, framelevel );
    rerr |= cpl_frame_set_type( result, frametype );
    rerr |= cpl_frame_set_tag( result, prod_tag );

    proplist = sph_utils_create_calib_propertylist( );

    if ( inherit_frame ){
        rerr |= cpl_dfs_setup_product_header( proplist, result,
                                      inframes, parlist,
                                      recid, pipelineid, SPH_COMMON_KEYWORD_VALUE_DIC_ID, inherit_frame );
    } else {
    	rerr |= cpl_dfs_setup_product_header( proplist, result,
                                  inframes, parlist,
                                  recid, pipelineid, SPH_COMMON_KEYWORD_VALUE_DIC_ID, cpl_frameset_get_first( inframes ) );
    }
    if ( rerr == CPL_ERROR_DATA_NOT_FOUND ) {
        cpl_error_reset();                          /* It seems as if cpl_dfs_setup_
                                                        product header always sets
                                                        the data not found error
                                                        even if the inframes list
                                                        is not null.
                                                        Need to check this with Lars Lundin
                                                    */



        rerr = CPL_ERROR_NONE;
    }
    rerr |= sph_fits_update_checksum( outframename );
    rerr |= cpl_propertylist_append_string( proplist, SPH_COMMON_KEYWORD_PRO_CATG, prod_tag);

    for (ii = 0; ii < cpl_propertylist_get_size(proplist); ++ii) {
        rerr |= sph_fits_update_property( outframename, cpl_propertylist_get( proplist, ii), 0);
    }

    fset = cpl_frameset_new( );
    rerr |= cpl_frameset_insert( fset, cpl_frame_duplicate( result ) );
    rerr |= cpl_dfs_update_product_header( fset );
    cpl_frameset_delete( fset );
    if ( rerr != CPL_ERROR_NONE ) {
        SPH_RAISE_CPL
    }
    cpl_frameset_insert(inframes,result);
    cpl_propertylist_delete( proplist );
    return result;
}


/*----------------------------------------------------------------------------*/
/**
 @brief    Create a standard list of valid DITs as a FITS table

 @param     outframename     The name of the new frame

 @return   new cpl_frame or NULL on error.

 Create a new standard calibration valid dit list.

 */
/*----------------------------------------------------------------------------*/
cpl_frame*
sph_utils_create_default_dit_list( const char* fname ) {
    cpl_frame*      result      = NULL;
    int             ii          = 0;
    int             jj          = 0;
    int             numnreads   = 10;
    int             numexps     = 500;
    int             numrows     = numnreads * numexps;
    float           exptime     = 1.3;
    int             rerr        = CPL_ERROR_NONE;
    int             nrow        = 0;
    cpl_table*        dit_table    = NULL;

    if (cpl_error_get_code() != CPL_ERROR_NONE) {
         sph_error_raise( SPH_UTILS_GENERAL,
                          __FILE__, __func__, __LINE__ ,
                          SPH_ERROR_WARNING,
                          "Attention cpl error has already been set!  The CPL message: %s\n",
                          cpl_error_get_message() );
         cpl_error_reset();
     }


    dit_table   = cpl_table_new(numrows);


    if (cpl_table_new_column(dit_table, "NREADOUTS", CPL_TYPE_INT)
            != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: Error adding column: %s", cpl_func,
                __LINE__, cpl_error_get_where());
        return NULL;

    }
    if (cpl_table_new_column(dit_table, "DIT", CPL_TYPE_DOUBLE)
            != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: Error adding column: %s", cpl_func,
                __LINE__, cpl_error_get_where());
        return NULL;
    }
    for ( ii = 0; ii < numnreads; ii++ ) {
        for ( jj = 0; jj < numexps; jj++) {
            cpl_table_set(dit_table, "NREADOUTS", nrow, ii+1);
            cpl_table_set(dit_table, "DIT", nrow, (jj+1) * exptime);
            nrow++;
        }
    }
    rerr = cpl_table_save(dit_table, NULL, NULL, fname, CPL_IO_DEFAULT);

    result = cpl_frame_new();
    cpl_frame_set_filename( result, fname );
    cpl_frame_set_group( result, CPL_FRAME_GROUP_CALIB );
    cpl_frame_set_level( result, CPL_FRAME_LEVEL_TEMPORARY );
    cpl_frame_set_type( result, CPL_FRAME_TYPE_TABLE );
    cpl_table_delete( dit_table );
    return result;
}


/*----------------------------------------------------------------------------*/
/**
 *    @brief Creates a new frameset with master_frames from the input frameset.
 *
 *    @param    inframes        the input frames to duplicate
 *
 *  @return    the new cpl_frameset or NULL in case of error
 *
 *  This function creates a copy of all input frames, where each frame in the
 *  new framelist points to a file which contains either a sph_master_frame
 *  or a sph_cube. The input frameset can contain frames of these types:
 *  <ul>
 *  <li>a raw frame with a single image (one extension, one plane)</li>
 *  <li>a sph_master_frame with 4 extensions and one plane each</li>
 *  <li>a sph_cube with 4 extensions and N planes each.</li>
 *  </ul>
 *  The function automatically transforms any raw frame inputs into a
 *  sph_master_frame by filling the image extension with the image and
 *  leaving the other extensions with filled with 0.
 *  The sph_master_frame and sph_cube inputs are copied to a new file unchanged.
 */
/*----------------------------------------------------------------------------*/
cpl_frameset*
sph_utils_create_master_frameset( cpl_frameset* inframes ) {
    cpl_frameset*            result        = NULL;
    cpl_frame*                curframe    = NULL;
    cpl_frame*                dupframe    = NULL;
    cpl_propertylist*        pl            = NULL;
    sph_master_frame*        mframe        = NULL;
    sph_keyword_manager*    keym        = NULL;
    const char*                    czFrame        = NULL;
    char                    czFrameNew[256];
    const char*                    czType        = NULL;
    short                    added        = 0;

    result = cpl_frameset_new();
    if ( !result ) {
        return NULL;
    }
    keym = sph_keyword_manager_new();
    if ( !keym ) {
        return NULL;
    }
    curframe = cpl_frameset_get_first( inframes );

    while ( curframe ) {
        dupframe = cpl_frame_duplicate( curframe );
        czFrame = cpl_frame_get_filename( curframe );
        strncpy( czFrameNew, czFrame, 256 );
        strcat( czFrameNew, ".tmp");
        sph_keyword_manager_load_keywords( czFrame,0 );
        czType = sph_keyword_manager_get_string( SPH_COMMON_KEYWORD_SPH_TYPE );
        added = 0;
        if ( czType ) {
            if ( !strcasecmp( czType, SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_CUBE ) ||
                 !strcasecmp( czType, SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_MASTER_FRAME ) )
            {
                cpl_frame_set_filename( dupframe, czFrameNew );
                cpl_frameset_insert( result, dupframe );
                added = 1;
            }
        }
        if ( !added ) {
            pl = sph_keyword_manager_load_properties( czFrame, 0 );
            mframe = sph_master_frame_new_empty();
            if ( mframe ) {
              mframe->image = cpl_image_load(czFrame, CPL_TYPE_DOUBLE, 0, 0 );
              if ( mframe->image ) {
                mframe->badpixelmap = cpl_image_new( cpl_image_get_size_x(mframe->image),
                                 cpl_image_get_size_y( mframe->image ), CPL_TYPE_INT );
                mframe->ncombmap = cpl_image_new( cpl_image_get_size_x(mframe->image),
                                  cpl_image_get_size_y( mframe->image ), CPL_TYPE_INT );
                mframe->rmsmap = cpl_image_new( cpl_image_get_size_x(mframe->image),
                                cpl_image_get_size_y( mframe->image ), CPL_TYPE_DOUBLE );
                sph_master_frame_save( mframe, czFrameNew, pl );
                sph_master_frame_delete( mframe ); mframe = NULL;
                cpl_frame_set_filename( dupframe, czFrameNew );
                cpl_frameset_insert( result, dupframe );
                added = 1;
              }
            }
            cpl_propertylist_delete(pl); pl = NULL;
        }
        curframe = cpl_frameset_get_next( inframes );
    }


    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Creates temporary copies of input frames for further processing.
 *
 * @param    inframes        the input frames to copy
 * @param    framesout        a frameset to which the temporary frames will be added
 *
 * return        error code of the operation
 *
 * This creates a temporary copy of the file pointer to by each frame and add
 * it to the output frameslist framesout. This function cann handle all RAW
 * input frames, and frames of type SPH_MASTER_FRAME and SPH_DOUBLE_IMAGE but
 * not of type SPH_CUBE. A frame of that type in the input causes a
 * CPL_ERROR_INCOMPATIBLE_INPUT to be returned.
 * @deprecated
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_utils_copy_frames( const cpl_frameset* inframes, cpl_frameset* framesout )
{
    sph_error_code             rerr         = CPL_ERROR_NONE;
    const cpl_frame*                curframe    = NULL;
    cpl_frame*                frameA        = NULL;
    cpl_propertylist*        pl            = NULL;
    sph_master_frame*        mframe        = NULL;
    sph_double_image*        dimage        = NULL;
    sph_keyword_manager*    keym        = NULL;
    const char*                    czFrame        = NULL;
    char                    czFrameA[2056];
    const char*                    czType        = NULL;
    short                    sphformat    = 0;
    cpl_image*                image        = NULL;

    keym = sph_keyword_manager_new();
    if ( !keym ) {
        return CPL_ERROR_NULL_INPUT;
    }

    curframe = cpl_frameset_get_first_const( inframes );

    while ( curframe ) {
        czFrame = cpl_frame_get_filename( curframe );
        sph_keyword_manager_load_keywords( czFrame, 0 );
        czType = sph_keyword_manager_get_string( SPH_COMMON_KEYWORD_SPH_TYPE );
        if ( !czType ) {
            cpl_error_reset();
        }
        sphformat = 0;
        strncpy( czFrameA, czFrame, 2056 );
        strcat( czFrameA, ".tmp");
        if ( czType ) {
            if ( !strcasecmp( czType, SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_CUBE ) )
            {
                return CPL_ERROR_INCOMPATIBLE_INPUT;
            }
            if ( !strcasecmp( czType, SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_MASTER_FRAME ) )
            {
                pl = sph_keyword_manager_load_properties( czFrame, 0 );
                mframe = sph_master_frame_load( czFrame, 0 );
                sph_master_frame_save( mframe, czFrameA, pl );
                sph_master_frame_delete( mframe );
                sphformat = 1;
            }
            if ( !strcasecmp( czType, SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_DOUBLE_IMAGE ) )
            {
                pl = sph_keyword_manager_load_properties( czFrame, 0 );
                dimage = sph_double_image_load( czFrame, 0 );
                sph_double_image_save( dimage, czFrameA, pl ,CPL_IO_CREATE);
                sph_double_image_delete( dimage );
                sphformat = 1;
            }
        }
        if ( !sphformat ) {
            image = cpl_image_load(czFrame,
                    CPL_TYPE_DOUBLE, 0, 0 );
            mframe = sph_master_frame_new_from_cpl_image( image );
            cpl_image_delete( image ); image = NULL;
            if ( mframe->properties ) {
                cpl_propertylist_delete( mframe->properties );
                mframe->properties = NULL;
            }
            mframe->properties = sph_keyword_manager_load_properties( czFrame, 0 );
            sph_master_frame_save( mframe, czFrameA, pl );
            sph_master_frame_delete( mframe );mframe = NULL;
        }
        frameA = cpl_frame_duplicate( curframe );
        cpl_frame_set_filename( frameA, czFrameA );
        cpl_frameset_insert( framesout, frameA );
        curframe = cpl_frameset_get_next_const( inframes );
    }
    return rerr;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Check frame against a DIT list and return the DIT ID for the frame

 @param     inframes        The frame to check for DIT ID
 @param     dit_list        The valid dit list to use for the check

 @return   dit id, or -1 on error.

 Check the validitiy of the EXPTIME and DET READOUT NO keywords and return the
 associated DIT ID.
 In case the frame has an invalid DIT value an SPH_INVALID_DIT
 error code is returned. The routine assumes the EXPTIME and DET READOUT NO is
 stored for the entire frame in the primary HDU.

 */
/*----------------------------------------------------------------------------*/
int
sph_utils_get_dit_id_from_cpl_table( const cpl_frame* inframe, const cpl_table* dit_table )
{
    cpl_propertylist*   proplist        = NULL;
    const char*               filename        = NULL;
    double              exptime         = 0.0;
    int                 readno          = 0;
    int                 ii              = 0;
    int                 found           = 0;
    int                 temp_nreads     = 0;
    int                 badpix          = 0;
    int                 dit_id          = -1;
    double              temp_exptime    = 0;

    filename        = cpl_frame_get_filename( inframe );
    proplist        = sph_keyword_manager_load_properties( filename, 0 );

    exptime = cpl_propertylist_get_double( proplist, SPH_COMMON_KEYWORD_EXPTIME );
    readno  = cpl_propertylist_get_int( proplist, SPH_COMMON_KEYWORD_NREADS );
    sph_error_raise( SPH_ERROR_INFO, __FILE__, __func__, __LINE__, SPH_ERROR_INFO, "Assing FITS dit from extime = %g and ROM = %d", exptime, readno );
    for ( ii = 0; ii < cpl_table_get_nrow(dit_table); ii++ )
    {
        temp_exptime    = cpl_table_get_double( dit_table,
                                                "DIT", ii, &badpix );
        temp_nreads     = cpl_table_get_int( dit_table,
                                                "NREADOUTS", ii, &badpix );
        if ( temp_nreads == readno && fabs( exptime - temp_exptime ) < DIT_EXPTIME_ACCURACY ) {
            found = found + 1;
            dit_id = ii + 1;
            sph_error_raise( SPH_ERROR_INFO, __FILE__, __func__, __LINE__, SPH_ERROR_INFO, "Found dit id %d", dit_id );
        }
    }
    cpl_propertylist_delete( proplist );
    if ( dit_id < 1 ) {
        sph_error_raise( SPH_ERROR_GENERAL,
                         __FILE__, __func__,
                         __LINE__,
                         SPH_ERROR_WARNING,
                         "Could not find a DIT id for the combination of "
                         "DIT = %g and NREADOUTS = %d "
                         "in the provided (or default) valid DIT table.",
                         exptime, readno );
    }
    return dit_id;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Check frame against a DIT list and return the DIT ID for the frame

 @param     inframes        The frame to check for DIT ID
 @param     dit_list        The frame of the valid dit list to use for the check

 @return   dit id, or -1 on error.

 Check the validitiy of the EXPTIME and DET READOUT NO keywords and return the
 associated DIT ID.
 In case the frame has an invalid DIT value an SPH_INVALID_DIT
 error code is returned. The routine assumes the EXPTIME and DET READOUT NO is
 stored for the entire frame in the primary HDU.

 */
/*----------------------------------------------------------------------------*/
int
sph_utils_get_dit_id( const cpl_frame* inframe, const cpl_frame* dit_list )
{
    const char*         dit_filename    = NULL;
    cpl_table*          dit_table       = NULL;
    int                    dit_id            = -1;

    dit_filename    = cpl_frame_get_filename( dit_list );

    dit_table = cpl_table_load( dit_filename, 1, 0 );
    if ( dit_table ) {
        dit_id = sph_utils_get_dit_id_from_cpl_table( inframe, dit_table );

        cpl_table_delete( dit_table );
    }
    return dit_id;
}


double
sph_utils_get_exposure_time( const cpl_frame* inframe ) {
	cpl_ensure(inframe,CPL_ERROR_ILLEGAL_INPUT,1.0);
	cpl_propertylist* pl = NULL;
	double exptime = 1.0;

    pl = sph_keyword_manager_load_properties(cpl_frame_get_filename(inframe),0);
	if (  cpl_propertylist_has(pl,SPH_COMMON_KEYWORD_EXPTIME) )
		exptime = cpl_propertylist_get_double(pl,SPH_COMMON_KEYWORD_EXPTIME);
	cpl_propertylist_delete(pl);pl = NULL;
	return exptime;
}

int
sph_utils_check_exposure_times_are_same( const cpl_frameset* inframes ) {
	cpl_ensure(inframes,CPL_ERROR_ILLEGAL_INPUT,0);
	cpl_frame* frame = NULL;
	double exptime = 1.0;

	frame = cpl_frameset_get_first_const( inframes );
	exptime = sph_utils_get_exposure_time(frame);
	while ( frame ) {
		if ( fabs(sph_utils_get_exposure_time(frame) - exptime ) > 0.000001 ) {
			return 0;
		}
		frame = cpl_frameset_get_next_const(inframes);
	}
	return 1;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Return the DITHER ID for the frame

 @param     inframe         The frame to return the dither id for
 @param     res             The dither resolution allowed

 @return   dither id, or -1 on error.

 This function returns a dither "ID" which is constructed from the x and y
 position of the detector. The ID allows for a given dither resolution --
 positional information below this resolution is lost.

 */
/*----------------------------------------------------------------------------*/
int
sph_utils_get_dither_id( const cpl_frame* inframe, float res,
                         const char* keyx,
                         const char* keyy )
{
    cpl_propertylist*   proplist        = NULL;
    const char*         filename        = NULL;
    long int            dit_id          = -1;
    double              dDitherx        = 0.0;
    double              dDithery        = 0.0;
    double              dID             = 0.0;

    filename        = cpl_frame_get_filename( inframe );
    proplist        = sph_keyword_manager_load_properties( filename, 0);

    dDitherx = cpl_propertylist_get_double( proplist, keyx );
    dDithery = cpl_propertylist_get_double( proplist, keyy );

    dID = ( dDitherx / res ) * ( dDithery / res );
    if ( dID > __INT_MAX__ ) {
        sph_error_raise( SPH_UTILS_TOO_FINE_DITHER,
                         __FILE__, __func__, __LINE__,
                         SPH_ERROR_ERROR,
                         "The dither resolution given "
                         "was too fine ! Decrease "
                         "it by at least a factor "
                         "of %lf.",
                         dID / (double)__INT_MAX__);
        return -1;
    }
    dit_id = (int)dID;
    cpl_propertylist_delete( proplist ); proplist = NULL;
    return dit_id;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Attach a propertylist to a frame
 *
 * @param plist        the propertylist to attach
 * @param frame        the frame to attach propertylist to
 * @param extension the extension to write propertylist to
 *
 * @return error code
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_utils_attach_plist( cpl_propertylist* plist, cpl_frame* frame, int ext ) {
    int            pcount        = 0;
    sph_error_code    rerr    = CPL_ERROR_NONE;

    if ( !plist || !frame || ext < 0 ) return CPL_ERROR_NULL_INPUT;

    for (pcount = 0; pcount < cpl_propertylist_get_size(plist); ++pcount) {
        rerr |= sph_fits_update_property( cpl_frame_get_filename( frame ), cpl_propertylist_get(plist, pcount), ext );
    }
    return rerr;
}


sph_error_code
sph_utils_cpl_image_fill_double( cpl_image * image, double value ){
    sph_error_code rerr     = CPL_ERROR_NONE;
    int            xx        = 0;
    int               yy        = 0;

    if ( !image ){
        SPH_NO_SELF
        return sph_error_get_last_code();
    }
    if ( cpl_image_get_type( image ) != CPL_TYPE_DOUBLE ){
        SPH_ERR("wrong CPL_TYPE")
        return sph_error_get_last_code();
    }

    for (xx = 0; xx < cpl_image_get_size_x( image ); ++xx) {
        for (yy = 0; yy < cpl_image_get_size_y( image ); ++yy) {
            cpl_image_set( image, xx + 1, yy + 1, value );
        }
    }

    if (cpl_error_get_code() != CPL_ERROR_NONE ){
        SPH_RAISE_CPL;
        rerr = sph_error_get_last_code();
    }

    return rerr;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief 	Remove all files from the given frames
 * @param	inframes	a pointer to the cpl_frameset
 * @return 	sphere error code
 *
 * Deletes all files on the disk from the given frames
 *
 */
/*----------------------------------------------------------------------------*/

sph_error_code sph_utils_delete_files( cpl_frameset* inframes )
{
	cpl_frame*		curframe	= NULL;
	cpl_ensure_code( inframes, CPL_ERROR_NULL_INPUT );

	SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;

	curframe = cpl_frameset_get_first( inframes );
	while ( curframe ){
		unlink( cpl_frame_get_filename( curframe ));
		curframe = cpl_frameset_get_next( inframes );
	}

	SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}


/*
 * A simple converting function from uppercase to lowcase
 */
void sph_utils_lowercase(char string[]){
    int i = 0;

       while ( string[i] )
       {
          string[i] = tolower(string[i]);
          i++;
       }

       return;
}

/*
 * A simple converting function from lowcase to uppercase
 */
void sph_utils_uppercase(char string[]){
    int i = 0;

       while ( string[i] )
       {
          string[i] = toupper(string[i]);
          i++;
       }

       return;
}


/* Return a newly allocated copy of STR,
   or 0 if out of memory. */
char *
sph_utils_strdup (str)
     const char *str;
{
  char *newstr;

  newstr = (char *) cpl_malloc (strlen (str) + 1);
  if (newstr)
    strcpy (newstr, str);
  return newstr;
}
