/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>
#include <float.h>

#include "sph_ifs_master_dark.h"
#include "sph_common_keywords.h"
#include <math.h>
#include "sph_error.h"
#include "sph_test.h"
#include "cutest_ifs_master_dark.h"
#include <Basic.h>
#include <CUnit.h>
#include "sph_test.h"
#include "cutest_ifs_master_dark.h"
#include "sph_utils.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup  cutest_ifs_master_dark  Unit test of ifs_master_dark recipe and \
 *                                       associated functions.
 *
 * This tests verifies the very basic requirements of the sph_ifs_master_dark
 * recipe. All functions of the recipe object are tested in only one,
 * standard, test scenario. In this scenario, the parameters for the
 * recipe are the standard default parameters. The input raw frames
 * are created using the sph_test_create_ifs_raw_dark_frameset function.
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/
int cutest_init_master_dark_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__, PACKAGE_BUGREPORT, CPL_MSG_DEBUG);
    return CPL_ERROR_NONE;
}

int cutest_clean_master_dark_testsuite(void) {
    sph_error_dump(SPH_ERROR_ERROR);
    return sph_end_test();
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Common setup for the unit tests
 *
 * @param    frames        Address of pointer to the input frames to set up
 * @param   params        Address of pointer to the input parameterlist to set up
 *
 * This sets up the standard recipe parameter list and the input frames.
 * The frames an params pointers should not be initiallised yet.
 *
 * On return the pointers will point to the corresponding cpl structures that
 * are now filled with the setup data for the test.
 *
 */
/*----------------------------------------------------------------------------*/
void cutest_ifs_master_dark_common_setup(cpl_frameset** frames,
        cpl_parameterlist** params, int nframes) {
    sph_error_code rerr = CPL_ERROR_NONE;
    cpl_error_reset();
    CU_ASSERT_PTR_NOT_NULL_FATAL( frames);
    CU_ASSERT_PTR_NOT_NULL_FATAL( params);

    *frames = cpl_frameset_new();
    rerr = sph_test_create_ifs_raw_dark_frameset(*frames, nframes);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE);

    *params = sph_ifs_master_dark_create_paramlist();

    rerr = cpl_error_get_code();
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE);
    CU_ASSERT_PTR_NOT_NULL_FATAL( *frames);
    CU_ASSERT_PTR_NOT_NULL_FATAL( *params);
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Common teardown for the unit tests
 *
 * @param    frames        pointer to the input frames to clean up
 * @param   params        pointer to the input parameterlist to clean up
 *
 * This cleans up the frameset and parameterlist
 */
/*----------------------------------------------------------------------------*/
void cutest_ifs_master_dark_common_teardown(cpl_frameset* frames,
        cpl_parameterlist* params) {

    CU_ASSERT_PTR_NOT_NULL_FATAL( frames);
    CU_ASSERT_PTR_NOT_NULL_FATAL( params);

    cpl_parameterlist_delete(params);
    cpl_frameset_delete(frames);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_ifs_master_dark check frames function.
 */
/*----------------------------------------------------------------------------*/
void cutest_ifs_master_dark_check_frames(void) {
    sph_ifs_master_dark* ifs_master_dark = NULL;
    cpl_parameterlist* inparams = NULL;
    cpl_frameset* inframes = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;

    /* Setup */
    cutest_ifs_master_dark_common_setup(&inframes, &inparams, 3);

    ifs_master_dark = sph_ifs_master_dark_new(inframes, inparams);
    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark);
    // Delete some stuff to avoid duplicates and memleaks.
    cpl_frameset_delete(ifs_master_dark->current_raw_frameset);
    cpl_frameset_delete(ifs_master_dark->rawframes);

    /* Running code... */
    rerr = sph_ifs_master_dark_check_frames(ifs_master_dark);

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark->inframes);

    CU_ASSERT_EQUAL( cpl_frameset_get_size( ifs_master_dark->inframes), 3);

    /* Terdown */
    sph_ifs_master_dark_delete(ifs_master_dark);
    cutest_ifs_master_dark_common_teardown(inframes, inparams);
    return;

}
void cutest_ifs_master_dark_check_params(void) {
    sph_ifs_master_dark* ifs_master_dark = NULL;
    cpl_parameterlist* inparams = NULL;
    cpl_frameset* inframes = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;

    /* Setup */
    cutest_ifs_master_dark_common_setup(&inframes, &inparams, 3);
    ifs_master_dark = sph_ifs_master_dark_new(inframes, inparams);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark);

    /* Running ... */

    rerr = sph_ifs_master_dark_check_params(ifs_master_dark);

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark->master_dark_outfilename);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark->static_badpix_filename);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark->framecomb_parameterlist);

    CU_ASSERT_STRING_EQUAL( ifs_master_dark->master_dark_outfilename,
            "master_dark.fits");

    CU_ASSERT_STRING_EQUAL( ifs_master_dark->static_badpix_filename,
            "static_badpixels.fits");

    CU_ASSERT_EQUAL(
            cpl_parameterlist_get_size( ifs_master_dark->framecomb_parameterlist),
            2);
    /* Terdown */
    cutest_ifs_master_dark_common_teardown(inframes, inparams);
    sph_ifs_master_dark_delete(ifs_master_dark);
    return;

}
void cutest_ifs_master_dark_new(void) {
    sph_ifs_master_dark* ifs_master_dark = NULL;
    cpl_parameterlist* inparams = NULL;
    cpl_frameset* inframes = NULL;

    cutest_ifs_master_dark_common_setup(&inframes, &inparams, 3);

    ifs_master_dark = sph_ifs_master_dark_new(inframes, inparams);

    CU_ASSERT_PTR_NOT_NULL( ifs_master_dark);
    /* Terdown */
    cutest_ifs_master_dark_common_teardown(inframes, inparams);
    sph_ifs_master_dark_delete(ifs_master_dark);
    return;
}
void cutest_ifs_master_dark_delete(void) {
    sph_ifs_master_dark* ifs_master_dark = NULL;
    cpl_parameterlist* inparams = NULL;
    cpl_frameset* inframes = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;

    /* Setup */
    cutest_ifs_master_dark_common_setup(&inframes, &inparams, 3);

    ifs_master_dark = sph_ifs_master_dark_new(inframes, inparams);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark);
    /* Run code... */
    rerr = sph_ifs_master_dark_delete(ifs_master_dark);
    /* Verify */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE);
    /* Terdown */
    cutest_ifs_master_dark_common_teardown(inframes, inparams);

    return;
}
void cutest_ifs_master_dark_run(void) {
    sph_ifs_master_dark* ifs_master_dark = NULL;
    cpl_parameterlist* inparams = NULL;
    cpl_frameset* inframes = NULL;
    cpl_image* bpixs = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;

    /* Setup */
    cutest_ifs_master_dark_common_setup(&inframes, &inparams, 3);

    ifs_master_dark = sph_ifs_master_dark_new(inframes, inparams);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark);

    /* Running */
    rerr |= sph_ifs_master_dark_run(ifs_master_dark);

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    bpixs = cpl_image_load(ifs_master_dark->master_dark_outfilename,
            CPL_TYPE_INT, 0, 1);
    CU_ASSERT_PTR_NOT_NULL_FATAL( bpixs);
    CU_ASSERT_TRUE( cpl_image_get_absflux( bpixs ) > 0);

    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE);
    cpl_image_delete(bpixs);
    bpixs = NULL;
    /* Terdown */
    cutest_ifs_master_dark_common_teardown(inframes, inparams);
    sph_ifs_master_dark_delete(ifs_master_dark);
    return;
}
void cutest_ifs_master_dark_returns_RON(void) {
    sph_ifs_master_dark* ifs_master_dark = NULL;
    cpl_parameterlist* inparams = NULL;
    cpl_frameset* inframes = NULL;
    cpl_image* bpixs = NULL;
    cpl_propertylist* pl = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;

    /* Setup */
    cutest_ifs_master_dark_common_setup(&inframes, &inparams, 3);

    ifs_master_dark = sph_ifs_master_dark_new(inframes, inparams);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark);

    /* Running */
    rerr |= sph_ifs_master_dark_run(ifs_master_dark);

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    bpixs = cpl_image_load(ifs_master_dark->master_dark_outfilename,
            CPL_TYPE_INT, 0, 1);
    CU_ASSERT_PTR_NOT_NULL_FATAL( bpixs);
    CU_ASSERT_TRUE( cpl_image_get_absflux( bpixs ) > 0);

    pl = cpl_propertylist_load(ifs_master_dark->master_dark_outfilename, 0);
    CU_ASSERT_TRUE( cpl_propertylist_has(pl,"ESO QC MEAN RON"));
    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE);
    cpl_image_delete(bpixs);
    bpixs = NULL;
    /* Teardown */
    cutest_ifs_master_dark_common_teardown(inframes, inparams);
    sph_ifs_master_dark_delete(ifs_master_dark);
    cpl_propertylist_delete(pl);
    return;
}

void cutest_ifs_master_dark_run_test_illegal_reject(void) {
    sph_ifs_master_dark* ifs_master_dark = NULL;
    cpl_parameterlist* inparams = NULL;
    cpl_frameset* inframes = NULL;
    cpl_parameter* p = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;

    /* Setup */
    cutest_ifs_master_dark_common_setup(&inframes, &inparams, 3);

    ifs_master_dark = sph_ifs_master_dark_new(inframes, inparams);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark);

    ifs_master_dark->clean_mean_reject_high = 2;
    ifs_master_dark->clean_mean_reject_low = 1;
    p = cpl_parameterlist_find(ifs_master_dark->framecomb_parameterlist,
            "clean_mean.reject_low");
    CU_ASSERT_PTR_NOT_NULL_FATAL( p);
    cpl_parameter_set_int(p, ifs_master_dark->clean_mean_reject_low);
    p = cpl_parameterlist_find(ifs_master_dark->framecomb_parameterlist,
            "clean_mean.reject_high");
    CU_ASSERT_PTR_NOT_NULL_FATAL( p);
    cpl_parameter_set_int(p, ifs_master_dark->clean_mean_reject_high);

    /* Running */
    rerr |= sph_ifs_master_dark_run(ifs_master_dark);

    /* Verification */
    CU_ASSERT_NOT_EQUAL( rerr, CPL_ERROR_NONE);
    sph_ifs_master_dark_delete(ifs_master_dark);

    /* Terdown */
    cutest_ifs_master_dark_common_teardown(inframes, inparams);
    return;
}

void cutest_ifs_master_dark_run_more_than50(void) {
    sph_ifs_master_dark* ifs_master_dark = NULL;
    cpl_parameterlist* inparams = NULL;
    cpl_frameset* inframes = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;

    /* Setup */
    inframes = cpl_frameset_new();
    rerr = sph_test_create_ifs_raw_dark_frameset_big(inframes, 50);
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE);

    inparams = sph_ifs_master_dark_create_paramlist();

    ifs_master_dark = sph_ifs_master_dark_new(inframes, inparams);

    CU_ASSERT_PTR_NOT_NULL_FATAL( ifs_master_dark);

    /* Running */
    rerr |= sph_ifs_master_dark_run(ifs_master_dark);

    /* Verification */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE);
    /* Terdown */
    cutest_ifs_master_dark_common_teardown(inframes, inparams);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit tests of techcal_master_dark recipe and associated functions
 */
/*----------------------------------------------------------------------------*/
int main(void) {
    int result = 0;

    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("master_dark_test", cutest_init_master_dark_testsuite,
            cutest_clean_master_dark_testsuite);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* add the tests to the suite */
    if (NULL
            == CU_add_test(pSuite, "sph_ifs_master_dark_new",
                    cutest_ifs_master_dark_new)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "sph_ifs_master_dark_check_frames",
                    cutest_ifs_master_dark_check_frames)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "sph_ifs_master_dark_check_params",
                    cutest_ifs_master_dark_check_params)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
//    if ( NULL == CU_add_test(pSuite, "sph_ifs_master_dark_run_more_than50",
//                             cutest_ifs_master_dark_run_more_than50 ) )
//    {
//       CU_cleanup_registry();
//       return CU_get_error();
//    }
    if (NULL
            == CU_add_test(pSuite, "sph_ifs_master_dark_returns_RON",
                    cutest_ifs_master_dark_returns_RON)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "sph_ifs_master_dark_run",
                    cutest_ifs_master_dark_run)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite,
                    "cutest_ifs_master_dark_run_test_illegal_reject",
                    cutest_ifs_master_dark_run_test_illegal_reject)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "sph_ifs_master_dark_delete",
                    cutest_ifs_master_dark_delete)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();
    sph_test_create_results_file("results.txt");
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
