/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <Basic.h>
#include <CUnit.h>
#include "sph_error.h"
#include "sph_test.h"
#include "sph_test_ngc_ir_simulator.h"
#include "sph_test_pupilimage_creator.h"
#include "sph_test_image_tools.h"
#include "sph_test_ird_fixtures.h"
#include "sph_ird_instrument_model.h"
#include "sph_ird_distortion_map.h"
#include "sph_ird_tags.h"
#include "sph_utils.h"

static
int cutest_sph_ird_distortion_map_init_testsuite(void);
static
int cutest_sph_ird_distortion_map_clean_testsuite(void);
static
void cutest_sph_ird_distortion_map_test_it(void);
static
void cutest_sph_ird_distortion_map_verify_result(void);
/*----------------------------------------------------------------------------*/
/**
 * @defgroup A CUnit Test Suite -- representing a collection of testcases
 * @par Synopsis:
 * @code
 * @endcode
 * @par Desciption:
 *
 * This module provides a collection of tests for one specific, distinct
 * module or set-up. The testing code is implemented using the CUnit
 * framework.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test MAIN function
 */
/*----------------------------------------------------------------------------*/
int main(void) {
    int result = 0;
    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("A cutest_sph_ird_distortion_map testsuite",
            cutest_sph_ird_distortion_map_init_testsuite,
            cutest_sph_ird_distortion_map_clean_testsuite);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* add the tests to the suite */
    if (NULL
            == CU_add_test(pSuite, "Testing distortion map run",
                    cutest_sph_ird_distortion_map_test_it)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
//    if ( NULL == CU_add_test(pSuite, "Testing distortion map verify",
//                             cutest_sph_ird_distortion_map_verify_result ) )
//    {
//       CU_cleanup_registry();
//       return CU_get_error();
//    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();
    sph_test_create_results_file("results.txt");
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_sph_ird_distortion_map_test_it(void) {
    cpl_frame* rawframe = NULL;
    cpl_frameset* rawframeset = NULL;
    sph_ird_instrument_model* model = NULL;
    cpl_parameterlist* parlist = NULL;
    sph_ird_distortion_map* recipe = NULL;
    cpl_frame* point_pattern_frame = NULL;
    cpl_propertylist* plist = NULL;

    rawframeset = cpl_frameset_new();
    model = sph_test_ird_fixtures_instrument_model(2048, 1024);
    CU_ASSERT_PTR_NOT_NULL_FATAL(model);
    rawframe = sph_test_ird_fixtures_create_dist_map_raw_frame(model, 63, 2.0,
            &point_pattern_frame);
    CU_ASSERT_PTR_NOT_NULL_FATAL(rawframe);
    cpl_frameset_insert(rawframeset, rawframe);
    cpl_frameset_insert(rawframeset, point_pattern_frame);

    parlist = sph_ird_distortion_map_create_paramlist();
    recipe = sph_ird_distortion_map_new(rawframeset, parlist);
    //recipe->polyfit_order = 2;
    //recipe->full_qc = 1;
    CU_ASSERT_EQUAL( sph_ird_distortion_map_run(recipe), CPL_ERROR_NONE);
    sph_ird_instrument_model_delete(model);
    sph_ird_distortion_map_delete(recipe);
    cpl_frameset_delete(rawframeset);
    cpl_parameterlist_delete(parlist);

    // Verifying coefficients are correct...
    plist = cpl_propertylist_load("distortion_map.fits", 0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(plist);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( plist, "ESO DRS DIST X COEFF 1_0"),
            -0.00765, 0.001);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( plist, "ESO DRS DIST Y COEFF 0_1"),
            -0.001658, 0.001);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( plist, "ESO DRS DIST X COEFF 0_0"),
            0.0, 0.02);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( plist, "ESO DRS DIST Y COEFF 0_0"),
            0.0, 0.02);
    CU_ASSERT_TRUE(cpl_propertylist_has(plist,"EXTNAME"));
    if (cpl_propertylist_has(plist, "EXTNAME")) {
        const char * cvalue = cpl_propertylist_get_string(plist,"EXTNAME");
        CU_ASSERT_EQUAL(cpl_error_get_code(), CPL_ERROR_NONE);
        CU_ASSERT_NSTRING_EQUAL(cvalue, "LEFT.DISTX.IMAGE", 10);
    }
    cpl_propertylist_delete(plist);
    plist = NULL;

    plist = cpl_propertylist_load("distortion_map.fits", 4);
    CU_ASSERT_PTR_NOT_NULL_FATAL(plist);
    CU_ASSERT_TRUE(cpl_propertylist_has(plist,"EXTNAME"));
    if (cpl_propertylist_has(plist, "EXTNAME")) {
        const char * cvalue = cpl_propertylist_get_string(plist,"EXTNAME");
        CU_ASSERT_EQUAL(cpl_error_get_code(), CPL_ERROR_NONE);
        CU_ASSERT_NSTRING_EQUAL(cvalue, "LEFT.DISTY.IMAGE", 10);
    }
    cpl_propertylist_delete(plist);

    plist = cpl_propertylist_load("distortion_map.fits", 8);
    CU_ASSERT_PTR_NOT_NULL_FATAL(plist);
    CU_ASSERT_TRUE(cpl_propertylist_has(plist,"EXTNAME"));
    if (cpl_propertylist_has(plist, "EXTNAME")) {
        const char * cvalue = cpl_propertylist_get_string(plist,"EXTNAME");
        CU_ASSERT_EQUAL(cpl_error_get_code(), CPL_ERROR_NONE);
        CU_ASSERT_NSTRING_EQUAL(cvalue, "RIGHT.DISTX.IMAGE", 10);
    }
    cpl_propertylist_delete(plist);

    return;
}

static
void cutest_sph_ird_distortion_map_verify_result(void) {
    cpl_frame* rawframe = NULL;
    cpl_frameset* rawframeset = NULL;
    sph_ird_instrument_model* model = NULL;
    cpl_parameterlist* parlist = NULL;
    sph_ird_distortion_map* recipe = NULL;
    cpl_frame* point_pattern_frame = NULL;
    cpl_frame* dummyframe = NULL;
    cpl_propertylist* plist = NULL;
    cpl_image* distimx = NULL;
    cpl_image* distimy = NULL;

    model = sph_test_ird_fixtures_instrument_model(2048, 1024);
    CU_ASSERT_PTR_NOT_NULL_FATAL(model);
    // Just need the point pattern here
    dummyframe = sph_test_ird_fixtures_create_dist_map_raw_frame(model, 63, 2.0,
            &point_pattern_frame);
    cpl_frame_delete(dummyframe);
    dummyframe = NULL;

    rawframeset = cpl_frameset_new();
    rawframe = cpl_frame_new();
    cpl_frame_set_filename(rawframe, "qc_image.fits");
    cpl_frame_set_group(rawframe, CPL_FRAME_GROUP_RAW);
    cpl_frame_set_tag(rawframe, SPH_IRD_TAG_DISTORTION_MAP_RAW);
    CU_ASSERT_PTR_NOT_NULL_FATAL(rawframe);
    cpl_frameset_insert(rawframeset, rawframe);
    cpl_frameset_insert(rawframeset, point_pattern_frame);

    parlist = sph_ird_distortion_map_create_paramlist();
    recipe = sph_ird_distortion_map_new(rawframeset, parlist);
    recipe->polyfit_order = 2;
    CU_ASSERT_EQUAL( sph_ird_distortion_map_run(recipe), CPL_ERROR_NONE);

    plist = cpl_propertylist_load("distortion_map.fits", 0);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( plist, "ESO DRS DIST X COEFF 0_0"),
            0.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( plist, "ESO DRS DIST X COEFF 1_0"),
            0.0, 0.01);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( plist, "ESO DRS DIST X COEFF 0_1"),
            0.0, 0.01);
    distimx = cpl_image_load("distortion_map.fits", CPL_TYPE_DOUBLE, 0, 0);
    distimy = cpl_image_load("distortion_map.fits", CPL_TYPE_DOUBLE, 0, 4);
    CU_ASSERT_PTR_NOT_NULL_FATAL( distimx);
    CU_ASSERT_PTR_NOT_NULL_FATAL( distimy);

    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_min(distimx), 0.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_max(distimx), 0.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_min(distimy), 0.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_min(distimy), 0.0, 0.5);

    cpl_image_delete(distimx);
    distimx = NULL;
    cpl_image_delete(distimy);
    distimy = NULL;

    distimx = cpl_image_load("distortion_map.fits", CPL_TYPE_DOUBLE, 0, 8);
    distimy = cpl_image_load("distortion_map.fits", CPL_TYPE_DOUBLE, 0, 12);
    CU_ASSERT_PTR_NOT_NULL_FATAL( distimx);
    CU_ASSERT_PTR_NOT_NULL_FATAL( distimy);

    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_min(distimx), 0.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_max(distimx), 0.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_min(distimy), 0.0, 0.5);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_min(distimy), 0.0, 0.5);

    cpl_image_delete(distimx);
    distimx = NULL;
    cpl_image_delete(distimy);
    distimy = NULL;

    cpl_propertylist_delete(plist);
    plist = NULL;

    sph_ird_instrument_model_delete(model);
    sph_ird_distortion_map_delete(recipe);
    cpl_frameset_delete(rawframeset);
    cpl_parameterlist_delete(parlist);
    return;
}

/*----------------------------------------------------------------------------*/
/*-                            INTERNAL HELPER FUNCTIONS                          */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to initiailise the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_sph_ird_distortion_map_init_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__, PACKAGE_BUGREPORT, CPL_MSG_DEBUG);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to clean the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_sph_ird_distortion_map_clean_testsuite(void) {
    sph_error_dump(SPH_ERROR_ERROR);
    return sph_end_test();
}
/**@}*/
