/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>

#include "sph_cube.h"
#include "sph_common_keywords.h"
#include "math.h"
#include "sph_error.h"
#include "sph_test.h"
#include <Basic.h>
#include <CUnit.h>
#include "sph_utils.h"
#include "sph_test.h"
#include "sph_fits.h"
static
void __cutest_sph_cube_finalise_file_setup(cpl_frameset* outframes);
/*----------------------------------------------------------------------------*/
/**
 * @defgroup cutest_sph_cube Unit test for the sph_cube module
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

int cutest_init_sph_cube_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__, PACKAGE_BUGREPORT, CPL_MSG_DEBUG);
    return 0;
}

int cutest_clean_sph_cube_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Close and dump errors.
     * -------------------------------------------------------------------*/
    sph_error_dump(SPH_ERROR_ERROR);
    return sph_end_test();
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_new(void) {
    sph_cube* acube = NULL;
    const char* fname = "testcube.fits";
    FILE* fp = NULL;

    unlink(fname);
    CU_ASSERT_PTR_NULL( sph_cube_new(NULL));
    cpl_error_reset();

    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);
    acube = sph_cube_new(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(acube);

    CU_ASSERT_PTR_NOT_NULL_FATAL(acube->filename);
    CU_ASSERT_STRING_EQUAL(acube->filename, fname);
    CU_ASSERT_EQUAL(acube->file_exists, 0);
    CU_ASSERT_EQUAL(acube->nx, 0);
    CU_ASSERT_EQUAL(acube->ny, 0);
    CU_ASSERT_EQUAL(acube->type, CPL_TYPE_UNSPECIFIED);
    CU_ASSERT_PTR_NOT_NULL(acube->proplist);

    // No file should be created yet...
    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NULL(fp);
    sph_cube_delete(acube);
    acube = NULL;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_new_dfs(void) {
    sph_cube* acube = NULL;
    const char* fname = "testcube.fits";
    FILE* fp = NULL;
    cpl_frameset* frameset = cpl_frameset_new();
    cpl_frame* aframe = cpl_frame_new();
    cpl_parameterlist* params = cpl_parameterlist_new();

    unlink(fname);
    CU_ASSERT_PTR_NULL(
            sph_cube_new_dfs(NULL, frameset,aframe,params,"ATAG","ARECIPE", "IRDIS", NULL));
    cpl_error_reset();
    CU_ASSERT_PTR_NULL(
            sph_cube_new_dfs(fname, frameset,aframe,params,NULL,"ARECIPE", "IRDIS", NULL));
    cpl_error_reset();

    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);
    acube = sph_cube_new_dfs(fname, frameset, aframe, params, "ATAG", "ARECIPE",
            "IRDIS", NULL);
    CU_ASSERT_PTR_NOT_NULL_FATAL(acube);

    CU_ASSERT_PTR_NOT_NULL_FATAL(acube->filename);
    CU_ASSERT_STRING_EQUAL(acube->filename, fname);
    CU_ASSERT_EQUAL(acube->file_exists, 0);
    CU_ASSERT_EQUAL(acube->nx, 0);
    CU_ASSERT_EQUAL(acube->ny, 0);
    CU_ASSERT_EQUAL(acube->type, CPL_TYPE_UNSPECIFIED);
    CU_ASSERT_PTR_NOT_NULL(acube->proplist);
    CU_ASSERT_PTR_EQUAL(acube->dfs_allframes, frameset);
    CU_ASSERT_PTR_EQUAL(acube->dfs_template_frame, aframe);
    CU_ASSERT_PTR_EQUAL(acube->dfs_paramlist, params);
    CU_ASSERT_STRING_EQUAL(acube->dfs_tag, "ATAG");
    CU_ASSERT_STRING_EQUAL(acube->dfs_recipe, "ARECIPE");
    CU_ASSERT_STRING_EQUAL(acube->dfs_pipename, "IRDIS");
    // No file should be created yet...
    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NULL(fp);
    sph_cube_delete(acube);
    acube = NULL;
    cpl_frameset_delete(frameset);
    frameset = NULL;
    cpl_frame_delete(aframe);
    aframe = NULL;
    cpl_parameterlist_delete(params);
    params = NULL;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_append_master_frame_newcube(void) {
    sph_cube* acube = NULL;
    sph_master_frame* mframe = NULL;
    const char* fname = "testcube2.fits";
    FILE* fp = NULL;

    unlink(fname);
    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);
    acube = sph_cube_new(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(acube);

    mframe = sph_master_frame_new(256, 256);
    CU_ASSERT_EQUAL( sph_cube_append_master(acube,mframe,1), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_cube_finalise_file(acube->filename), CPL_ERROR_NONE);

    CU_ASSERT_EQUAL(acube->file_exists, 1);
    CU_ASSERT_EQUAL(acube->nx, 256);
    CU_ASSERT_EQUAL(acube->ny, 256);

    // file should have been created...
    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fp);
    fclose(fp);
    sph_cube_delete(acube);
    acube = NULL;
    sph_master_frame_delete(mframe);
    mframe = NULL;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_append_master_frame(void) {
    sph_cube* acube = NULL;
    sph_master_frame* mframe = NULL;
    const char* fname = "testcube2.fits";
    FILE* fp = NULL;

    unlink(fname);
    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);
    acube = sph_cube_new(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(acube);

    mframe = sph_master_frame_new(256, 256);
    CU_ASSERT_EQUAL( sph_cube_append_master(acube,mframe,1), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_cube_finalise_file(acube->filename), CPL_ERROR_NONE);

    CU_ASSERT_EQUAL(acube->file_exists, 1);
    CU_ASSERT_EQUAL(acube->nx, 256);
    CU_ASSERT_EQUAL(acube->ny, 256);

    // file should have been created...
    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fp);
    fclose(fp);

    CU_ASSERT_EQUAL( sph_cube_append_master(acube,mframe,1), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_cube_finalise_file(acube->filename), CPL_ERROR_NONE);

    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname,0), 2);

    sph_cube_delete(acube);
    acube = NULL;
    sph_master_frame_delete(mframe);
    mframe = NULL;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_append_image(void) {
    cpl_image* image = NULL;
    const char* fname = "testcube_mext.fits";
    const char* fname0 = "testcube_mext_e00.fits";
    FILE* fp = NULL;
    cpl_propertylist* pl = NULL;

    unlink(fname);
    pl = cpl_propertylist_new();
    cpl_propertylist_update_int(pl, "MYKEY", 1000);
    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);

    image = cpl_image_new(256, 256, CPL_TYPE_DOUBLE);
    CU_ASSERT_EQUAL( sph_cube_append_image(fname,image,pl,0), CPL_ERROR_NONE);

    // file should have been created...
    fp = fopen(fname0, "r");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fp);
    fclose(fp);

    cpl_propertylist_update_int(pl, "MYKEY2", 1000);
    CU_ASSERT_EQUAL( sph_cube_append_image(fname,image,pl,0), CPL_ERROR_NONE);

    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname0,0), 2);

    cpl_propertylist_delete(pl);
    pl = NULL;

    pl = cpl_propertylist_load(fname0, 0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(pl);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY"), 1000);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY2"), 1000);

    cpl_propertylist_delete(pl);
    pl = NULL;
    cpl_image_delete(image);
    image = NULL;
    unlink(fname0);
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_append_image_props_ok(void) {
    cpl_image* image = NULL;
    const char* fname = "testcube_mext.fits";
    const char* fname0 = "testcube_mext_e00.fits";
    FILE* fp = NULL;
    cpl_propertylist* pl = NULL;

    unlink(fname);
    pl = cpl_propertylist_new();
    cpl_propertylist_update_int(pl, "MYKEY", 1000);
    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);

    image = cpl_image_new(256, 256, CPL_TYPE_DOUBLE);
    CU_ASSERT_EQUAL( sph_cube_append_image(fname,image,pl,0), CPL_ERROR_NONE);

    // file should have been created...
    fp = fopen(fname0, "r");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fp);
    fclose(fp);

    cpl_propertylist_update_int(pl, "MYKEY2", 1000);
    CU_ASSERT_EQUAL( sph_cube_append_image(fname,image,pl,0), CPL_ERROR_NONE);

    CU_ASSERT_EQUAL( sph_cube_append_image(fname,image,NULL,1), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname0,0), 2);

    CU_ASSERT_EQUAL( sph_cube_append_image(fname,image,NULL,0), CPL_ERROR_NONE);

    cpl_propertylist_delete(pl);
    pl = NULL;

    pl = cpl_propertylist_load(fname0, 0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(pl);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY"), 1000);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY2"), 1000);

    cpl_propertylist_delete(pl);
    pl = NULL;
    sph_cube_finalise_file(fname);
    pl = cpl_propertylist_load(fname, 0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(pl);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY"), 1000);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY2"), 1000);
    cpl_propertylist_delete(pl);
    pl = NULL;

    cpl_image_delete(image);
    image = NULL;
    unlink(fname0);
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_finalise_file_nothing_todo(void) {
    const char* fname = "testcube_mext.fits";
    const char* fname0 = "testcube_mext_e00.fits";
    const char* fname1 = "testcube_mext_e01.fits";
    FILE* fp = NULL;

    unlink(fname);
    // file should not exist...
    fp = fopen(fname0, "r");
    CU_ASSERT_PTR_NULL_FATAL(fp);
    fp = fopen(fname1, "r");
    CU_ASSERT_PTR_NULL_FATAL(fp);

    CU_ASSERT_EQUAL( sph_cube_finalise_file(fname), CPL_ERROR_NONE);

    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NULL_FATAL(fp);
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_finalise_file(void) {
    const char* fname = "testcube_mext.fits";
    const char* fname0 = "testcube_mext_e00.fits";
    const char* fname1 = "testcube_mext_e01.fits";
    FILE* fp = NULL;
    cpl_imagelist* imlist = NULL;
    cpl_propertylist* pl = NULL;

    unlink(fname);
    __cutest_sph_cube_finalise_file_setup(NULL);
    CU_ASSERT_EQUAL( sph_cube_finalise_file(fname), CPL_ERROR_NONE);

    // file should have been created...
    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fp);
    fclose(fp);

    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname,0), 5);

    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname,1), 5);

    cpl_propertylist_delete(pl);
    pl = NULL;

    pl = cpl_propertylist_load(fname, 0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(pl);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY"), 1000);
    CU_ASSERT_EQUAL(cpl_propertylist_has(pl,"MYKEY2"), 0);

    cpl_propertylist_delete(pl);
    pl = NULL;
    pl = cpl_propertylist_load(fname, 1);
    CU_ASSERT_PTR_NOT_NULL_FATAL(pl);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY"), 1000);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY2"), 1000);

    cpl_propertylist_delete(pl);
    pl = NULL;

    cpl_imagelist_delete(imlist);
    imlist = NULL;
    unlink(fname0);
    unlink(fname1);
    // Finalising a file twice should give no error.
    CU_ASSERT_EQUAL( sph_cube_finalise_file(fname), CPL_ERROR_NONE);
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_finalise_all(void) {
    const char* fname = "testcube_mext.fits";
    const char* fname0 = "testcube_mext_e00.fits";
    const char* fname1 = "testcube_mext_e01.fits";
    FILE* fp = NULL;
    cpl_imagelist* imlist = NULL;
    cpl_propertylist* pl = NULL;
    cpl_frameset* frames = NULL;
    cpl_frame* aframe = NULL;
    frames = cpl_frameset_new();
    aframe = cpl_frame_new();
    cpl_frame_set_tag(aframe, "NONE");
    cpl_frame_set_group(aframe, CPL_FRAME_GROUP_CALIB);
    cpl_frame_set_filename(aframe, fname);
    cpl_frameset_insert(frames, aframe);

    unlink(fname);
    __cutest_sph_cube_finalise_file_setup(NULL);
    // making sure there is no cube before test
    unlink(fname);
    CU_ASSERT_EQUAL( sph_cube_finalise_all(frames), CPL_ERROR_NONE);

    // file should have been created...
    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fp);
    fclose(fp);

    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname,0), 5);

    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname,1), 5);

    cpl_propertylist_delete(pl);
    pl = NULL;

    pl = cpl_propertylist_load(fname, 0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(pl);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY"), 1000);
    CU_ASSERT_EQUAL(cpl_propertylist_has(pl,"MYKEY2"), 0);

    cpl_propertylist_delete(pl);
    pl = NULL;
    pl = cpl_propertylist_load(fname, 1);
    CU_ASSERT_PTR_NOT_NULL_FATAL(pl);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY"), 1000);
    CU_ASSERT_EQUAL(cpl_propertylist_get_int(pl,"MYKEY2"), 1000);

    cpl_propertylist_delete(pl);
    pl = NULL;

    cpl_imagelist_delete(imlist);
    imlist = NULL;
    unlink(fname0);
    unlink(fname1);
    cpl_frameset_delete(frames);
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_append_double_image_newcube(void) {
    sph_cube* acube = NULL;
    sph_double_image* dimage = NULL;
    const char* fname = "testcube2.fits";
    FILE* fp = NULL;

    unlink(fname);

    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);
    acube = sph_cube_new(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(acube);

    dimage = sph_double_image_new(256, 256);
    CU_ASSERT_EQUAL(cpl_error_get_code(), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_cube_append_double_image(acube,dimage,1),
            CPL_ERROR_NONE);
    CU_ASSERT_EQUAL(cpl_error_get_code(), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_cube_finalise_file(acube->filename), CPL_ERROR_NONE);

    CU_ASSERT_EQUAL(acube->file_exists, 1);
    CU_ASSERT_EQUAL(acube->nx, 256);
    CU_ASSERT_EQUAL(acube->ny, 256);

    // file should have been created...
    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fp);
    fclose(fp);

    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname,0), 1);

    CU_ASSERT_EQUAL(cpl_error_get_code(), CPL_ERROR_NONE);
    sph_cube_delete(acube);
    acube = NULL;
    sph_double_image_delete(dimage);
    dimage = NULL;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_cube_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_cube_append_double_image(void) {
    sph_cube* acube = NULL;
    sph_double_image* dimage = NULL;
    const char* fname = "testcube2.fits";
    FILE* fp = NULL;

    unlink(fname);
    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);
    acube = sph_cube_new(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(acube);

    dimage = sph_double_image_new(256, 256);
    CU_ASSERT_EQUAL(cpl_error_get_code(), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_cube_append_double_image(acube,dimage,1),
            CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_cube_finalise_file(acube->filename), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL(cpl_error_get_code(), CPL_ERROR_NONE);
    CU_ASSERT_EQUAL(acube->file_exists, 1);
    CU_ASSERT_EQUAL(acube->nx, 256);
    CU_ASSERT_EQUAL(acube->ny, 256);

    // file should have been created...
    fp = fopen(fname, "r");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fp);
    fclose(fp);

    CU_ASSERT_EQUAL( sph_cube_append_double_image(acube,dimage,1),
            CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( sph_cube_finalise_file(acube->filename), CPL_ERROR_NONE);

    CU_ASSERT_EQUAL( sph_fits_get_nplanes(fname,0), 2);

    sph_cube_delete(acube);
    acube = NULL;
    sph_double_image_delete(dimage);
    dimage = NULL;
}

/*----------------------------------------------------------------------------*/
/**
 @brief   Main to run the tests.
 */
/*----------------------------------------------------------------------------*/
int main(void) {
    int result = 0;
    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("cutest_sph_cube", cutest_init_sph_cube_testsuite,
            cutest_clean_sph_cube_testsuite);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL == CU_add_test(pSuite, "Testing cube_new", cutest_sph_cube_new)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "Testing cube_new_dfs",
                    cutest_sph_cube_new_dfs)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "Testing cube_append_master_frame_newcube",
                    cutest_sph_cube_append_master_frame_newcube)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "Testing cube_append_master_frame",
                    cutest_sph_cube_append_master_frame)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "Testing cube_append_double_image_newcube",
                    cutest_sph_cube_append_double_image_newcube)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "Testing cube_append_double_image_frame",
                    cutest_sph_cube_append_double_image)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "Testing cube_append_image_frame",
                    cutest_sph_cube_append_image)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "Testing cube_append_image_frame_props_ok",
                    cutest_sph_cube_append_image_props_ok)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "Testing cube_finalise_file",
                    cutest_sph_cube_finalise_file)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "Testing cube_finalise_file_nothing",
                    cutest_sph_cube_finalise_file_nothing_todo)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "Testing cube_finalise_all",
                    cutest_sph_cube_finalise_all)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();
    sph_test_create_results_file("results.txt");
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}
static
void __cutest_sph_cube_finalise_file_setup(cpl_frameset* outframes) {
    cpl_image* image = NULL;
    const char* fname = "testcube_mext.fits";
    const char* fname0 = "testcube_mext_e00.fits";
    const char* fname1 = "testcube_mext_e01.fits";
    cpl_imagelist* imlist = NULL;
    cpl_propertylist* pl = NULL;
    int np = 5;
    int pp = 0;
    cpl_frame* aframe = NULL;
    pl = cpl_propertylist_new();
    cpl_propertylist_update_int(pl, "MYKEY", 1000);
    sph_filemanager_add_tmp_file(fname);
    CU_ASSERT_PTR_NOT_NULL_FATAL(fname);

    imlist = cpl_imagelist_new();
    for (pp = 0; pp < np; ++pp) {
        image = cpl_image_new(256, 256, CPL_TYPE_DOUBLE);
        cpl_image_add_scalar(image, pp * 1.1);
        cpl_imagelist_set(imlist, image, pp);
    }
    cpl_imagelist_save(imlist, fname0, CPL_TYPE_UNSPECIFIED, pl, CPL_IO_CREATE);
    cpl_imagelist_delete(imlist);
    imlist = NULL;

    cpl_propertylist_update_int(pl, "MYKEY2", 1000);
    imlist = cpl_imagelist_new();
    for (pp = 0; pp < np; ++pp) {
        image = cpl_image_new(256, 256, CPL_TYPE_DOUBLE);
        cpl_image_add_scalar(image, pp * 1.1);
        cpl_imagelist_set(imlist, image, pp);
    }
    cpl_imagelist_save(imlist, fname1, CPL_TYPE_UNSPECIFIED, pl, CPL_IO_CREATE);
    cpl_imagelist_delete(imlist);
    imlist = NULL;

    if (outframes) {
        aframe = cpl_frame_new();
        cpl_frame_set_tag(aframe, "NONE");
        cpl_frame_set_group(aframe, CPL_FRAME_GROUP_CALIB);
        cpl_frame_set_filename(aframe, fname0);
        cpl_frameset_insert(outframes, aframe);
        aframe = cpl_frame_new();
        cpl_frame_set_tag(aframe, "NONE");
        cpl_frame_set_group(aframe, CPL_FRAME_GROUP_CALIB);
        cpl_frame_set_filename(aframe, fname1);
        cpl_frameset_insert(outframes, aframe);
    }
    cpl_propertylist_delete(pl);
    pl = NULL;
}

/**@}*/
