/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <Basic.h>
#include <CUnit.h>
#include <cpl.h>
#include <cpl_image_io.h>
#include "sph_error.h"
#include "sph_test.h"
#include "sph_ird_instrument_model.h"
#include "sph_test_ngc_ir_simulator.h"
#include "sph_test_pupilimage_creator.h"
#include "sph_test_image_tools.h"
#include "sph_utils.h"
#include "sph_fits.h"
#include "sph_time.h"
#include "sph_skycalc.h"
#include "sph_coord.h"
#include "sph_keyword_manager.h"
#include "sph_common_keywords.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup A CUnit Test Suite -- representing a collection of testcases
 * @par Synopsis:
 * @code
 * @endcode
 * @par Desciption:
 *
 * This module provides a collection of tests for one specific, distinct
 * module or set-up. The testing code is implemented using the CUnit
 * framework.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

static
void cutest_sph_extract_angles_test_bad_input(void);
static
void cutest_sph_extract_angles_test_irdis(void);
static
int cutest_init_testsuite(void);
static
int cutest_clean_testsuite(void);

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test MAIN function
 */
/*----------------------------------------------------------------------------*/
int main(void) {
    int result = 0;
    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("Exctract angles functionality testsuite",
            cutest_init_testsuite, cutest_clean_testsuite);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* add the tests to the suite */

    if (NULL
            == CU_add_test(pSuite, "Testing for null input errors",
                    cutest_sph_extract_angles_test_bad_input)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite, "Testing IRDIS angles correctly extracted",
                    cutest_sph_extract_angles_test_irdis)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();
    sph_test_create_results_file("results.txt");
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/*----------------------------------------------------------------------------*/
/*-                            TESTS                                          */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_sph_extract_angles_test_bad_input(void) {
    cpl_image* testimage = NULL;
    cpl_frame* input_frame = NULL;
    cpl_vector* angles = NULL;

    testimage = sph_test_image_tools_create_grid_image(256, 256, 10, 10,
            CPL_TYPE_INT, 10000, 2.0);

    CU_ASSERT_PTR_NOT_NULL_FATAL(testimage);

    input_frame = sph_test_cubify_image(testimage, "TEST_CUBE_TAG",
            CPL_FRAME_GROUP_RAW);

    CU_ASSERT_EQUAL(sph_extract_angles_from_cube( NULL, 0.0, 0.0, &angles),
            CPL_ERROR_NULL_INPUT);
    CU_ASSERT_EQUAL(sph_extract_angles_from_cube( input_frame, 0.0, 0.0, NULL),
            CPL_ERROR_NULL_INPUT);

    cpl_image_delete(testimage);
    testimage = NULL;
    cpl_frame_delete(input_frame);
    input_frame = NULL;
    cpl_vector_delete(angles);
    angles = NULL;

    cpl_error_reset();
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_sph_extract_angles_test_irdis(void) {
    cpl_image* testimage = NULL;
    cpl_frame* input_frame = NULL;
    cpl_vector* angles = NULL;
    double mjd = 0.0;

    testimage = sph_test_image_tools_create_grid_image(256, 256, 10, 10,
            CPL_TYPE_INT, 10000, 2.0);

    CU_ASSERT_PTR_NOT_NULL_FATAL(testimage);

    input_frame = sph_test_cubify_image(testimage, "TEST_CUBE_TAG",
            CPL_FRAME_GROUP_RAW);

    sph_test_frame_add_datestamp(input_frame);

    sph_test_frame_add_key_string(input_frame, 0,
            SPH_COMMON_KEYWORD_DET_FRAM_UTC, "2000-01-05T13:00:00");

    sph_time_mjd_from_string(&mjd, "2000-01-05T12:00:00");

    sph_test_frame_add_key_double(input_frame, 0, SPH_COMMON_KEYWORD_MJD_OBS,
            mjd);

    sph_test_frame_add_key_int(input_frame, 0, SPH_COMMON_KEYWORD_EPOCH, 2000);
    sph_test_frame_add_key_double(input_frame, 0, SPH_COMMON_KEYWORD_DEC, 0.0);
    sph_test_frame_add_key_double(input_frame, 0, SPH_COMMON_KEYWORD_RA, 0.0);

    CU_ASSERT_EQUAL(
            sph_extract_angles_from_cube( input_frame, SPH_IRD_INSTRUMENT_MODEL_DEROT_OFFSETELEV,0.0, &angles),
            CPL_ERROR_NONE);

    CU_ASSERT_PTR_NOT_NULL_FATAL( angles);

    CU_ASSERT_EQUAL(cpl_vector_get_size(angles), 1);

    // TODO: insert check for actual values once we have verified with some real data

    cpl_image_delete(testimage);
    testimage = NULL;
    cpl_frame_delete(input_frame);
    input_frame = NULL;
    cpl_vector_delete(angles);
    angles = NULL;

    cpl_error_reset();
    return;
}
/*----------------------------------------------------------------------------*/
/*-                            INTERNAL HELPER FUNCTIONS                          */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to initiailise the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_init_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__, PACKAGE_BUGREPORT, CPL_MSG_DEBUG);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to clean the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_clean_testsuite(void) {
    sph_error_dump(SPH_ERROR_ERROR);
    sph_end_test();
    return 0;
}

/**@}*/
