/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include "sph_strehl.h"

#include <assert.h>
#include <string.h>
#include <stdio.h>

/*-----------------------------------------------------------------------------
 Define
 -----------------------------------------------------------------------------*/

#ifndef SPH_STREHL_TABLE_IRDIS
#define SPH_STREHL_TABLE_IRDIS "spher-calib/ird/sph_ird_filt_table.fits"
#endif

#ifndef SPH_STREHL_TABLE_ZIMPOL
#define SPH_STREHL_TABLE_ZIMPOL "spher-calib/zpl/sph_zpl_filt_table.fits"
#endif

#define SPH_STREHL_RAW "sph_strehl.fits"

#ifndef SPH_STREHL_NX
/* A pretty much random, even number */
#define SPH_STREHL_NX 798
#endif

#ifndef SPH_STREHL_NY
/* A pretty much random number */
#define SPH_STREHL_NY 877
#endif

/* For the purpose of the test, using ten times bigger pixel scales, so can
 * use smaller images (otherwise the test runs for ages on some machines)
 */
#ifndef CUTEST_SPH_STREHL_PSCALE_IRDIS
#define CUTEST_SPH_STREHL_PSCALE_IRDIS 10.0 * SPH_STREHL_PSCALE_IRDIS
#endif

#ifndef CUTEST_SPH_STREHL_PSCALE_ZIMPOL
#define CUTEST_SPH_STREHL_PSCALE_ZIMPOL 10.0 * SPH_STREHL_PSCALE_ZIMPOL
#endif

/*-----------------------------------------------------------------------------
 Function declarations
 -----------------------------------------------------------------------------*/

static void sph_strehl_test_one(const char *, const char *, int);
static void sph_strehl_fill_header(cpl_propertylist *, int);
static void sph_strehl_save_raw(cpl_frame *, const cpl_image *, const char *,
        int);

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test MAIN function
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {

    char * base;
    char *end;
    int ndir = 5; /* Depth of source directory */

    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);

    assert(argc > 0);

    base = cpl_strdup(argv[0]);

    while ((end = strrchr(base, '/')) > base && --ndir >= 0) {
        *end = 0;
    }
    if (end != NULL) {
        end[1] = 0; /* Truncate after '/' */

        sph_strehl_test_one(base, SPH_STREHL_TABLE_IRDIS, 'I');
        sph_strehl_test_one(base, SPH_STREHL_TABLE_ZIMPOL, 'Z');
    }

    cpl_free(base);

    return cpl_test_end(0);
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
 @internal
 @brief Test one arm
 @param path   Initial path to filter frame
 @param base   Filter frame (with remainder of path)
 @param arm    Arm indicator ('I'/'Z')
 @return   void

 */
/*----------------------------------------------------------------------------*/
static void sph_strehl_test_one(const char * path, const char * base, int arm) {

    char * file = cpl_sprintf("%s%s", path, base);
    FILE * test = fopen(file, "r");

    if (test != NULL) {
        cpl_frame * filterframe = cpl_frame_new();
        cpl_frame * rawframe = cpl_frame_new();
        cpl_propertylist * qclist = cpl_propertylist_new();
        cpl_parameterlist * parlist = cpl_parameterlist_new();
        cpl_image * imtmp = cpl_image_new(SPH_STREHL_NX, SPH_STREHL_NY,
                CPL_TYPE_FLOAT);
        cpl_image * image = cpl_image_new(SPH_STREHL_NX, SPH_STREHL_NY,
                CPL_TYPE_FLOAT);
        FILE * stream =
                cpl_msg_get_level() > CPL_MSG_INFO ?
                        fopen("/dev/null", "a") : stdout;
        cpl_error_code code;

        cpl_msg_info(cpl_func, "Testing %s", file);
        cpl_test_zero(fclose(test));
        cpl_test_nonnull(stream);

        code = cpl_frame_set_filename(filterframe, file);
        cpl_test_eq_error(code, CPL_ERROR_NONE);

        code = cpl_frame_set_group(filterframe, CPL_FRAME_GROUP_CALIB);
        cpl_test_eq_error(code, CPL_ERROR_NONE);

        code = cpl_image_fill_noise_uniform(image, 0.0, 1.0);
        cpl_test_eq_error(code, CPL_ERROR_NONE);

        if (arm == 'I') {
            /* FIXME: Use pixel scale (2" = 163pixel) */
            const double sig = 2.2;

            if (cpl_msg_get_level() <= CPL_MSG_DEBUG) {
                /* Fail on pure noise */
                sph_strehl_save_raw(rawframe, image, SPH_STREHL_RAW, arm);

                code = sph_strehl_irdis(qclist, image, image, rawframe,
                        filterframe, "test", parlist);
                cpl_test_eq_error(code, CPL_ERROR_ILLEGAL_OUTPUT);
            }

            /* FIXME: Use something better than a Gaussian */
            code = cpl_image_fill_gaussian(imtmp, SPH_STREHL_NX / 4,
                    SPH_STREHL_NY / 2, 1000.0, sig, sig);
            cpl_test_eq_error(code, CPL_ERROR_NONE);

            code = cpl_image_add(image, imtmp);
            cpl_test_eq_error(code, CPL_ERROR_NONE);

            code = cpl_image_fill_gaussian(imtmp, 3 * SPH_STREHL_NX / 4,
                    SPH_STREHL_NY / 2, 10000.0, sig, sig);
            cpl_test_eq_error(code, CPL_ERROR_NONE);

            code = cpl_image_add(image, imtmp);
            cpl_test_eq_error(code, CPL_ERROR_NONE);

            sph_strehl_save_raw(rawframe, image, SPH_STREHL_RAW, arm);

            code = sph_strehl_irdis(qclist, image, image, rawframe, filterframe,
                    "test", parlist);
            cpl_test_eq_error(code, CPL_ERROR_NONE);

        } else if (arm == 'Z') {
            const double sig = 2.2; /* FIXME: Use pixel scale */

            if (cpl_msg_get_level() <= CPL_MSG_DEBUG) {
                /* Fail on pure noise */
                sph_strehl_save_raw(rawframe, image, SPH_STREHL_RAW, arm);

                code = sph_strehl_zimpol(qclist, image, image, rawframe,
                        filterframe, "test", parlist);
                cpl_test_eq_error(code, CPL_ERROR_ILLEGAL_OUTPUT);
            }

            /* FIXME: Use something better than a Gaussian */
            code = cpl_image_fill_gaussian(imtmp, SPH_STREHL_NX / 2,
                    SPH_STREHL_NY / 2, 10000.0, sig, sig);
            cpl_test_eq_error(code, CPL_ERROR_NONE);

            code = cpl_image_add(image, imtmp);
            cpl_test_eq_error(code, CPL_ERROR_NONE);

            sph_strehl_save_raw(rawframe, image, SPH_STREHL_RAW, arm);

            code = sph_strehl_zimpol(qclist, image, image, rawframe,
                    filterframe, "test", parlist);
            cpl_test_eq_error(code, CPL_ERROR_NONE);
        }

        cpl_propertylist_dump(qclist, stream);

        cpl_image_delete(image);
        cpl_image_delete(imtmp);
        cpl_frame_delete(rawframe);
        cpl_frame_delete(filterframe);
        cpl_propertylist_delete(qclist);
        cpl_parameterlist_delete(parlist);

        if (stream != stdout)
            cpl_test_zero( fclose(stream));

        if (cpl_test_get_failed() == 0)
            cpl_test_zero(remove(SPH_STREHL_RAW));

    } else {
        cpl_msg_info(cpl_func, "Not testing %s", file);
    }

    cpl_free(file);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @internal
 @brief Fill propertylist corresponding to a raw frame
 @param self   Propertylist to fill
 @param arm    Arm indicator ('I'/'Z')
 @return   void

 */
/*----------------------------------------------------------------------------*/
static void sph_strehl_fill_header(cpl_propertylist * self, int arm) {

    cpl_error_code code;

    code = cpl_propertylist_append_double(self, SPH_STREHL_PSCALE,
            arm == 'I' ? CUTEST_SPH_STREHL_PSCALE_IRDIS : CUTEST_SPH_STREHL_PSCALE_ZIMPOL);
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    /* Set filter both both arms */
    code = cpl_propertylist_append_string(self, SPH_STREHL_FILTER_IRD1,
            "CntK1");
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    code = cpl_propertylist_append_string(self, SPH_STREHL_FILTER_IRD2,
            "CLEAR");
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    code = cpl_propertylist_append_string(self, SPH_STREHL_FILTER_ZPL0, "HeI");
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    code = cpl_propertylist_append_string(self, SPH_STREHL_FILTER_ZPL1, "OPEN");
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    code = cpl_propertylist_append_string(self, SPH_STREHL_FILTER_ZPL2, "Ha");
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @internal
 @brief Create a raw frame
 @param self     The frame to fill
 @param image    Image to save (only dimensions are used)
 @param filename The name of the FITS file
 @param arm    Arm indicator ('I'/'Z')
 @return   void

 */
/*----------------------------------------------------------------------------*/
static void sph_strehl_save_raw(cpl_frame * self, const cpl_image * image,
        const char * filename, int arm) {

    cpl_propertylist * plist = cpl_propertylist_new();
    cpl_error_code code;

    sph_strehl_fill_header(plist, arm);

    code = cpl_image_save(image, filename, CPL_TYPE_UNSPECIFIED, plist,
            CPL_IO_CREATE);
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    code = cpl_frame_set_filename(self, filename);
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    code = cpl_frame_set_group(self, CPL_FRAME_GROUP_RAW);
    cpl_test_eq_error(code, CPL_ERROR_NONE);

    cpl_propertylist_delete(plist);

    return;
}
