/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>

#include "sph_triple_image.h"
#include "sph_common_keywords.h"
#include "sph_zpl_keywords.h"
#include "sph_filemanager.h"

#include <math.h>
#include "sph_error.h"
#include "sph_test.h"
#include <Basic.h>
#include <CUnit.h>
#include "sph_test.h"
#include "sph_test_image_tools.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cutest_sph_triple_image Unit test for the sph_triple_image module
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

static sph_master_frame* create_mframe(double value) {
    sph_master_frame* mframe;
    int nx = 256;
    int ny = 256;
    mframe = sph_master_frame_new(nx, ny);
    cpl_image_delete(mframe->badpixelmap);
    mframe->badpixelmap = NULL;
    mframe->badpixelmap = sph_test_image_tools_create_flat_image_int(nx, ny, 0);
    cpl_image_delete(mframe->rmsmap);
    mframe->rmsmap = NULL;
    mframe->rmsmap = sph_test_image_tools_create_flat_image_double(nx, ny,
            value);
    cpl_image_delete(mframe->ncombmap);
    mframe->ncombmap = NULL;
    mframe->ncombmap = sph_test_image_tools_create_flat_image_double(nx, ny,
            value);
    cpl_image_delete(mframe->image);
    mframe->image = NULL;
    mframe->image = sph_test_image_tools_create_flat_image_double(nx, ny,
            value);
    return mframe;
}

static
int cutest_init_sph_triple_image_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/

    sph_init_test(__FILE__, PACKAGE_BUGREPORT, CPL_MSG_DEBUG);
    return CPL_ERROR_NONE;
}

static
int cutest_clean_sph_triple_image_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Close and dump errors.
     * -------------------------------------------------------------------*/
    sph_error_dump(SPH_ERROR_ERROR);
    return sph_end_test();

}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_triple_image_new_empty function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_triple_image_new_empty(void) {
    sph_triple_image* trimage = NULL;

    trimage = sph_triple_image_new_empty();
    CU_ASSERT_PTR_NOT_NULL_FATAL( trimage);

    sph_triple_image_delete(trimage);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_triple_image_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_triple_image_new(void) {
    sph_triple_image* trimage = NULL;
    const int nx = 100;
    const int ny = 200;

    trimage = sph_triple_image_new(nx, ny);
    CU_ASSERT_PTR_NOT_NULL_FATAL( trimage);

    CU_ASSERT_EQUAL( cpl_image_get_size_x( trimage->iframe->image ), nx);
    CU_ASSERT_EQUAL( cpl_image_get_size_y( trimage->iframe->badpixelmap ), ny);
    CU_ASSERT_EQUAL( cpl_image_get_size_x( trimage->qframe->image ), nx);
    CU_ASSERT_EQUAL( cpl_image_get_size_y( trimage->qframe->badpixelmap ), ny);
    CU_ASSERT_EQUAL( cpl_image_get_size_x( trimage->uframe->image ), nx);
    CU_ASSERT_EQUAL( cpl_image_get_size_y( trimage->uframe->badpixelmap ), ny);

    sph_triple_image_delete(trimage);
    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE);

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_triple_image_new_from_master_frames function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_triple_image_new_from_master_frames(void) {
    sph_triple_image* trimage = NULL;
    sph_master_frame* mframe1 = NULL;
    sph_master_frame* mframe2 = NULL;
    sph_master_frame* mframe3 = NULL;
    cpl_error_code rerr = CPL_ERROR_NONE;

    mframe1 = create_mframe(10.0);
    mframe2 = create_mframe(20.0);
    mframe3 = create_mframe(30.0);

    trimage = sph_triple_image_new_from_master_frames(mframe1, mframe2,
            mframe3);
    CU_ASSERT_PTR_NOT_NULL_FATAL( trimage);

    rerr = sph_triple_image_quality_check(trimage);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_IFRAME),
            10.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_QFRAME),
            20.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_UFRAME),
            30.0, 1e-20);
    sph_triple_image_delete(trimage);

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_triple_image_duplicate function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_triple_image_duplicate(void) {
    sph_triple_image* trimage = NULL;
    sph_triple_image* trimage_copy = NULL;
    sph_master_frame* mframe1 = NULL;
    sph_master_frame* mframe2 = NULL;
    sph_master_frame* mframe3 = NULL;
    cpl_error_code rerr = CPL_ERROR_NONE;

    mframe1 = create_mframe(10.0);
    mframe2 = create_mframe(20.0);
    mframe3 = create_mframe(30.0);

    trimage = sph_triple_image_new_from_master_frames(mframe1, mframe2,
            mframe3);
    CU_ASSERT_PTR_NOT_NULL_FATAL( trimage);
    trimage_copy = sph_triple_image_duplicate(trimage);
    sph_triple_image_delete(trimage);

    rerr = sph_triple_image_quality_check(trimage_copy);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage_copy->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_IFRAME),
            10.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage_copy->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_QFRAME),
            20.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage_copy->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_UFRAME),
            30.0, 1e-20);
    sph_triple_image_delete(trimage_copy);

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_triple_image_save_load function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_triple_image_save_load(void) {
    sph_triple_image* trimage = NULL;
    sph_master_frame* mframe1 = NULL;
    sph_master_frame* mframe2 = NULL;
    sph_master_frame* mframe3 = NULL;
    cpl_error_code rerr = CPL_ERROR_NONE;

    mframe1 = create_mframe(10.0);
    mframe2 = create_mframe(20.0);
    mframe3 = create_mframe(30.0);

    trimage = sph_triple_image_new_from_master_frames(mframe1, mframe2,
            mframe3);
    CU_ASSERT_PTR_NOT_NULL_FATAL( trimage);

    //save triple image without qc parameters
    rerr = sph_triple_image_save(trimage, "test_trimage.fits.tmp", NULL);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    rerr = sph_triple_image_delete(trimage);
    trimage = NULL;
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    //load triple image & calculate qc params to test save/load triple image functions
    trimage = sph_triple_image_load("test_trimage.fits.tmp", 0);

    rerr = sph_triple_image_quality_check(trimage);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_IFRAME),
            10.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_QFRAME),
            20.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_UFRAME),
            30.0, 1e-20);

    //save triple image with qc params to check save/load triple image functions (with qc params)
    rerr = sph_triple_image_save(trimage, "test_trimage.fits.tmp", NULL);
    sph_triple_image_delete(trimage);
    trimage = NULL;
    trimage = sph_triple_image_load("test_trimage.fits.tmp", 0);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_IFRAME),
            10.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_QFRAME),
            20.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_UFRAME),
            30.0, 1e-20);
    sph_triple_image_delete(trimage);
    trimage = NULL;

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_triple_image_save_load function.
 */
/*----------------------------------------------------------------------------*/
void cutest_sph_triple_image_add_subtract_divide(void) {
    sph_triple_image* trimage = NULL;
    sph_triple_image* trimage2 = NULL;
    sph_master_frame* mframe1 = NULL;
    sph_master_frame* mframe2 = NULL;
    sph_master_frame* mframe3 = NULL;
    cpl_error_code rerr = CPL_ERROR_NONE;

    mframe1 = create_mframe(10.0);
    mframe2 = create_mframe(20.0);
    mframe3 = create_mframe(30.0);

    trimage = sph_triple_image_new_from_master_frames(mframe1, mframe2,
            mframe3);
    CU_ASSERT_PTR_NOT_NULL_FATAL( trimage);
    trimage2 = sph_triple_image_new_from_master_frames(create_mframe(100.0),
            create_mframe(200.0), create_mframe(300.0));
    CU_ASSERT_PTR_NOT_NULL_FATAL( trimage2);

    rerr = sph_triple_image_add_triple_image(trimage, trimage2);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
    rerr = sph_triple_image_quality_check(trimage);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_IFRAME),
            110.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_QFRAME),
            220.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_UFRAME),
            330.0, 1e-20);

    rerr = sph_triple_image_subtract_triple_image(trimage, trimage2);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
    rerr = sph_triple_image_quality_check(trimage);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_IFRAME),
            10.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_QFRAME),
            20.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_UFRAME),
            30.0, 1e-20);

    rerr = sph_triple_image_divide_double(trimage, 10.0);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
    rerr = sph_triple_image_quality_check(trimage);
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);

    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_IFRAME),
            1.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_QFRAME),
            2.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double(trimage->qclist, SPH_COMMON_KEYWORD_QC_MEAN_TRIPLEIMAGE_UFRAME),
            3.0, 1e-20);

    sph_triple_image_delete(trimage);
    trimage = NULL;
    sph_triple_image_delete(trimage2);
    trimage2 = NULL;

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief   Main to run the tests.
 */
/*----------------------------------------------------------------------------*/
int main(void) {
    int result = 0;
    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("cutest_sph_triple_image",
            cutest_init_sph_triple_image_testsuite,
            cutest_clean_sph_triple_image_testsuite);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* add the tests to the suite */

    if (NULL
            == CU_add_test(pSuite, "sph_triple_image_new_empty",
                    cutest_sph_triple_image_new_empty)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_triple_image_new",
                    cutest_sph_triple_image_new)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_triple_image_new_from_master_frames",
                    cutest_sph_triple_image_new_from_master_frames)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_triple_image_duplicate",
                    cutest_sph_triple_image_duplicate)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_triple_image_save_load",
                    cutest_sph_triple_image_save_load)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_triple_image_add_subtract_divide",
                    cutest_sph_triple_image_add_subtract_divide)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();
    sph_test_create_results_file("results.txt");
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
