/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cpl.h>

#include "sph_zpl_exposure_imaging.h"
#include "sph_utils.h"
#include "sph_zpl_keywords.h"
#include "sph_common_keywords.h"
#include "sph_keyword_manager.h"
#include "sph_filemanager.h"
#include "sph_master_frame.h"
#include "sph_framecombination.h"
#include "sph_error.h"
#include "sph_test.h"
#include "sph_test_image_tools.h"
#include "sph_test_zpl_simulator.h"
#include <Basic.h>
#include <CUnit.h>

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cutest_zpl_exposure_imaging Unit test of sph_zpl_exposure_imaging
 *
 */
/*----------------------------------------------------------------------------*/
static
void cutest_zpl_exposure_imaging_create_head_file(const char* infile) {
    FILE* fp;
    char outfilename[256];

    sprintf(outfilename, "%s.head", infile);
    fp = fopen(outfilename, "w");
    if (fp != NULL) {
        fprintf(fp, "ESO DRS ZPL TEST IMAGING = 48.5\n");
        fclose(fp);
    }

}

sph_keyword_manager* cutest_zpl_exposure_imaging_create_cfg_file(void) {
    FILE* fp;
    sph_keyword_manager* km = NULL;

    fp = fopen("sphere.cfg", "w");
    fprintf(
            fp,
            "SPH.KEYWORDS.ESO_DRS_ZPL_TEST_IMAGING.NAME = \"ESO DRS ZPL TEST IMAGING\"\n");
    fprintf(fp, "SPH.KEYWORDS.ESO_DRS_ZPL_TEST_IMAGING.TYPE = \"DOUBLE\"\n");
    fprintf(fp, "SPH.KEYWORDS.ESO_DRS_ZPL_TEST_IMAGING.VALUE = \"10.0\"\n");
    fclose(fp);

    km = sph_keyword_manager_new();
    sph_keyword_manager_read_config("sphere.cfg");

    return km;
}

static int cutest_init_zpl_exposure_imaging_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__, PACKAGE_BUGREPORT, CPL_MSG_DEBUG);
    unlink("sphere.cfg");
    return CPL_ERROR_NONE;
}

static int cutest_clean_zpl_exposure_imaging_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Close and dump errors.
     * -------------------------------------------------------------------*/
    sph_error_dump(SPH_ERROR_ERROR);
    return sph_end_test();
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_new_empty function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_new_empty(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;

    zplexp = sph_zpl_exposure_imaging_new_empty();
    CU_ASSERT_PTR_NOT_NULL_FATAL( zplexp);

    CU_ASSERT_TRUE(
            cpl_propertylist_has( zplexp->properties, SPH_COMMON_KEYWORD_SPH_TYPE ));
    CU_ASSERT_STRING_EQUAL(
            cpl_propertylist_get_string( zplexp->properties, SPH_COMMON_KEYWORD_SPH_TYPE ),
            SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_PREPROC_ZPL_EXP_IMG);
    sph_zpl_exposure_imaging_delete(zplexp);
    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE);

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_new(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    const int nx = 100;
    const int ny = 200;

    zplexp = sph_zpl_exposure_imaging_new(nx, ny);
    CU_ASSERT_PTR_NOT_NULL_FATAL( zplexp);

    CU_ASSERT_TRUE(
            cpl_propertylist_has( zplexp->properties, SPH_COMMON_KEYWORD_SPH_TYPE ));
    CU_ASSERT_STRING_EQUAL(
            cpl_propertylist_get_string( zplexp->properties, SPH_COMMON_KEYWORD_SPH_TYPE ),
            SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_PREPROC_ZPL_EXP_IMG);

    CU_ASSERT_EQUAL( cpl_image_get_size_x( zplexp->image_odd ), nx);
    CU_ASSERT_EQUAL( cpl_image_get_size_y( zplexp->image_odd ), ny);

    CU_ASSERT_EQUAL( cpl_image_get_size_x( zplexp->image_even ), nx);
    CU_ASSERT_EQUAL( cpl_image_get_size_y( zplexp->image_even ), ny);

    sph_zpl_exposure_imaging_delete(zplexp);
    CU_ASSERT_EQUAL( cpl_error_get_code(), CPL_ERROR_NONE);

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_save function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_save(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    sph_zpl_exposure_imaging* zexpload = NULL;
    cpl_frame* zexpframe = NULL;

    zplexp = sph_test_zpl_simulator_create_exposure_imaging(100, 100);

    /* internal test of ph_test_zpl_simulator_create_exposure_imaging & _2flat */
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zplexp->image_odd ),
            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zplexp->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    // save zpl exp imaging frame
    zexpframe = sph_filemanager_create_temp_frame("test_zplexp_imaging.fits",
            "PREPROC IMAGING", CPL_FRAME_GROUP_NONE);
    sph_zpl_exposure_imaging_save(zplexp, zexpframe, NULL);
    //printf("Filename fo zexpframe: %s \n", cpl_frame_get_filename( zexpframe ));
    //load a zpl exposure imaging from pure zpl exp imaging frame
    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 0);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ),
            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);
    cpl_frame_delete(zexpframe);
    sph_zpl_exposure_imaging_delete(zplexp);
    sph_zpl_exposure_imaging_delete(zexpload);
    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_save function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_save_load_with_ovsc_table(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    sph_zpl_exposure_imaging* zexpload = NULL;
    cpl_frame* zexpframe = NULL;
    cpl_propertylist* ovsc_pl = NULL;
    double a[] = { 100.0, 1.0, 200.0, 2.0 };

    zplexp = sph_test_zpl_simulator_create_exposure_imaging(100, 100);

    /* internal test of ph_test_zpl_simulator_create_exposure_imaging & _2flat */
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zplexp->image_odd ),
            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zplexp->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    // save zpl exp imaging frame
    zexpframe = sph_filemanager_create_temp_frame("test_zplexp_imaging.fits",
            "PREPROC IMAGING", CPL_FRAME_GROUP_NONE);
    ovsc_pl = sph_test_zpl_simulator_create_ovsc_propertylist_imaging(a);
    //printf("ovsc_pl: ADU1_OVSC_MEAN = %f \n", cpl_propertylist_get_double( ovsc_pl, ADU1_OVSC_MEAN ) );

    if (zplexp->ovsc)
        cpl_propertylist_delete(zplexp->ovsc);
    zplexp->ovsc = cpl_propertylist_duplicate(ovsc_pl);
    cpl_propertylist_delete(ovsc_pl);
    sph_zpl_exposure_imaging_save(zplexp, zexpframe, NULL);

    //printf("Filename fo zexpframe: %s \n", cpl_frame_get_filename( zexpframe ));
    //load a zpl exposure imaging from pure zpl exp imaging frame
    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 0);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ),
            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpload->ovsc, ADU1_OVSC_MEAN ), a[0],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpload->ovsc, ADU1_OVSC_RMS ), a[1],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpload->ovsc, ADU2_OVSC_MEAN ), a[2],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpload->ovsc, ADU2_OVSC_RMS ), a[3],
            1e-20);

    sph_zpl_exposure_imaging_delete(zplexp);
    sph_zpl_exposure_imaging_delete(zexpload);
    cpl_frame_delete(zexpframe);
    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_save_open and _append function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_save_append(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    sph_zpl_exposure_imaging* zplexp_append = NULL;
    sph_zpl_exposure_imaging* zexpload = NULL;
    cpl_frame* zexpframe = NULL;

    zplexp = sph_test_zpl_simulator_create_exposure_imaging(100, 100);
    CU_ASSERT_PTR_NOT_NULL_FATAL(zplexp);

    /* internal test of ph_test_zpl_simulator_create_exposure_imaging & _2flat */
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zplexp->image_odd ),
            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zplexp->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    // save open zpl exp imaging frame: plane = 0
    zexpframe = sph_filemanager_create_temp_frame("test_zplexp_imaging.fits",
            "PREPROC IMAGING", CPL_FRAME_GROUP_NONE);
    sph_zpl_exposure_imaging_save_open(zplexp, zexpframe, NULL, NULL);
    //printf("Filename fo zexpframe: %s \n", cpl_frame_get_filename( zexpframe ));
    // save append zpl expimaging : plane = 1

    zplexp_append = sph_test_zpl_simulator_create_exposure_imaging_2flat_double(
            100, 100, 10.0, 20.0);
    sph_zpl_exposure_imaging_save_append(zplexp_append, zexpframe, NULL);
    //sph_cube_finalise_file( cpl_frame_get_filename( zexpframe ));
    sph_zpl_exposure_imaging_finalize_file(zexpframe, NULL);

    //load a zpl exposure imaging from zpl exp imaging frame cube: plane = 0
    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 0);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ),
            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);
    sph_zpl_exposure_imaging_delete(zexpload);

    //load a zpl exposure imaging from zpl exp imaging frame cube: plane = 1
    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 1);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ), 10.0,
                            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ), 20.0, 1e-20);

    cpl_frame_delete(zexpframe);
    sph_zpl_exposure_imaging_delete(zplexp);
    sph_zpl_exposure_imaging_delete(zplexp_append);
    sph_zpl_exposure_imaging_delete(zexpload);
    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_load function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_load(void) {
    sph_zpl_exposure_imaging* zexpload = NULL;
    cpl_frame* zexpcube = NULL;

    //create flat image cube
    zexpcube = sph_test_zpl_simulator_create_exposure_imaging_cube_frame(100,
            100, 5, "test_zplexp_imaging_cube.fits");

    //load a zpl exposure imaging from zpl exp imaging cube frame
    CU_ASSERT_PTR_NOT_NULL_FATAL(zexpcube);

    //load zpl exposure imaging at the plane number 2 from cube of zpl imaging exposures
    zexpload = sph_zpl_exposure_imaging_load(zexpcube, 2);
    CU_ASSERT_PTR_NOT_NULL_FATAL(zexpload);

    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ),
                            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    cpl_frame_delete(zexpcube);
    sph_zpl_exposure_imaging_delete(zexpload);
    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_test_zpl_simulator_create_exposure_imaging_2flat_double function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_frames_double(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    cpl_frame* zexpframe = NULL;
    cpl_frame* zexpcube = NULL;
    cpl_frameset* zexpframes = NULL;
    sph_error_code rerr = CPL_ERROR_NONE;
    cpl_propertylist* pl = NULL;
    sph_master_frame* mf_image_odd = NULL;
    //sph_master_frame*    mf_image_even= NULL;
    cpl_parameterlist* parlist = NULL;
    cpl_parameter* p = NULL;

    const int nz = 5; //nframes
    const int nx = 100;
    const int ny = 100;
    double image_odd = 100.0;
    double image_even = 200.0;
    double dV = 100.0;

    pl = cpl_propertylist_new();
    cpl_propertylist_append_int(pl, SPH_ZPL_KEYWORD_CAMERA_NAME,
            SPH_ZPL_KEYWORD_VALUE_CAMERA1_ID);
    zexpcube = sph_test_zpl_simulator_frame_template(
            "test_zplexpimg_cube_cam1_double.fits");

    zexpframes = cpl_frameset_new();
    for (int i = 0; i < nz; i++) {
        zplexp = sph_test_zpl_simulator_create_exposure_imaging_2flat_double(nx,
                ny, image_odd + i * dV, image_even + i * dV);

        zexpframe = sph_filemanager_create_temp_frame(
                "test_zplexpimg_single_frame.fits", "NONE",
                CPL_FRAME_GROUP_NONE);
        rerr = sph_zpl_exposure_imaging_save(zplexp, zexpframe, NULL);

        CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
        if (rerr != CPL_ERROR_NONE) {
            SPH_ERR( "could not save zpl exposure imaging");
            return;
        }
        cpl_frameset_insert(zexpframes, zexpframe);
        sph_zpl_exposure_imaging_delete(zplexp);
    }CU_ASSERT_EQUAL(cpl_frameset_get_size( zexpframes ), nz);

    rerr =
            sph_test_zpl_simulator_create_cube_from_exposure_imaging_frames_double(
                    zexpframes, zexpcube);
    //CU_ASSERT_NOT_EQUAL_FATAL(rerr, CPL_ERROR_NONE);

    parlist = cpl_parameterlist_new();
    cpl_parameterlist_append(parlist, p);
    p = cpl_parameter_new_enum("zpl.test.coll_alg", CPL_TYPE_INT,
            "No description", "zpl.test", 1, 3, 0, 1, 2);
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);

    cpl_parameterlist_append(parlist, p);
    p = cpl_parameter_new_range("zpl.test.coll_alg.clean_mean.reject_high",
            CPL_TYPE_INT, "No description", "zpl.test", 0, 0, 20);
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);

    cpl_parameterlist_append(parlist, p);
    p = cpl_parameter_new_range("zpl.test.coll_alg.clean_mean.reject_low",
            CPL_TYPE_INT, "No description", "zpl.test", 0, 0, 20);
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);

    cpl_parameterlist_append(parlist, p);
    p = cpl_parameter_new_range("zpl.test.clean_mean.sigma", CPL_TYPE_DOUBLE,
            "No description", "zpl.test", 5.0, 0.0, 20.);
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(parlist, p);

    mf_image_odd =
            sph_framecombination_new_master_frame_from_cpl_frame_multi_ext(
                    zexpcube, 0, SPH_COLL_ALG_MEAN, parlist);
    CU_ASSERT_PTR_NOT_NULL_FATAL(mf_image_odd);
    sph_master_frame_save(mf_image_odd, "test_mf_image_odd.fits.tmp", NULL);
    //printf("absflux: %.3f\n", cpl_image_get_absflux( mf_zero_odd->image ));
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_absflux( mf_image_odd->image ),
            300.0*nx*ny, 1e-20);

    cpl_frame_delete(zexpcube);
    cpl_frameset_delete(zexpframes);
    cpl_parameterlist_delete(parlist);
    cpl_propertylist_delete(pl);

    sph_master_frame_delete(mf_image_odd);
    sph_filemanager_clean();

    return;

}

void cutest_test_zpl_exposure_imaging_new_from_images(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    sph_zpl_exposure_imaging* zexpload = NULL;
    cpl_frame* zexpframe = NULL;
    cpl_image* image_odd = NULL;
    cpl_image* image_even = NULL;

    image_odd = sph_test_image_tools_create_flat_image_double(100, 100, 10.0);
    image_even = sph_test_image_tools_create_flat_image_double(100, 100, 11.0);

    zplexp = sph_zpl_exposure_imaging_new_from_cplimages(image_odd, image_even);
    cpl_image_delete(image_odd);
    cpl_image_delete(image_even);
    CU_ASSERT_PTR_NOT_NULL_FATAL(zplexp);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zplexp->image_odd ), 10.0,
                            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
                           cpl_image_get_mean( zplexp->image_even ), 11.0, 1e-20);

    // save zpl exp frame
    zexpframe = sph_filemanager_create_temp_frame("test_zplexp_imaging.fits",
            "PREPROC", CPL_FRAME_GROUP_NONE);
    sph_zpl_exposure_imaging_save(zplexp, zexpframe, NULL);
    //load a zpl exposure imaging from pure zpl exp frame
    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 0);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ), 10.0,
                            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ), 11.0, 1e-20);

    sph_zpl_exposure_imaging_delete(zplexp);
    sph_zpl_exposure_imaging_delete(zexpload);
    cpl_frame_delete(zexpframe);

    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_load function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_load_with_head(void) {
    sph_zpl_exposure_imaging* zexpload = NULL;
    sph_zpl_exposure_imaging* zexptest = NULL;
    cpl_frame* test_frame = NULL;
    sph_keyword_manager* keyman = NULL;
    cpl_frame* zexpcube = NULL;

    cpl_error_reset();

    //create sphere.cfg
    keyman = cutest_zpl_exposure_imaging_create_cfg_file();

    CU_ASSERT_PTR_NOT_NULL_FATAL( keyman);

    //create flat image cube
    zexpcube = sph_test_zpl_simulator_create_exposure_imaging_cube_frame(100,
            100, 5, "test_zplexp_imaging_cube.fits");

    //load a zpl exposure imaging from zpl exp imaging cube frame
    CU_ASSERT_PTR_NOT_NULL_FATAL(zexpcube);

    //create .head file
    cutest_zpl_exposure_imaging_create_head_file("test_zplexp_imaging_cube");

    //load zpl exposure imaging at the plane number 2 from cube of zpl imaging exposures
    zexpload = sph_zpl_exposure_imaging_load(zexpcube, 0);
    CU_ASSERT_PTR_NOT_NULL_FATAL(zexpload);

    CU_ASSERT_TRUE(
            cpl_propertylist_has( zexpload->properties, "ESO DRS ZPL TEST IMAGING"));
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpload->properties, "ESO DRS ZPL TEST IMAGING"),
            48.5, 0.001);

        CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ),
                                ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    test_frame = sph_filemanager_create_temp_frame(
            "test_zplexp_imaging.fits.tmp", "NONE", CPL_FRAME_GROUP_NONE);
    sph_zpl_exposure_imaging_save(zexpload, test_frame, NULL);

    unlink("test_zplexp_cube_imaging.fits");
    unlink("test_zplexp_cube_imaging.head");

    zexptest = sph_zpl_exposure_imaging_load(test_frame, 0);
    CU_ASSERT_TRUE(
            cpl_propertylist_has( zexpload->properties, "ESO DRS ZPL TEST IMAGING"));
    CU_ASSERT_DOUBLE_EQUAL(cpl_propertylist_get_double( zexpload->properties,
                                                        "ESO DRS ZPL TEST IMAGING"),
                           48.5, 0.001);

    cpl_frame_delete(zexpcube);
    sph_zpl_exposure_imaging_delete(zexpload);
    cpl_frame_delete(test_frame);
    sph_zpl_exposure_imaging_delete(zexptest);

    sph_keyword_manager_delete();

    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_save function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_duplicate_with_ovsc_table(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    sph_zpl_exposure_imaging* zexpload = NULL;
    sph_zpl_exposure_imaging* zexpdupl = NULL;
    cpl_frame* zexpframe = NULL;
    cpl_propertylist* ovsc_pl = NULL;
    double a[] = { 100.0, 1.0, 200.0, 2.0 };

    zplexp = sph_test_zpl_simulator_create_exposure_imaging(100, 100);

    /* internal test of ph_test_zpl_simulator_create_exposure_imaging & _2flat */
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zplexp->image_odd ),
            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zplexp->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    // save zpl exp imaging frame
    zexpframe = sph_filemanager_create_temp_frame(
            "test_zplexp_imaging_duplicate.fits", "PREPROC IMAGING",
            CPL_FRAME_GROUP_NONE);
    ovsc_pl = sph_test_zpl_simulator_create_ovsc_propertylist_imaging(a);
    //printf("ovsc_pl: ADU1_OVSC_MEAN = %f \n", cpl_propertylist_get_double( ovsc_pl, ADU1_OVSC_MEAN ) );

    if (zplexp->ovsc)
        cpl_propertylist_delete(zplexp->ovsc);
    zplexp->ovsc = cpl_propertylist_duplicate(ovsc_pl);
    cpl_propertylist_delete(ovsc_pl);
    sph_zpl_exposure_imaging_save(zplexp, zexpframe, NULL);

    //printf("Filename fo zexpframe: %s \n", cpl_frame_get_filename( zexpframe ));
    //load a zpl exposure imaging from pure zpl exp imaging frame
    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 0);
    zexpdupl = sph_zpl_exposure_imaging_duplicate(zexpload);

    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpdupl->image_odd ),
                            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpdupl->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpdupl->ovsc, ADU1_OVSC_MEAN ), a[0],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpdupl->ovsc, ADU1_OVSC_RMS ), a[1],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpdupl->ovsc, ADU2_OVSC_MEAN ), a[2],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpdupl->ovsc, ADU2_OVSC_RMS ), a[3],
            1e-20);

    sph_zpl_exposure_imaging_delete(zplexp);
    sph_zpl_exposure_imaging_delete(zexpload);
    sph_zpl_exposure_imaging_delete(zexpdupl);
    cpl_frame_delete(zexpframe);
    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_open/_append/_finalize functions.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_open_append_finalize_with_ovsc_table(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    sph_zpl_exposure_imaging* zplexp_append = NULL;
    sph_zpl_exposure_imaging* zexpload = NULL;
    cpl_frame* zexpframe = NULL;
    cpl_propertylist* ovsc_pl = NULL;
    cpl_table* ovsc_table = NULL;
    double a[8] = { 100.0, 1.0, 200.0, 2.0 };
    double b[8] = { 110.0, 11.0, 210.0, 12.0 };

    zplexp = sph_test_zpl_simulator_create_exposure_imaging(100, 100);
    CU_ASSERT_PTR_NOT_NULL_FATAL(zplexp);

    /* internal test of ph_test_zpl_simulator_create_exposure_imaging & _2flat */
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zplexp->image_odd ),
                            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zplexp->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    // save open zpl exp imaging frame: plane = 0
    zexpframe = sph_filemanager_create_temp_frame(
            "test_zplexp_imaging_finalize.fits", "PREPROC IMAGING",
            CPL_FRAME_GROUP_NONE);
    ovsc_pl = sph_test_zpl_simulator_create_ovsc_propertylist_imaging(a);
    if (zplexp->ovsc)
        cpl_propertylist_delete(zplexp->ovsc);
    zplexp->ovsc = cpl_propertylist_duplicate(ovsc_pl);
    cpl_propertylist_delete(ovsc_pl);
    ovsc_pl = NULL;

    ovsc_table = sph_zpl_exposure_imaging_ovsc_table_create_empty();
    sph_zpl_exposure_imaging_save_open(zplexp, zexpframe, NULL, ovsc_table);
    //printf("Filename fo zexpframe: %s \n", cpl_frame_get_filename( zexpframe ));
    // save append zpl expimaging : plane = 1

    zplexp_append = sph_test_zpl_simulator_create_exposure_imaging_2flat_double(
            100, 100, 10.0, 20.0);
    ovsc_pl = sph_test_zpl_simulator_create_ovsc_propertylist_imaging(b);
    if (zplexp_append->ovsc)
        cpl_propertylist_delete(zplexp_append->ovsc);
    zplexp_append->ovsc = cpl_propertylist_duplicate(ovsc_pl);
    cpl_propertylist_delete(ovsc_pl);
    ovsc_pl = NULL;

    sph_zpl_exposure_imaging_save_append(zplexp_append, zexpframe, ovsc_table);

    sph_zpl_exposure_imaging_finalize_file(zexpframe, ovsc_table);
    //load a zpl exposure imaging from zpl exp imaging frame cube: plane = 0

    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 0);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ),
                            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_propertylist_get_double( zexpload->ovsc,
                                                        ADU1_OVSC_MEAN ), a[0],
                           1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_propertylist_get_double( zexpload->ovsc, ADU1_OVSC_RMS ),
                           a[1], 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_propertylist_get_double( zexpload->ovsc, ADU2_OVSC_MEAN ),
                           a[2], 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_propertylist_get_double( zexpload->ovsc, ADU2_OVSC_RMS ),
                           a[3], 1e-20);
    sph_zpl_exposure_imaging_delete(zexpload);

    //load a zpl exposure imaging from zpl exp imaging frame cube: plane = 1
    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 1);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ), 10.0,
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zexpload->image_even ), 20.0, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_propertylist_get_double( zexpload->ovsc, ADU1_OVSC_MEAN ),
                           b[0], 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpload->ovsc, ADU1_OVSC_RMS ), b[1],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpload->ovsc, ADU2_OVSC_MEAN ), b[2],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpload->ovsc, ADU2_OVSC_RMS ), b[3],
            1e-20);
    sph_zpl_exposure_imaging_delete(zexpload);

    cpl_frame_delete(zexpframe);
    sph_zpl_exposure_imaging_delete(zplexp);
    sph_zpl_exposure_imaging_delete(zplexp_append);
    cpl_table_delete(ovsc_table);
    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_zpl_exposure_imaging_save function.
 */
/*----------------------------------------------------------------------------*/
void cutest_test_zpl_exposure_imaging_subtract_overscans(void) {
    sph_zpl_exposure_imaging* zplexp = NULL;
    sph_zpl_exposure_imaging* zexpload = NULL;
    sph_zpl_exposure_imaging* zexpdupl = NULL;
    cpl_frame* zexpframe = NULL;
    cpl_propertylist* ovsc_pl = NULL;
    double a[] = { 10.0, 1.0, 20.0, 2.0 };

    zplexp = sph_test_zpl_simulator_create_exposure_imaging(100, 100);

    /* internal test of ph_test_zpl_simulator_create_exposure_imaging & _2flat */
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zplexp->image_odd ),
            ZERO_ODD_DOUBLE, 1e-20);
    CU_ASSERT_DOUBLE_EQUAL(cpl_image_get_mean( zplexp->image_even ),
                           ZERO_EVEN_DOUBLE, 1e-20);

    // save zpl exp imaging frame
    zexpframe = sph_filemanager_create_temp_frame(
            "test_zplexp_imaging_subtract_overscans.fits", "PREPROC IMAGING",
            CPL_FRAME_GROUP_NONE);
    ovsc_pl = sph_test_zpl_simulator_create_ovsc_propertylist_imaging(a);
    //printf("ovsc_pl: ADU1_OVSC_MEAN = %f \n", cpl_propertylist_get_double( ovsc_pl, ADU1_OVSC_MEAN ) );
    if (zplexp->ovsc)
        cpl_propertylist_delete(zplexp->ovsc);
    zplexp->ovsc = cpl_propertylist_duplicate(ovsc_pl);
    cpl_propertylist_delete(ovsc_pl);

    //subtract overscans & save
    sph_zpl_exposure_imaging_subtract_overscans(zplexp);
    sph_zpl_exposure_imaging_save(zplexp, zexpframe, NULL);

    //printf("Filename fo zexpframe: %s \n", cpl_frame_get_filename( zexpframe ));
    //load a zpl exposure imaging from pure zpl exp imaging frame
    zexpload = sph_zpl_exposure_imaging_load(zexpframe, 0);
    zexpdupl = sph_zpl_exposure_imaging_duplicate(zexpload);

    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_odd ),
            (ZERO_ODD_DOUBLE - a[0] + ZERO_ODD_DOUBLE - a[2])/2.0, 1e-10);
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_mean( zexpload->image_even ),
                            (ZERO_EVEN_DOUBLE - a[0] + ZERO_EVEN_DOUBLE - a[2] )/2.0,
                            1e-10);

    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpdupl->ovsc, ADU1_OVSC_MEAN ), a[0],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpdupl->ovsc, ADU1_OVSC_RMS ), a[1],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpdupl->ovsc, ADU2_OVSC_MEAN ), a[2],
            1e-20);
    CU_ASSERT_DOUBLE_EQUAL(
            cpl_propertylist_get_double( zexpdupl->ovsc, ADU2_OVSC_RMS ), a[3],
            1e-20);

    sph_zpl_exposure_imaging_delete(zplexp);
    sph_zpl_exposure_imaging_delete(zexpload);
    sph_zpl_exposure_imaging_delete(zexpdupl);
    cpl_frame_delete(zexpframe);

    sph_filemanager_clean();

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test MAIN function
 */
/*----------------------------------------------------------------------------*/
int main(void) {
    int result = 0;
    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("cutest_zpl_exposure_imaging",
            cutest_init_zpl_exposure_imaging_testsuite,
            cutest_clean_zpl_exposure_imaging_testsuite);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* add the tests to the suite */

    if (NULL
            == CU_add_test(pSuite, "sph_zpl_exposure_imaging_new_empty",
                    cutest_test_zpl_exposure_imaging_new_empty)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_zpl_exposure_imaging_new",
                    cutest_test_zpl_exposure_imaging_new)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_zpl_exposure_imaging_save",
                    cutest_test_zpl_exposure_imaging_save)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_zpl_exposure_imaging_load",
                    cutest_test_zpl_exposure_imaging_load)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    //test for the sph_test_zpl_simulator_create_exposure_imaging_2flat_double
    if (NULL
            == CU_add_test(pSuite, "sph_zpl_exposure_imaging_frames_double",
                    cutest_test_zpl_exposure_imaging_frames_double)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_zpl_exposure_imaging_new_from_images",
                    cutest_test_zpl_exposure_imaging_new_from_images)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_zpl_exposure_imaging_save_append",
                    cutest_test_zpl_exposure_imaging_save_append)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite, "sph_zpl_exposure_imaging_load_with_head",
                    cutest_test_zpl_exposure_imaging_load_with_head)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite,
                    "sph_zpl_exposure_imaging_save_load_with_ovsc_tabl",
                    cutest_test_zpl_exposure_imaging_save_load_with_ovsc_table)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite,
                    "sph_zpl_exposure_imaging_duplicate_with_ovsc_tabl",
                    cutest_test_zpl_exposure_imaging_duplicate_with_ovsc_table)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(
                    pSuite,
                    "sph_zpl_exposure_imaging_open_append_finalize_with_ovsc_table",
                    cutest_test_zpl_exposure_imaging_open_append_finalize_with_ovsc_table)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    if (NULL
            == CU_add_test(pSuite,
                    "sph_zpl_exposure_imaging_subtract_overscans",
                    cutest_test_zpl_exposure_imaging_subtract_overscans)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();
    sph_test_create_results_file("results.txt");
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}
