/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>

#include "sph_dataset_stats.h"
#include "sph_dataset.h"
#include "sph_common_keywords.h"
#include <math.h>
#include "sph_error.h"
#include "sph_test.h"
#include <Basic.h>
#include <CUnit.h>
#include "sph_test.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

/*----------------------------------------------------------------------------*/
/**
 * @defgroup Requirement Test: sph_dataset_peak_find_subpeaks
 * 
 * The requirement tests for the function sph_dataset_peak_find_subpeaks              
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

sph_dataset*
reqtest_util_dataset_new_load(void) {
    sph_dataset* dat = NULL;
    cpl_vector* lams = NULL;
    cpl_vector* vals = NULL;
    int ii = 0;

    dat = sph_dataset_new(100, 1.0, 100.0);
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat);

    lams = dat->xvalues;
    vals = dat->yvalues;
    CU_ASSERT_PTR_NOT_NULL_FATAL( lams);
    CU_ASSERT_PTR_NOT_NULL_FATAL( vals);

    for (ii = 0; ii < cpl_vector_get_size(lams); ++ii) {
        cpl_vector_set(vals, ii, (double) ii);
    }

    CU_ASSERT_EQUAL( cpl_vector_get( lams, 0 ), 1.0);
    CU_ASSERT_EQUAL( cpl_vector_get( lams, 99 ), 100.0);
    return dat;
}
sph_dataset*
reqtest_util_dataset_new_gauss(void) {
    sph_dataset* dat = NULL;
    cpl_vector* lams = NULL;
    cpl_vector* vals = NULL;
    double val = 0.0;
    int ii = 0;

    dat = sph_dataset_new(100, 1.0, 100.0);
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat);

    lams = dat->xvalues;
    vals = dat->yvalues;
    CU_ASSERT_PTR_NOT_NULL_FATAL( lams);
    CU_ASSERT_PTR_NOT_NULL_FATAL( vals);

    for (ii = 0; ii < cpl_vector_get_size(lams); ++ii) {
        val = cpl_vector_get(lams, ii) - 50.0;
        val = exp(-val * val / 25.0);
        cpl_vector_set(vals, ii, val);
    }

    return dat;
}
sph_dataset*
reqtest_util_dataset_new_double_gauss(void) {
    sph_dataset* dat = NULL;
    cpl_vector* lams = NULL;
    cpl_vector* vals = NULL;
    double val = 0.0;
    double valB = 0.0;
    int ii = 0;

    dat = sph_dataset_new(100, 1.0, 100.0);
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat);

    lams = dat->xvalues;
    vals = dat->yvalues;
    CU_ASSERT_PTR_NOT_NULL_FATAL( lams);
    CU_ASSERT_PTR_NOT_NULL_FATAL( vals);

    for (ii = 0; ii < cpl_vector_get_size(lams); ++ii) {
        val = cpl_vector_get(lams, ii) - 25.0;
        val = 5.0 * exp(-val * val / 400.0);
        valB = cpl_vector_get(lams, ii) - 75.0;
        valB = exp(-valB * valB / 400.0);
        val = val + valB;
        cpl_vector_set(vals, ii, val);
    }

    return dat;
}
sph_dataset*
reqtest_util_dataset_new_no_peak(gsl_rng* pRNG) {
    sph_dataset* dat = NULL;
    cpl_vector* lams = NULL;
    cpl_vector* vals = NULL;
    double val = 0.0;
    int ii = 0;

    dat = sph_dataset_new(100, 1.0, 100.0);
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat);

    lams = dat->xvalues;
    vals = dat->yvalues;
    CU_ASSERT_PTR_NOT_NULL_FATAL( lams);
    CU_ASSERT_PTR_NOT_NULL_FATAL( vals);

    for (ii = 0; ii < cpl_vector_get_size(lams); ++ii) {
        val = gsl_ran_gaussian(pRNG, 1.0);
        cpl_vector_set(vals, ii, val);
    }

    return dat;
}
int reqtest_util_dataset_add_noise(sph_dataset* self, gsl_rng* pRNG,
        double noise_sigma) {
    int ii = 0;
    double oldv = 0.0;

    CU_ASSERT_PTR_NOT_NULL_FATAL( self);

    for (ii = 0; ii < cpl_vector_get_size(self->yvalues); ++ii) {
        oldv = cpl_vector_get(self->yvalues, ii);
        oldv = oldv + gsl_ran_gaussian(pRNG, noise_sigma);
        cpl_vector_set(self->yvalues, ii, oldv);
    }

    return CPL_ERROR_NONE;
}

int reqtest_util_write_result_file(const char* fname, sph_dataset_stats* stats) {
    FILE* resfile = NULL;

    resfile = fopen(fname, "w");
    fprintf(resfile, "================= Result of Requirements test ====\n");
    fprintf(resfile, "\n");

    fclose(resfile);
    return 0;
}

int reqtest_init_dataset_stats_find_peaks_testsuite(void) {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/

#if defined CPL_VERSION_CODE && CPL_VERSION_CODE >= CPL_VERSION(4, 0, 0)
    cpl_init(CPL_INIT_DEFAULT);
#else
    cpl_init();
#endif

    cpl_msg_set_time_on();
    cpl_msg_set_component_on();
    cpl_msg_set_domain("Testing");
    cpl_msg_set_domain_on();
    cpl_msg_set_level(CPL_MSG_OFF);
    cpl_msg_set_log_level(CPL_MSG_DEBUG);

    sph_error_new();
    return 0;
}

int reqtest_clean_dataset_stats_find_peaks_testsuite(void) {
    sph_error_dump(SPH_ERROR_ERROR);
    cpl_end();
    return 0;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_dataset_stats_find_peaks function.
 
 This function tests the beahviour for a dataset with one peak and no 
 noise added. 
 */
/*----------------------------------------------------------------------------*/
void reqtest_dataset_stats_find_peaks_one_peak_no_noise(void) {
    sph_dataset_stats* dat_stats = NULL;
    sph_dataset* dat = NULL;
    int rerr = CPL_ERROR_NONE;

    /* Setup and run ...*/
    dat = reqtest_util_dataset_new_gauss();
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat);

    dat_stats = sph_dataset_stats_new(dat);
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat_stats);

    rerr = sph_dataset_stats_find_peaks(dat_stats, 0.1, 3);

    /*Verify */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( dat_stats->npeaks, 1);
    CU_ASSERT_DOUBLE_EQUAL( dat_stats->peakloc[0], 50.0, 0.05);
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_dataset_stats_find_peaks function.
 
 This function tests the beahviour for a dataset with two peaks and 
 a noise added. 
 
 */
/*----------------------------------------------------------------------------*/
void reqtest_dataset_stats_find_peaks_two_peaks_noisy(void) {
    sph_dataset_stats* dat_stats = NULL;
    sph_dataset* dat = NULL;
    int rerr = CPL_ERROR_NONE;

    /* Setup and run ...*/
    dat = reqtest_util_dataset_new_double_gauss();
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat);

    dat_stats = sph_dataset_stats_new(dat);
    CU_ASSERT_PTR_NOT_NULL_FATAL( dat_stats);

    rerr = sph_dataset_stats_find_peaks(dat_stats, 0.1, 3);

    /*Verify */
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE);
    CU_ASSERT_EQUAL( dat_stats->npeaks, 2);
    CU_ASSERT_DOUBLE_EQUAL( dat_stats->peakloc[0], 25.0, 2.05);
    CU_ASSERT_DOUBLE_EQUAL( dat_stats->peakloc[1], 75.0, 2.05);
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit tests of techcal_master_dark recipe and associated functions
 */
/*----------------------------------------------------------------------------*/
int main(void) {
    int result = 0;
    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("dataset_stats_find_peaks_test",
            reqtest_init_dataset_stats_find_peaks_testsuite,
            reqtest_clean_dataset_stats_find_peaks_testsuite);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite,
                    "sph_dataset_stats_find_peaks_one_peak_no_noise",
                    reqtest_dataset_stats_find_peaks_one_peak_no_noise)) {
        CU_cleanup_registry();
        return CU_get_error();
    }
    if (NULL
            == CU_add_test(pSuite,
                    "sph_dataset_stats_find_peaks_two_peaks_noisy",
                    reqtest_dataset_stats_find_peaks_two_peaks_noisy)) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();
    sph_test_create_results_file("results.txt");
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
